/*
 * Part of Xlab Interface
 *
 * Created: 2018-10-26
 * Qt Creator 4.4.1 Based on Qt 5.9.2
 *  Author: Krzysztof Markiewicz
 *  www.obbo.pl
 *
 * This program is distributed under the terms of the GNU General Public License
 */

#include "multimeterdisplay.h"
#include "ui_multimeterdisplay.h"
#include <QFontMetrics>
#include <QtWidgets>
#include <QtMath>
#include <QPixmap>
#include <QClipboard>
#include <QString>
#include <QDoubleValidator>
#include <QLocale>
#include <QPalette>
#include "multimeterdisplayoptions.h"


MultimeterDisplay::MultimeterDisplay(QWidget *parent):
    QWidget(parent),
    ui(new Ui::MultimeterDisplay)
{
    ui->setupUi(this);

    lineEditBase = new QLineEdit(this);
    lineEditBase->setObjectName(QStringLiteral("__qt__passive_label"));
    lineEditBase->setVisible(false);
    lineEditBase->setValidator(new QDoubleValidator(this));
    background = new QWidget(this);
    background->setObjectName(QStringLiteral("__qt__passive_label"));
    background->setGeometry(ui->output->geometry());
    background->lower();
    ui->output->setAttribute(Qt::WA_TranslucentBackground);
    preparePowers();
    this->alarm.level = ALARM_STATE_OK;
    this->alarm.bottomCritical = -pow(10, 25);
    this->alarm.bottomWarning = -pow(10, 24);
    this->alarm.topWarning = pow(10, 24);
    this->alarm.topCritical = pow(10, 25);
    this->alarm.external = false;
    this->alarm.enabled = false;
    this->measure.value = 0.0;
    this->measure.power = 0;
    this->measure.precision = 0;
    this->measure.infinity = false;
    this->measure.nan = false;
    this->label = "";
    this->hold = false;
    this->range.max = pow(10, 25);
    this->range.min = -pow(10, 25);
    this->range.check = false;
    this->range.exceeded = 0;
    this->extreme.enable = false;
    this->extreme.external = false;
    this->appendLabel = true;
    this->appendNewLine = true;
    this->warningColorName = "#f6f278";
    this->criticalColorName = "#f6b0b0";
    this->extend.enable = false;
    this->extend.precision = 2;
    this->accuracy.unit = 0;
    this->accuracy.percent = 0;
    this->accuracy.visible = false;
    this->offset.enabled = false;
    ui->labelRangeOver->setVisible(false);
    ui->labelRangeIn->setVisible(true);
    ui->labelRangeBelow->setVisible(false);
    this->bargraphVisible = false;
    ui->bargraph->setVisible(false);
    ui->bargraph->setMajorTickRising(true);
    ui->bargraph->setRange(this->range.min, this->range.max);
    this->buttonsVisible = false;
    this->buttonsEnabled = this->buttonsVisible;
    this->contexMenuEnable = false;
#ifdef Q_OS_WIN
    this->valueFont = QFont("Consolas", 11, QFont::Normal);
#elif defined(Q_OS_MAC)
    this->valueFont = QFont("Consolas", 11, QFont::Normal);
#else
    this->valueFont = QFont("Cantarell", 11, QFont::Normal);
#endif
    setValueFont(this->valueFont.family(), this->valueFont.pointSize());
    ui->multimeterDispalyButtons->setVisible(false);
    ui->multimeterDispalyButtons->connectButtons(this, SLOT(toolButtonHoldChanged(bool)), SLOT(toolButtonClearChanged()),
                                                 SLOT(toolButtonExtendChanged(bool)), SLOT(toolButtonOffsetChanged(bool)));
    policy.setHorizontalPolicy(QSizePolicy::Preferred);
    policy.setVerticalPolicy(QSizePolicy::Fixed);
    policy.setRetainSizeWhenHidden(true);
    ui->placeHolder->setSizePolicy(policy);
    ui->placeHolder->setVisible(false);
    setPolicyRetain();
    ui->labelValueMin->setVisible(false);
    ui->labelValueMax->setVisible(false);
    ui->labelOffsetBase->setVisible(false);

    connect(ui->labelOffsetBase, SIGNAL(clicked()), this, SLOT(editOffsetBase()));
    connect(lineEditBase, SIGNAL(editingFinished()), this, SLOT(onLineEditBaseEditingFinished()));
    connect(lineEditBase, SIGNAL(returnPressed()), this, SLOT(onLineEditBaseReturnPressed()));

    refreshBackground();
    addContexMenu();
    scaleLayout();
    setStretch();
    enableExtreme(this->extreme.enable);
    setUnit(this->unit);
    setValue(this->measure.value);
}

MultimeterDisplay::~MultimeterDisplay()
{
    delete ui;
}

void MultimeterDisplay::connectOptions(MultimeterDisplayOptions *options, const QObject *object, const char *slot, const char *signal)
{
    configuration = options;
    connect(this, SIGNAL(configurationRequest(QString)), object, slot);
    connect(object, signal, this, SLOT(applyConfiguration()));
}

void MultimeterDisplay::preparePowers(void)
{
    this->mapPower.insert(24, "Y");
    this->mapPower.insert(21, "Z");
    this->mapPower.insert(18, "E");
    this->mapPower.insert(15, "P");
    this->mapPower.insert(12, "T");
    this->mapPower.insert(9, "G");
    this->mapPower.insert(6, "M");
    this->mapPower.insert(3, "k");
    this->mapPower.insert(2, "h");
    this->mapPower.insert(1, "da");
    this->mapPower.insert(0, "");
    this->mapPower.insert(-1, "d");
    this->mapPower.insert(-2, "c");
    this->mapPower.insert(-3, "m");
    this->mapPower.insert(-6, QString::fromStdWString(L"\u00b5"));
    this->mapPower.insert(-9, "n");
    this->mapPower.insert(-12, "p");
    this->mapPower.insert(-15, "f");
    this->mapPower.insert(-18, "a");
    this->mapPower.insert(-21, "z");
    this->mapPower.insert(-24, "y");
}

void MultimeterDisplay::applyPolicy(QSizePolicy policy)
{
    ui->labelValueMin->setSizePolicy(policy);
    ui->labelValueMax->setSizePolicy(policy);
    ui->labelOffsetBase->setSizePolicy(policy);
}

void MultimeterDisplay::setPolicyRetain(void)
{
    if ((this->buttonsEnabled) || (this->extreme.enable)) policy.setRetainSizeWhenHidden(true);
    else policy.setRetainSizeWhenHidden(false);
    applyPolicy(policy);
}

void MultimeterDisplay::toolButtonHoldChanged(bool checked)
{
    setHold(checked);
}

void MultimeterDisplay::toolButtonClearChanged(void)
{
    clearExtreme();
}

void MultimeterDisplay::toolButtonExtendChanged(bool checked)
{
    enableExtendedPrecision(checked);
}

void MultimeterDisplay::toolButtonOffsetChanged(bool checked)
{
    enableOffset(checked, measure.value);
}

void MultimeterDisplay::setButtonsVisible(bool visible)
{
    ui->multimeterDispalyButtons->setVisible(visible);
    if (visible) enableButtons(true);
}

void MultimeterDisplay::enableButtons(bool enable)
{
    this->buttonsEnabled = enable;
    setPolicyRetain();
}

void MultimeterDisplay::enableContexMenu(bool enable)
{
    this->contexMenuEnable = enable;
    addContexMenu();
}

void MultimeterDisplay::addContexMenu(void)
{
    ui->output->setContextMenuPolicy(Qt::CustomContextMenu);
    if (this->contexMenuEnable) connect(ui->output, SIGNAL(customContextMenuRequested(const QPoint&)), this, SLOT(showContextMenu(const QPoint&)));
    else disconnect(ui->output, SIGNAL(customContextMenuRequested(const QPoint&)), this, SLOT(showContextMenu(const QPoint&)));
}

void MultimeterDisplay::showContextMenu(const QPoint& pos)
{
    QPoint globalPos = ui->output->mapToGlobal(pos);
    QMenu outputMenu;
    outputMenu.addAction(tr("Copy as text"), this, SLOT(copyAsText()));
    outputMenu.addAction(tr("Copy as image"), this, SLOT(copyAsImage()));
    outputMenu.addAction(tr("Copy as double"), this, SLOT(copyAsDouble()));
    if (configuration != nullptr) {
        outputMenu.addSeparator();
        actionSettings = new QAction(QIcon(":/icons/menu_settings.png"), tr("&Settings..."), this);
        actionSettings->setShortcuts(QKeySequence::Save);
        actionSettings->setIconText(tr("Settings..."));
        actionSettings->setShortcutContext(Qt::ApplicationShortcut);
        connect(actionSettings, SIGNAL(triggered()), this, SLOT(callSettings()));
        outputMenu.addAction(actionSettings);
    }
    outputMenu.exec(globalPos);
}

void MultimeterDisplay::loadRangePixmap(QSize size)
{
    int height;
    height = int(0.5 * size.height());
    QList<int> divider = {12, 14, 16, 20, 24, 28, 32, 40, 48, 56, 64, 72};
    int i;
    for (i = 0; i < divider.count(); ++i) {
        if ((divider[i]) >= height) break;
    }
    height = divider[i];
    pixelmapBelowRange.load(":/icons/mmd_rangebelow_" + QString::number(height) + ".png");
    pixelmapInRange.load(":/icons/mmd_rangein.png");
    pixelmapInRange = pixelmapInRange.scaled(height * 2, height);
    pixelmapOverRange.load(":/icons/mmd_rangeover_" + QString::number(height) + ".png");
}

void MultimeterDisplay::copyAsText(void)
{
    QClipboard *clipboard = QApplication::clipboard();
    QString text;
    if (this->appendLabel) text = ui->labelLabel->text() + " = ";
    text += ui->labelValue->text() + " " + ui->labelUnit->text();
    if (this->appendNewLine) text += "\n";
    clipboard->setText(text);
}

void MultimeterDisplay::copyAsImage(void)
{
    QClipboard *clipboard = QApplication::clipboard();
    QPixmap image;
    if (this->appendLabel) image = this->grab(QRect(QPoint(0, 0), QSize(this->width(), this->height())));
    else image = ui->output->grab(QRect(QPoint(0, 0), QSize(ui->output->width(), ui->output->height())));
    clipboard->setPixmap(image);
}

void MultimeterDisplay::copyAsDouble(void)
{
    QClipboard *clipboard = QApplication::clipboard();
    QString text;
    if (this->extend.enable) text = valueToString(this->measure.value, 0, this->measure.precision - extend.precision);
    else text = valueToString(this->measure.value, 0, this->measure.precision);
    if (this->appendNewLine) text += "\n";
    clipboard->setText(text);
}

void MultimeterDisplay::scaleLayout(void)
{
    QSize space = getLayoutSpace(this->valueFont);

    ui->layoutOutput->setSpacing(space.height() / 10);
    ui->layoutOutput->setStretch(0,10);
    ui->layoutOutput->setStretch(1,50);
    ui->layoutOutput->setStretch(2,10);
    ui->layoutOutput->setStretch(3,10);
    ui->layoutTop->setSpacing(space.width() / 4);
    ui->layoutCentral->setContentsMargins(space.width() / 2, 0, space.width() / 2, 0);
    ui->layoutCentral->setSpacing(space.width() / 2);
    loadRangePixmap(space);
    ui->labelRangeOver->setPixmap(pixelmapOverRange);
    ui->labelRangeIn->setPixmap(pixelmapInRange);
    ui->labelRangeBelow->setPixmap(pixelmapBelowRange);
    scaleLayoutSpacing();
    int stretchLabel = 0;
    if (this->label != "") stretchLabel = this->stretch.label;
    int stretchAccuracy = 0;
    if (this->accuracy.visible) stretchAccuracy = this->stretch.accuracy;
    refreshStretch(stretchLabel, this->stretch.value, this->stretch.unit, stretchAccuracy);
}

void MultimeterDisplay::scaleLayoutSpacing(void)
{
    if (this->label != "") ui->layout->setSpacing(getLayoutSpace(this->valueFont).width());
    else  ui->layout->setSpacing(0);
}

void MultimeterDisplay::setStretch(int label, int value, int unit, int accuracy)
{
    this->stretch.label = label;
    this->stretch.value = value;
    this->stretch.unit = unit;
    this->stretch.accuracy = accuracy;
    refreshStretch(this->stretch.label, this->stretch.value, this->stretch.unit, this->stretch.accuracy);
}

void MultimeterDisplay::refreshStretch(int label, int value, int unit, int accuracy)
{
    ui->layout->setStretch(0, label);
    ui->layout->setStretch(1, value + unit + accuracy);
    ui->layoutCentral->setStretch(0, int(0.1 * (value + unit)));
    ui->layoutCentral->setStretch(1, value);
    ui->layoutCentral->setStretch(2, unit);
    ui->layoutCentral->setStretch(3, accuracy);
}

void MultimeterDisplay::setBackgroundColor(QString color)
{
    if (!this->isEnabled()) color = this->backgroundColorName;
    background->setStyleSheet("background: " + color + ";");
    ui->bargraph->setBackgroundColor(QColor(color));
}

void MultimeterDisplay::refreshBackground(void)
{
    showAlarm(this->alarm.level);
}

void MultimeterDisplay::setColor(QString color)
{
    this->colorName = color;
    ui->bargraph->setBarColor(color);
    ui->labelValue->setStyleSheet(QString("QLabel:enabled {color: " + color + ";}"));
    ui->labelUnit->setStyleSheet(QString("QLabel:enabled {color: " + color + ";}"));
}

void MultimeterDisplay::applyConfiguration(void)
{
    if (configuration != nullptr) {
        setColor(configuration->foregroundColorName);
        this->backgroundColorName = configuration->backgroundColorName;
        this->warningColorName = configuration->warningColorName;
        this->criticalColorName = configuration->criticalColorName;
        refreshBackground();
        this->appendLabel =  configuration->appendLabel;
        this->appendNewLine = configuration->appendNewLine;
        setValueFont(configuration->valueFontFamily, this->valueFont.pointSize());
        setLabelFont(configuration->labelFontFamily, this->labelFont.pointSize());
        setExtendedPrecision(configuration->extendPrecision);
    }
}

void MultimeterDisplay::callSettings()
{
    if (configuration != nullptr) emit configurationRequest(configuration->getName());
}

void MultimeterDisplay::setHold(bool hold)
{
    this->hold = hold;
    ui->multimeterDispalyButtons->setButtonHoldChecked(this->hold);
}

void MultimeterDisplay::setAccuracy(double unit, double percent)
{
    if (unit < 0) unit = -unit;
    this->accuracy.unit = unit;
    if (percent < 0) percent = -percent;
    this->accuracy.percent = percent;
    showAccuracy();
}

void MultimeterDisplay::setAccuracy(double unit, QString percent)
{
    if (unit < 0) unit = -unit;
    this->accuracy.unit = unit;
    int index;
    index = percent.indexOf(QRegExp("ppm", Qt::CaseInsensitive));
    if (index > 0 ) {
        percent.replace(QRegExp("ppm", Qt::CaseInsensitive), "").trimmed();
        this->accuracy.percent = percent.toFloat() / pow(10, 4);
    }
    index = percent.indexOf(QRegExp("ppb", Qt::CaseInsensitive));
    if (index > 0 ) {
        percent.replace(QRegExp("ppb", Qt::CaseInsensitive), "").trimmed();
        this->accuracy.percent = percent.toFloat() / pow(10, 7);
    }
    index = percent.indexOf(QRegExp("%", Qt::CaseInsensitive));
    if (index > 0 ) {
        percent.replace(QRegExp("%", Qt::CaseInsensitive), "").trimmed();
        this->accuracy.percent = percent.toFloat();
    }
    if (this->accuracy.percent < 0) this->accuracy.percent = -this->accuracy.percent;
    showAccuracy();
}

void MultimeterDisplay::setAccuracyVisible(bool visible)
{
    this->accuracy.visible = visible;
    showAccuracy();
}

void MultimeterDisplay::setExtreme(double min, double max)
{
    this->extreme.min = min;
    this->extreme.max = max;
    this->extreme.external = true;
    showExtreme();
}

void MultimeterDisplay::setUnit(QString unit)
{
    this->unit = unit;
    showUnit();
}

void MultimeterDisplay::setLabel(QString label)
{
    this->label = QString(tr("%1")).arg(label);
    ui->labelLabel->setText(this->label);
    scaleLayout();
}

void MultimeterDisplay::setRange(double min, double max)
{
    if (min <= max) {
        this->range.min = min;
        this->range.max = max;
    } else {
        this->range.min = max;
        this->range.max = min;
    }
    ui->bargraph->setRange(this->range.min, this->range.max);
}

void MultimeterDisplay::setRangeExceeded(int state)
{
    if (state < 0) state = RANGE_STATE_BelowRange;
    else if (state > 0) state = RANGE_STATE_OverRange;
    if (this->range.exceeded != state) {
        this->range.exceeded = state;
        showRangeExceeded(this->range.exceeded);
    }
}

void MultimeterDisplay::enableCheckRange(bool enable)
{
    this->range.check = enable;
}

void MultimeterDisplay::setPrecision(int precision)
{
    if (precision < -PRECISION_MAX) precision = -PRECISION_MAX;
    if (precision > PRECISION_MAX) precision = PRECISION_MAX;
    this->measure.precision = precision;
}

void MultimeterDisplay::setUnitPower(int power)
{
    if (power < -POWER_MAX) power = -POWER_MAX;
    if (power > POWER_MAX) power = POWER_MAX;
    while (!mapPower.contains(power)) power++;
    this->measure.power = power;
    showUnit();
}

void MultimeterDisplay::setExtendedPrecision(int precision)
{
    if (configuration != nullptr) this->extend.precision = configuration->validateExtendPrecision(precision);
    if (this->extend.enable) showValue();
}

void MultimeterDisplay::setValueFont(QString family, int size)
{
    this->valueFont.setFamily(family);
    this->valueFont.setPointSize(size);
    ui->labelValue->setFont(this->valueFont);
    ui->labelUnit->setFont(this->valueFont);
    this->extraFont.setFamily(family);
    this->extraFont.setPointSize(int(0.4 * size));
    ui->labelValueMin->setFont(this->extraFont);
    ui->labelValueMax->setFont(this->extraFont);
    ui->labelAccuracy->setFont(this->extraFont);
    ui->labelOffsetBase->setFont(this->extraFont);
    lineEditBase->setFont(this->extraFont);
    scaleLayout();
}

void MultimeterDisplay::setLabelFont(QString family, int size)
{
    labelFont.setFamily(family);
    labelFont.setPointSize(size);
    ui->labelLabel->setFont(labelFont);
}

void MultimeterDisplay::setValueFontSize(int size)
{
    setValueFont(this->valueFont.family(), size);
}

void MultimeterDisplay::setValueFontFamily(QString family)
{
    setValueFont(family, this->valueFont.pointSize());
}

void MultimeterDisplay::setLabelFontSize(int size)
{
    setLabelFont(this->labelFont.family(), size);
}

void MultimeterDisplay::setAlarm(double bCritical, double bWarning, double tWarining, double tCritical)
{
    this->alarm.topCritical = tCritical;
    this->alarm.topWarning = tWarining;
    this->alarm.bottomWarning = bWarning;
    this->alarm.bottomCritical = bCritical;
}

void MultimeterDisplay::setAlarmLevel(int level)
{
    switch (level) {
        case ALARM_STATE_Disabled:
            this->alarm.enabled = false;
        case ALARM_STATE_OK:
        case ALARM_STATE_Warning:
        case ALARM_STATE_Critical:
            break;
        default:
            level = ALARM_STATE_Unknown;
    };
    this->alarm.external = true;
    if (this->alarm.level != level) {
        this->alarm.level = level;
        showAlarm(this->alarm.level);
    }
}

int MultimeterDisplay::readAlarm(void)
{
    return this->alarm.level;
}

void MultimeterDisplay::enableAlarm(bool enable)
{
    this->alarm.enabled = enable;
    if (!this->alarm.enabled) this->alarm.level = ALARM_STATE_Disabled;
    this->alarm.external = false;
    showAlarm(this->alarm.level);
}

void MultimeterDisplay::enableBargraph(bool enable)
{
    this->bargraphVisible = enable;
    setBackgroundColor(ui->output->palette().base().color().name());
    setColor(ui->output->palette().foreground().color().name());
    ui->bargraph->setVisible(this->bargraphVisible);
}

void MultimeterDisplay::enableExtreme(bool enable)
{
    this->extreme.enable = enable;
    this->extreme.external = false;
    if (this->extreme.enable) clearExtreme();
    setPolicyRetain();
    showExtreme();
}

void MultimeterDisplay::clearExtreme(void)
{
    this->extreme.external = false;
    this->extreme.min = this->measure.value;
    this->extreme.max = this->measure.value;
    showExtreme();
}

void MultimeterDisplay::enableExtendedPrecision(bool extended)
{
    this->extend.enable = extended;
    ui->multimeterDispalyButtons->setButtonExtendChecked(this->extend.enable);
    showValue();
}

void MultimeterDisplay::setValue(double value)
{
    if (!hold) {
        if (qIsFinite(value)) {
            if (this->extreme.enable) {
                if (!extreme.external) {
                    if (this->measure.value < this->extreme.min) this->extreme.min = this->measure.value;
                    if (this->measure.value > this->extreme.max) this->extreme.max = this->measure.value;
                }
            }
            this->measure.infinity = false;
            this->measure.nan = false;
            if (this->range.check) {
                if (value <= this->range.min) {
                    value = this->range.min;
                    if (this->range.exceeded != RANGE_STATE_BelowRange) {
                        this->range.exceeded = RANGE_STATE_BelowRange;
                        showRangeExceeded(this->range.exceeded);
                    }
                } else if (value >= this->range.max) {
                    value = this->range.max;
                    if (this->range.exceeded != RANGE_STATE_OverRange) {
                        this->range.exceeded = RANGE_STATE_OverRange;
                        showRangeExceeded(this->range.exceeded);
                    }
                } else
                    if (this->range.exceeded != RANGE_STATE_InRange) {
                        this->range.exceeded = RANGE_STATE_InRange;
                        showRangeExceeded(this->range.exceeded);
                    }
            }
            if (this->alarm.enabled) {
                if (!this->alarm.external) {
                    int oldAlarmLevel = this->alarm.level;
                    this->alarm.level = ALARM_STATE_OK;
                    if ((value < this->alarm.bottomWarning) || (value > this->alarm.topWarning)) this->alarm.level = ALARM_STATE_Warning;
                    if ((value < this->alarm.bottomCritical) || (value > this->alarm.topCritical)) this->alarm.level = ALARM_STATE_Critical;
                    if (oldAlarmLevel != this->alarm.level) showAlarm(this->alarm.level);
                }
            }
            this->measure.value = value;
        } else if (qIsInf(value)) {
            this->measure.infinity = true;
        } else this->measure.nan = true;
        showValue();
    }
}

void MultimeterDisplay::setOffsetBase(double base)
{
    this->offset.base = base;
    ui->labelOffsetBase->setText("base: " + valueToString(this->offset.base, this->measure.power, this->measure.precision) + " " + getCurrentUnit());
    showValue();
}

void MultimeterDisplay::enableOffset(bool enable, double base)
{
    this->offset.enabled = enable;
    if (this->offset.enabled) setOffsetBase(base);
    ui->labelOffsetBase->setVisible(this->offset.enabled);
    ui->multimeterDispalyButtons->setButtonOffsetChecked(this->offset.enabled);
    showValue();
}

void MultimeterDisplay::editOffsetBase(void)
{
    QSize space = getLayoutSpace(this->extraFont);

    lineEditBase->setText(valueToString(this->offset.base, this->measure.power, this->measure.precision));
    lineEditBase->setGeometry(ui->labelOffsetBase->geometry().left() + ui->output->geometry().left() - space.width() / 8,
                              ui->labelOffsetBase->geometry().top() + ui->output->geometry().top() - space.height() / 8,
                              ui->labelOffsetBase->geometry().width() + space.width() / 4, ui->labelOffsetBase->geometry().height() + space.height() / 4);
    lineEditBase->setVisible(true);
    lineEditBase->setFocus();
}

void MultimeterDisplay::onLineEditBaseEditingFinished(void)
{
    lineEditBase->setVisible(false);
}

void MultimeterDisplay::onLineEditBaseReturnPressed(void)
{
    double value = lineEditBase->text().replace(",", ".").toDouble() * pow(10, this->measure.power);
    if (value > this->range.max) value = this->range.max;
    else if (value < this->range.min) value = this->range.min;
    setOffsetBase(value);
    lineEditBase->setVisible(false);
}

void MultimeterDisplay::showUnit(void)
{
    ui->labelUnit->setText(getCurrentUnit());
}

QString MultimeterDisplay::getCurrentUnit(void)
{
    return mapPower.value(this->measure.power) + unit;
}

void MultimeterDisplay::showRangeExceeded(int state)
{
    ui->labelRangeBelow->setVisible(false);
    ui->labelRangeIn->setVisible(false);
    ui->labelRangeOver->setVisible(false);
    switch(state) {
    case RANGE_STATE_BelowRange:
        ui->labelRangeBelow->setVisible(true);
        break;
    case RANGE_STATE_OverRange:
        ui->labelRangeOver->setVisible(true);
        break;
    default:
        ui->labelRangeIn->setVisible(true);
    }
    if (this->range.check) emit this->rangeExceeded(state);
}

int MultimeterDisplay::checkPrecision(int precision)
{
    if (precision > 0) return precision;
    else return 0;
}

void MultimeterDisplay::showValue(void)
{
    if (this->measure.infinity) {
        ui->labelValue->setText(INFINITY_STRING);
        return;
    }
    if (this->measure.nan) {
        ui->labelValue->setText(NAN_STRING);
        return;
    }
    if (this->offset.enabled) {
        if (this->extend.enable) ui->labelValue->setText(valueToString(this->measure.value - this->offset.base, this->measure.power,
                                                                 this->measure.precision - extend.precision));
        else ui->labelValue->setText(valueToString(this->measure.value - this->offset.base, this->measure.power, this->measure.precision));
    } else {
        if (this->extend.enable) ui->labelValue->setText(valueToString(this->measure.value, this->measure.power, this->measure.precision - extend.precision));
        else ui->labelValue->setText(valueToString(this->measure.value, this->measure.power, this->measure.precision));
    }
    showExtreme();
    showAccuracy();
    if (this->bargraphVisible) ui->bargraph->setValue(this->measure.value);
}

void MultimeterDisplay::setInfinity(void)
{
    this->measure.infinity = true;
    showValue();
}

void MultimeterDisplay::setNaN(void)
{
    this->measure.nan = true;
    showValue();
}

void MultimeterDisplay::showAccuracy(void)
{
    if (this->accuracy.visible) {
        double a = accuracy.percent * 0.01 * measure.value;
        if (a < 0) a = -a;
        a += this->accuracy.unit;
        ui->labelAccuracy->setText("±" + valueToString(a, this->measure.power, this->measure.precision) + " " + getCurrentUnit());
        ui->labelAccuracy->setVisible(true);
    } else {
         ui->labelAccuracy->setVisible(false);
    }
}

void MultimeterDisplay::showExtreme(void)
{
    ui->labelValueMin->setVisible(this->extreme.enable);
    ui->labelValueMax->setVisible(this->extreme.enable);
    double min = this->extreme.min;
    double max = this->extreme.max;
    if (this->offset.enabled) {
        min -= this->offset.base;
        max -= this->offset.base;
    }
    if (this->extreme.enable) {
        ui->labelValueMin->setText(tr("min: ") + valueToString(min, this->measure.power, this->measure.precision) + " " + getCurrentUnit());
        ui->labelValueMax->setText(tr("max: ") + valueToString(max, this->measure.power, this->measure.precision) + " " + getCurrentUnit());
    }
}

QString MultimeterDisplay::valueToString(double value, int power, int precision)
{
    QLocale c;   // national / international
    c.setNumberOptions(QLocale::OmitGroupSeparator);
    precision = checkPrecision(power - precision);
    return c.toString(value / (pow(10, power)), 'f', precision);
}

void MultimeterDisplay::showAlarm(int state)
{
    if (this->alarm.enabled) {
    switch (state) {
    case ALARM_STATE_Warning:
        setBackgroundColor(this->warningColorName);
        return;
    case ALARM_STATE_Critical:
        setBackgroundColor(this->criticalColorName);
        return;
    default:
        // alarm disbled
        // state normal
        // state unknown
        setBackgroundColor(this->backgroundColorName);
    };
    } else {
        setBackgroundColor(this->backgroundColorName);
    }
}

QSize MultimeterDisplay::getLayoutSpace(QFont font)
{
    QFontMetrics fm(font);
    QSize space;
    space.setWidth(fm.width('0'));
    space.setHeight(fm.height());
    return space;
}

void MultimeterDisplay::paintEvent(QPaintEvent *)
{
    QStyleOption opt;
    opt.init(this);
    QPainter p(this);
    style()->drawPrimitive(QStyle::PE_Widget, &opt, &p, this);

    if (this->isEnabled()) {
        if (this->isBackgroundColor) {
            showAlarm(this->alarm.level);
            this->isBackgroundColor = false;
        }
    } else {
        if (!this->isBackgroundColor) {
            showAlarm(this->alarm.level);
            this->isBackgroundColor = true;
        }
    }
    if (configuration == nullptr) this->backgroundColorName = ui->output->palette().base().color().name();
}

void MultimeterDisplay::resizeEvent(QResizeEvent *)
{
    background->setGeometry(ui->output->geometry());
}


