#include "libusbdevice.h"
#include "libusbdeviceinfo.h"

LibUsbDevice::LibUsbDevice(QObject* parent) :
    QObject(parent)
{
    usbConnectionLost = false;
    enableEventThread = true;
    isDeviceConnected = false;
    isInitialiazed = false;
    deviceHandle = NULL;
    context = NULL;
    pcToUsbDeviceTransfer = NULL;
    usbDeviceToPcTransfer = NULL;
    dataAvailable = false;
    hasHotPlugSupport = false;
    deviceFound = NULL;
    connect(&serial,SIGNAL(newData(int)),this,SLOT(newDataAvailable(int)));
    connect(&serial,SIGNAL(clear()),this,SLOT(clearData()));
}
void LibUsbDevice::reset(){
    usbConnectionLost = false;
    enableEventThread = true;
    isDeviceConnected = false;
    isInitialiazed = false;
    deviceHandle = NULL;
    context = NULL;
    pcToUsbDeviceTransfer = NULL;
    usbDeviceToPcTransfer = NULL;
    dataAvailable = false;
    hasHotPlugSupport = false;
    deviceFound = NULL;
}

void LibUsbDevice::clearData(){
    for(int i = 0; i < LEN_BULK_IN_BUFFER; i++)
        chData[i] = 0;
}

extern "C" int LIBUSB_CALL hotplugAttachCallback (libusb_context*, libusb_device*, libusb_hotplug_event, void* ref)
{
    LibUsbDevice* cthis = static_cast<LibUsbDevice*>(ref);
    cthis->openDevice();
    qDebug()<<"Hotplug attach";
    if(cthis->isDeviceConnected)
        cthis->asyncBulkReadTransfer();
    return 0;
}

extern "C" int LIBUSB_CALL hotplugDetachCallback (libusb_context*, libusb_device*, libusb_hotplug_event, void* ref)
{
    qDebug()<<"Hotplug detach";
    LibUsbDevice* cthis = static_cast<LibUsbDevice*>(ref);
    if(cthis->isDeviceConnected)
    {
        if(cthis->usbDeviceToPcTransfer)
        {
            libusb_cancel_transfer(cthis->usbDeviceToPcTransfer);
        }
        if(cthis->pcToUsbDeviceTransfer)
        {
            libusb_cancel_transfer(cthis->pcToUsbDeviceTransfer);
        }
        libusb_release_interface(cthis->deviceHandle, INTERFACE_NUMBER);
        libusb_close(cthis->deviceHandle);
        cthis->isDeviceConnected = false;
    }
  return 0;
}

extern "C" void LIBUSB_CALL asyncBulkReadTransferCallback(struct libusb_transfer* transfer)
{
    LibUsbDevice* cthis = static_cast<LibUsbDevice*>(transfer->user_data);
    cthis->dataAvailable = true;
    if(cthis->enableEventThread)
    {
        cthis->dataLength = transfer->actual_length;
        libusb_submit_transfer(transfer);
    }

}

extern "C" void LIBUSB_CALL asyncBulkWriteTransferCallback(struct libusb_transfer*)
{

}

void LibUsbDevice::initializeDevice()
{    
    int status;
    status = libusb_init(&context); //initialize the library for the session we just declared
    if(status != 0)
    {
        cstatus = tr("Initialization Error!");
        qDebug()<<"Initialization Error! "<<status; //there was an error
        isInitialiazed = false;
        return;
    }
    isInitialiazed = true;
    if(libusb_has_capability (LIBUSB_CAP_HAS_HOTPLUG)&&0)
    {
        hasHotPlugSupport = true;
        //status = libusb_hotplug_register_callback (context, LIBUSB_HOTPLUG_EVENT_DEVICE_ARRIVED, LIBUSB_HOTPLUG_ENUMERATE , VENDOR_ID,
        //  PRODUCT_ID, CLASS_ID, hotplugAttachCallback, this, &hotplugHandle[0]);
        if (LIBUSB_SUCCESS != status)
        {
            hasHotPlugSupport = false;
        }

        //status = libusb_hotplug_register_callback (context, LIBUSB_HOTPLUG_EVENT_DEVICE_LEFT, LIBUSB_HOTPLUG_ENUMERATE , VENDOR_ID,
        // PRODUCT_ID, CLASS_ID, hotplugDetachCallback, this, &hotplugHandle[1]);
        if (LIBUSB_SUCCESS != status)
        {
            hasHotPlugSupport = false;
        }
        future = QtConcurrent::run (this,&LibUsbDevice::eventThread);
    }
    libusb_set_debug(context, LIBUSB_LOG_LEVEL_INFO); //set verbosity level to 3
}

void LibUsbDevice::openDevice()
{
    size_t i = 0;
    int status;

    status = libusb_get_device_list(context, &devs);
    if (status < 0){
        cstatus = tr("Cannot open device!");
        qDebug()<<"No USB Devices to List";
        return;
    }
    while ((dev = devs[i++]) != NULL){
        status = libusb_get_device_descriptor(dev, &deviceDesc);
        if (status < 0) {
            break;
        }
        if (deviceDesc.idVendor == VENDOR_ID && deviceDesc.idProduct == PRODUCT_ID){
            status = libusb_open(dev, &deviceHandle);
            if (status) {
                cstatus = tr("Cannot open device!");
                qDebug()<<"Cannot open device--"<<i;
                deviceHandle = NULL;
            } else {
                qDebug()<<dev<<" Device Opened";
                isDeviceConnected = true;
                break;
            }
        }
    }
    libusb_free_device_list(devs, 1);
    if(deviceHandle == NULL) {
        cstatus = tr("Cannot open device!");
        qDebug()<<"Cannot open device";
        emit serial.connectionStatus("Cannot open device");
        return;
    }
    if(libusb_kernel_driver_active(deviceHandle, 0) == 1) {
        qDebug()<<"Kernel Driver Active";
        if(libusb_detach_kernel_driver(deviceHandle, 0) == 0) { //detach it
            qDebug()<<"Kernel Driver Detached!";
        }
    }
    status = libusb_claim_interface(deviceHandle, INTERFACE_NUMBER);
    if(status != LIBUSB_SUCCESS) {
        qDebug()<<"Cannot Claim Interface";
        libusb_close(deviceHandle);
        isDeviceConnected = false;
        return;
    }
}

void LibUsbDevice::closeDevice()
{
    if(isDeviceConnected)
    {
        qDebug()<<"Signal to disconnect";
        usbConnectionLost = true;
        qDebug()<<"Cancel USB to PC transfer";
        if(usbDeviceToPcTransfer)
        {
            libusb_cancel_transfer(usbDeviceToPcTransfer);
        }
        qDebug()<<"Cancel PC to USB transfer";
        if(pcToUsbDeviceTransfer)
        {
            libusb_cancel_transfer(pcToUsbDeviceTransfer);
        }
        enableEventThread = false;
        if(hasHotPlugSupport)
        {
            // libusb_hotplug_deregister_callback(context, hotplugHandle[0]);
            // libusb_hotplug_deregister_callback(context, hotplugHandle[1]);
        }
        qDebug()<<"Wait for Finished";
        future.waitForFinished();
        qDebug()<<"Release interface";
        libusb_release_interface(deviceHandle, INTERFACE_NUMBER);
        qDebug()<<"Close";
        libusb_close(deviceHandle);
        qDebug()<<"Exit";
        libusb_exit(context);
        isDeviceConnected=false;
    }
    else if(isInitialiazed)
    {
        enableEventThread = false;
        if(hasHotPlugSupport)
        {
            // libusb_hotplug_deregister_callback(context, hotplugHandle[0]);
            // libusb_hotplug_deregister_callback(context, hotplugHandle[1]);
        }
        future.waitForFinished();
        libusb_exit(context);
    }
}

void LibUsbDevice::eventThread() {
    while (enableEventThread) {
        libusb_handle_events(context);
    }
    return;
}

int LibUsbDevice::controlReadTransfer(uint8_t command, unsigned char* buffer, int length, uint16_t index, uint16_t value) {
    int bytesRead;

    if(!checkIsDeviceConnected()) return -99;
    qDebug()<<"USB Transfer: cmd="<<QString::number(command)<<" index="<<QString::number(index)<<" value="<<QString::number(value);
    bytesRead = libusb_control_transfer(deviceHandle,0xC0,command,value,index,buffer,length,1000);
    if(bytesRead<0) libUsbTransferResultError(bytesRead);
    else if(length>0) qDebug()<<"Return "<<QString::number(bytesRead)<<" bytes";
    return bytesRead;
}

void LibUsbDevice::controlWriteTransfer(uint8_t command, uint16_t index, uint16_t value)
{
    controlReadTransfer(command,NULL,0,index,value);
}

void LibUsbDevice::asyncBulkReadTransfer()
{
    if (!checkIsDeviceConnected()) return;
    int status;
    usbDeviceToPcTransfer = libusb_alloc_transfer(0);
    libusb_fill_bulk_transfer( usbDeviceToPcTransfer, deviceHandle, USB_ENDPOINT_IN,
                               chData, LEN_BULK_IN_BUFFER, asyncBulkReadTransferCallback, this, 1000); // <----------1000
    status = libusb_submit_transfer(usbDeviceToPcTransfer);
    if (status != 0) {
        cstatus = tr("Bulk read transfer failed!");
        qDebug()<<"Transfer Failed! "<<status;
        libUsbTransferResultError(status);
        return;
    }
    if (!future.isRunning()) {
        future = QtConcurrent::run (this,&LibUsbDevice::eventThread);
    }
}

void LibUsbDevice::bulkWriteTransfer(unsigned char* buffer, int length)
{
    if(!checkIsDeviceConnected()) return;
    int status;
    int actual;
    status = libusb_bulk_transfer(deviceHandle, USB_ENDPOINT_OUT, buffer, length, &actual, 1000);
    if(status < 0) {
        cstatus = tr("Bulk write transfer failed!");
        qDebug()<<"Bulk write transfer failed! "<<status;
        libUsbTransferResultError(status);
        return;
    }
}

void LibUsbDevice::libUsbTransferResultError(int errCode) {
    usbConnectionLost = true;
    if (usbConnectionLost) emit connectionLost();
    qDebug()<<"Emit ConnectionLost";
    if ( errCode == LIBUSB_ERROR_IO) qDebug()<<"Transfer error: LIBUSB_ERROR_IO";
    else if (errCode == LIBUSB_ERROR_INVALID_PARAM) qDebug()<<"Transfer error: LIBUSB_ERROR_INVALID_PARAM";
    else if (errCode == LIBUSB_ERROR_ACCESS) qDebug()<<"Transfer error: LIBUSB_ERROR_ACCESS";
    else if (errCode == LIBUSB_ERROR_NO_DEVICE) qDebug()<<"Transfer error: LIBUSB_ERROR_NO_DEVICE";
    else if (errCode == LIBUSB_ERROR_NOT_FOUND) qDebug()<<"Transfer error: LIBUSB_ERROR_NOT_FOUND";
    else if (errCode == LIBUSB_ERROR_BUSY) qDebug()<<"Transfer error: LIBUSB_ERROR_BUSY";
    else if (errCode == LIBUSB_ERROR_TIMEOUT) qDebug()<<"Transfer error: LIBUSB_ERROR_TIMEOUT";
    else if (errCode == LIBUSB_ERROR_OVERFLOW) qDebug()<<"Transfer error: LIBUSB_ERROR_OVERFLOW";
    else if (errCode == LIBUSB_ERROR_PIPE) qDebug()<<"Transfer error: LIBUSB_ERROR_PIPE";
    else if (errCode == LIBUSB_ERROR_INTERRUPTED) qDebug()<<"Transfer error: LIBUSB_ERROR_INTERRUPTED";
    else if (errCode == LIBUSB_ERROR_NO_MEM) qDebug()<<"Transfer error: LIBUSB_ERROR_NO_MEM";
    else if (errCode == LIBUSB_ERROR_NOT_SUPPORTED) qDebug()<<"Transfer error: LIBUSB_ERROR_NOT_SUPPORTED";
    else if (errCode == LIBUSB_ERROR_OTHER) qDebug()<<"Transfer error: LIBUSB_ERROR_OTHER";
}

void LibUsbDevice::newDataAvailable(int size) {
    dataAvailable = true;
    dataLength = size;
}

bool LibUsbDevice::checkIsDeviceConnected() {
    bool result = true;

    if((!isDeviceConnected) || (usbConnectionLost)) {
        result = false;
        qDebug()<<"Device not connected";
    }
    return result;
}
