/*
 * Part of Xlab Interface
 *
 * Created: 2017-12-16
 * Qt Creator 4.4.1 Based on Qt 5.9.2
 *  Author: Krzysztof Markiewicz
 *  www.obbo.pl
 *
 * This program is distributed under the terms of the GNU General Public License
 */

#include <qglobal.h>
#ifdef Q_OS_MACOS
    #include <Python.h>
#endif
#include "pythonconsole.h"
#include "pythonconsoleoptions.h"
#include "PythonQt.h"
#include "PythonQt_QtAll.h"
#include "PythonQtScriptingConsole.h"
#include "pythonxlabobject.h"
#include "pythonxlabobjectawg.h"
#include "xlabinterface.h"
#include <QKeySequence>
#include <QFontMetrics>



PythonConsole::PythonConsole(QWidget *parent, PythonConsoleOptions *options, bool *isRunning)
    : QMainWindow(parent), pyConsole(nullptr), pyObject(nullptr), pyObjectAWG(nullptr)
{
    stateRunning = isRunning;
    *stateRunning = true;
    configuration = options;
    setWindowTitle(QCoreApplication::applicationName() + " - " + tr("Python Console"));
    setWindowFlags(this->windowFlags() & ~Qt::WindowMinMaxButtonsHint );
    resizeTimer = new QTimer(this);
    connect(resizeTimer, SIGNAL(timeout()), this, SLOT(resizeTimeout()));
    lineEditWindowSize = new QLineEdit(this);
    lineEditWindowSize->setVisible(false);
    lineEditWindowSize->setFont(QFont("Courier", 11));
    lineEditWindowSize->setAlignment(Qt::AlignCenter);
    lineEditWindowSize->setReadOnly(true);
    createMenu();
    initPythonConsole();

    if ((!configuration->restoreSize) || (configuration->sizeWidth == 0) || (configuration->sizeHeight == 0)) {
        configuration->sizeWidth = configuration->sizeWidthDefault;
        configuration->sizeHeight = configuration->sizeHeightDefault;
    }
    lineEditWindowSize->raise();
    applyConfiguration();
    connect(parent, SIGNAL(applyConfigurationRequest()), this, SLOT(applyConfiguration()));
}

PythonConsole::~PythonConsole()
{
    qDebug() << "Destruct Python console";
    deletePythonConsole();
    *stateRunning = false;
}

void PythonConsole::initPythonConsole()
{
    qDebug() << "Init PythonQt";
    PythonQt::init(PythonQt::IgnoreSiteModule | PythonQt::RedirectStdOut);
    PythonQt_QtAll::init();
    PythonQtObjectPtr  mainContext = PythonQt::self()->getMainModule();
    if (pyConsole == nullptr) {
        pyConsole = new PythonQtScriptingConsole(this, mainContext);
        pyConsole->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOn);
        setCentralWidget(pyConsole);
    }
    if (pyObject == nullptr) {
        pyObject = new PythonXlabObject;
        pyObject->setXlabQt(static_cast<XlabInterface*>(parentWidget()));
        mainContext.addObject("xlab", pyObject);
    }
    if (pyObjectAWG == nullptr) {
        pyObjectAWG = new PythonXlabObjectAWG;
        pyObjectAWG->setXlabQt(static_cast<XlabInterface*>(parentWidget()));
        mainContext.addObject("xlabAWG", pyObjectAWG);
    }
}

void PythonConsole::applyConfiguration()
{
    setFont(configuration->fontFamily, configuration->fontSize);
    setMinimumWidth(colsToPixel(configuration->sizeWidthMinimum));
    if (configuration->sizeWidthMaximum != 0) {
        setMaximumWidth(colsToPixel(configuration->sizeWidthMaximum));
    }
    setMinimumHeight(rowsToPixel(configuration->sizeHeightMinimum));
    if (configuration->sizeHeightMaximum != 0) {
        setMaximumHeight(colsToPixel(configuration->sizeHeightMaximum));
    }
    pyConsole->setFont(currentFont);
    pyConsole->setStyleSheet("PythonQtScriptingConsole {border: none; background-color: "
                             + configuration->backgroundColorName
                             + "; color: " + configuration->colorName + "; }");
    configuration->sizeWidth = configuration->validateSizeWidth(configuration->sizeWidth);
    configuration->sizeHeight = configuration->validateSizeHeight(configuration->sizeHeight);
    this->resize(colsToPixel(configuration->sizeWidth), rowsToPixel(configuration->sizeHeight));
    this->updateGeometry();
}

void PythonConsole::setFont(QString family, int size)
{
    currentFont.setFamily(family);
    currentFont.setPointSize(size);
}

void PythonConsole::deletePythonConsole()
{
    delete pyObject;
    pyObject = nullptr;
    delete pyObjectAWG;
    pyObjectAWG = nullptr;
    delete pyConsole;
    pyConsole = nullptr;
}

int PythonConsole::colsToPixel(int cols)
{
    int sb = pyConsole->style()->pixelMetric(QStyle::PM_ScrollBarExtent);
    int dm = int(pyConsole->document()->documentMargin());
    QFontMetrics fm(currentFont);
    return cols * fm.width('x') + sb + 2 * dm;
}

int PythonConsole::pixelToCols(int width)
{
    int sb = pyConsole->style()->pixelMetric(QStyle::PM_ScrollBarExtent);
    int dm = int(pyConsole->document()->documentMargin());
    QFontMetrics fm(currentFont);
    return int(trunc((width - sb - 2 * dm) / fm.width('x')));
}

int PythonConsole::rowsToPixel(int rows)
{
    int mb = menuBar()->height();
    int dm = int(pyConsole->document()->documentMargin());
    QFontMetrics fm(currentFont);
    return rows * fm.height() + mb + 2 * dm;
}

int PythonConsole::pixelToRows(int height)
{
    int mb = menuBar()->height();
    int dm = int(pyConsole->document()->documentMargin());
    QFontMetrics fm(currentFont);
    return int(trunc((height - mb - 2 * dm) / fm.height()));
}

void PythonConsole::createMenu()
{
    QList<QKeySequence> shortcuts;

    mainMenu = menuBar()->addMenu(tr("&Action"));

    actionOpenScript = new QAction(QIcon(":/icons/menu_openfile.png"), tr("&Open..."), this);
    actionOpenScript->setIconText(tr("Open"));
    shortcuts.clear();
    shortcuts.push_back(QKeySequence(tr("Ctrl+O")));
    actionOpenScript->setShortcuts(shortcuts);
    actionOpenScript->setShortcutContext(Qt::ApplicationShortcut);
    connect(actionOpenScript, SIGNAL(triggered()), this, SLOT(openScript()));
    mainMenu->addAction(actionOpenScript);

    actionClear = new QAction(QIcon(":/icons/menu_clear.png"), tr("Clear"), this);
    actionClear->setIconText(tr("Clear"));
    actionClear->setShortcutContext(Qt::ApplicationShortcut);
    connect(actionClear, SIGNAL(triggered()), this, SLOT(clearConsole()));
    mainMenu->addAction(actionClear);

    actionHide = new QAction(QIcon(".png"), tr("&Hide"), this);
    actionHide->setIconText(tr("Hide"));
    shortcuts.clear();
    shortcuts.push_back(QKeySequence(tr("Ctrl+H")));
    actionHide->setShortcuts(shortcuts);
    actionHide->setShortcutContext(Qt::ApplicationShortcut);
    connect(actionHide, SIGNAL(triggered()), this, SLOT(hideConsole()));
    mainMenu->addAction(actionHide);

    actionSettings = new QAction(QIcon(":/icons/menu_settings.png"), tr("&Settings..."), this);
    actionSettings->setShortcuts(QKeySequence::Save);
    actionSettings->setIconText(tr("Settings..."));
    actionSettings->setShortcutContext(Qt::ApplicationShortcut);
    connect(actionSettings, SIGNAL(triggered()), this, SLOT(callSettings()));
    mainMenu->addAction(actionSettings);

    mainMenu->addSeparator();

    actionClose = new QAction(QIcon(":/icons/menu_exit.png"), tr("&Close"), this);
    actionClose->setIconText(tr("Close"));
    shortcuts.clear();
    shortcuts.push_back(QKeySequence(tr("Alt+X")));
    actionClose->setShortcuts(shortcuts);
    actionClose->setShortcutContext(Qt::ApplicationShortcut);
    connect(actionClose, SIGNAL(triggered()), this, SLOT(closeConsole()));
    mainMenu->addAction(actionClose);
}

void PythonConsole::callSettings()
{
    emit configurationRequest(configuration->getName());
}

void PythonConsole::clearConsole()
{
    pyConsole->clear();
}

void PythonConsole::hideConsole()
{
    this->setVisible(false);
}

void PythonConsole::closeConsole()
{
    this->close();
}

void PythonConsole::openScript()
{
    QString fileName = QFileDialog::getOpenFileName(this,
        tr("Open Python Script"), "",
        tr("Python Script (*.py);;All Files (*)"));
    if (!fileName.isEmpty()) runScript(fileName);
}

void PythonConsole::evaluateScript(QString evalCode, bool notify)
{
	PythonQtObjectPtr  mainContext = PythonQt::self()->getMainModule();
	mainContext.evalScript(evalCode.trimmed() + "\n");
	if (notify) {
        printConsole(tr("Evaluated ") + QString::number(countLines(evalCode)) + tr(" line(s)."));
        pyConsole->appendCommandPrompt();
	}
}

void PythonConsole::runScript(QString fileName, bool notify)
{
    QDir dir = QDir::currentPath();
	QDir newDir = QDir(fileName.mid(0, fileName.lastIndexOf(QDir::separator())));
    QDir::setCurrent(newDir.absolutePath());
	QFile file(fileName);
	if (!file.open(QIODevice::ReadOnly)) {
        if (notify) printConsole(tr("Unable to open file: ") + fileName);
        qWarning() << "Unable to open file: " << newDir.absolutePath();
        return;
    }
    QFileInfo f(fileName);
    if (notify) printConsole(tr("Opening file: ") + f.fileName() + " ...");
    qDebug() << "Opening file: " << newDir.absolutePath();
    QString evalCode = QString(file.readAll());
    evaluateScript(evalCode, notify);
	QDir::setCurrent (dir.absolutePath());
}

int PythonConsole::countLines(QString evalCode)
{
    int lines = evalCode.count("\n");
    if (!evalCode.endsWith("\n")) {
        lines++;
    }
    return lines;
}

void PythonConsole::printConsole(QString string)
{
    PythonQtObjectPtr  mainContext = PythonQt::self()->getMainModule();
    mainContext.evalScript("print '" + string + "'");
}

void PythonConsole::resizeTimeout()
{
   resizeTimer->stop();
   lineEditWindowSize->setVisible(false);
}

void PythonConsole::resizeEvent(QResizeEvent *event)
{
    resizeTimer->stop();
    if ((event->oldSize().width() > 0) & (event->oldSize().height() > 0)) {    
        int w = pixelToCols(event->size().width());
        int h = pixelToRows(event->size().height());
        lineEditWindowSize->setText("txt: " + QString::number(w) + " x " + QString::number(h));
        configuration->sizeWidth = w;
        configuration->sizeHeight = h;

        int wpx, hpx;
        QFontMetrics fm(lineEditWindowSize->font());
        wpx = fm.width(lineEditWindowSize->text()) + 8;
        hpx = fm.height() + 4;
        lineEditWindowSize->setGeometry((this->width() - wpx) / 2, hpx / 2 + (this->height() - hpx) / 2, wpx, hpx);
        lineEditWindowSize->setVisible(true);
        resizeTimer->start(RESIZE_TIMEOUT);
    }
}
