#include "xlabinterface.h"
#include "ui_xlabinterface.h"
#include <stdio.h>


XlabInterface::XlabInterface(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::XlabInterface)
{
    QFontDatabase::addApplicationFont(":/fonts/cour.ttf");
    QFontDatabase::addApplicationFont(":/fonts/SourceCodeVariable-Italic.ttf");
    QFontDatabase::addApplicationFont(":/fonts/SourceCodeVariable-Roman.ttf");

#ifdef Q_OS_WIN
    // qApp->setFont(QFont("Courier New", 9, QFont::Normal));
#elif defined(Q_OS_MAC)
    // qApp->setFont(QFont("Courier New", 10, QFont::Normal));
#else
    qApp->setFont(QFont("Cantarell", 10, QFont::Normal));
#endif

    // build main configuration
    map_Option = new QMap<QString, QVariant*>;

    tomap_GlobalOptions = new GlobalOptions;
    v_GlobalOptions = new QVariant;
    v_GlobalOptions->setValue(tomap_GlobalOptions);
    map_Option->insert("GlobalOptions", v_GlobalOptions);

    tomap_PythonConsoleOptions = new PythonConsoleOptions;
    v_PythonConsoleOptions = new QVariant;
    v_PythonConsoleOptions->setValue(tomap_PythonConsoleOptions);
    map_Option->insert("PythonConsoleOptions", v_PythonConsoleOptions);

    tomap_ScopeOptions = new ScopeOptions;
    v_ScopeOptions = new QVariant;
    v_ScopeOptions->setValue(tomap_ScopeOptions);
    map_Option->insert("ScopeOptions", v_ScopeOptions);

    tomap_MultimeterDisplayOptions = new MultimeterDisplayOptions;
    v_MultimeterDisplayOptions = new QVariant;
    v_MultimeterDisplayOptions->setValue(tomap_MultimeterDisplayOptions);
    map_Option->insert("MultimeterDisplayOptions", v_MultimeterDisplayOptions);

    const float DEFAULT_DPI = 96.0;
    tomap_GlobalOptions->scaleFactorX = logicalDpiX() / DEFAULT_DPI;
    tomap_GlobalOptions->scaleFactorY = logicalDpiY() / DEFAULT_DPI;

    /*
     * QDir::addSearchPath()
     * resource with : instead of :/ as the prefix, the resource will be looked up using the search path list
     *
     * QDir::setSearchPaths("docs", QStringList(":/embeddedDocuments"));
     * QFile file("docs:design.odf"); // will look in the :/embeddedDocuments resource path
    */

    ui->setupUi(this);

    logging = false;
    loggingCounter = 0;
    logsToSave = "";
    lastFrameScope = 0;

    bufferSize = 768;
    qApp->processEvents();
    //Sleep(100);

    recordingWaves = false;
    recordWaveFile = new QFile();
    loadRecordedWave = false;

    m_repaint = false;
    m_prevTabIndex = 0;
    ui->mainTabWidget->setCurrentIndex(MAIN_TAB_Osciloscope);

    QFile fileTheme;
#ifdef Q_OS_WIN
    fileTheme.setFileName(":/themes/theme-win.qss");
#elif defined(Q_OS_MAC)
    fileTheme.setFileName(":/themes/theme-mac.qss");
#else
    fileTheme.setFileName(":/themes/theme-linux.qss");
#endif
    fileTheme.open(QFile::ReadOnly);
    QString styleSheet = QLatin1String(fileTheme.readAll());
    this->setStyleSheet(styleSheet);
    fileTheme.close();

    //ui->labelHertz->setStyleSheet("QLabel{font-size: 32px};");
    ui->connectLabel->setStyleSheet("QLabel{font-size: 12px};");

    changeWindowTitle("Interface - Not Connected");
    setupAppSetting();

    xAxisMax = 2 * bufferSize;
    xAxisMin = 256;
    yAxisMax = 1280;
    yAxisMin = 512;
    gridSize = 64;        // 1dz = 64
    ui->ch1PositionSlider->setMaximum(yAxisMax / 2 - rangeMax / 4);
    ui->ch1PositionSlider->setMinimum(-(yAxisMax / 2 - rangeMax / 4));
    ui->ch2PositionSlider->setMaximum(yAxisMax / 2 - rangeMax / 4);
    ui->ch2PositionSlider->setMinimum(-(yAxisMax / 2 - rangeMax / 4));
    ui->chdPositionSlider->setMaximum(yAxisMax / 2 - rangeMax / 4 - 120);
    ui->chdPositionSlider->setMinimum(-(yAxisMax / 2 - rangeMax / 4 + 120));

    ui->ch1CaptureSlider->setMaximum(yAxisMax / 2 - rangeMax / 4);
    ui->ch1CaptureSlider->setMinimum(-(yAxisMax / 2 - rangeMax / 4));
    ui->ch2CaptureSlider->setMaximum(yAxisMax / 2 - rangeMax / 4);
    ui->ch2CaptureSlider->setMinimum(-(yAxisMax / 2 - rangeMax / 4));
    ui->chdCaptureSlider->setMaximum(yAxisMax / 2 - rangeMax / 4 - 120);
    ui->chdCaptureSlider->setMinimum(-(yAxisMax / 2 - rangeMax / 4 + 120));

    triggerSetRange(4, (bufferSize - 8), 127, TRIGGRT_WINDOW_VALUE_MAX, TRIGGRT_WINDOW_DISTANCE_MIN);


    xmax = 256;
    hCursorAPos = 200;
    hCursorBPos = 100;
    vCursorInit();
    vCursorSetRange(8, xAxisMax - 8);
    triggerInit();
    hCursorInit();
    hCursorAPosCh1 = 100;
    hCursorBPosCh1 = -100;
    hCursorAPosCh2 = 100;
    hCursorBPosCh2 = -100;

    awg_wave_update_disable = false;

    itemIsSelected = false;
    captureRef = false;
    saveWave = false;
    displayLoadedWave = false;
    mode = MAIN_MODE_Scope;

    clearBuffer(datacache, DATACACHE_SIZE);
    scopeInitDatacache();

    connect(&customThemeDialog, SIGNAL(applyCustomTheme(int,CustomColors*)), this, SLOT(setTheme(int, CustomColors*)));
    setupValues();
    setupGrid(ui->plotterWidget);

    connect(ui->plotterWidget, SIGNAL(mouseMove(QMouseEvent*)), this, SLOT(moveCursor(QMouseEvent*)));
    connect(ui->plotterWidget, SIGNAL(mousePress(QMouseEvent*)), this, SLOT(selectItem(QMouseEvent*)));
    connect(ui->plotterWidget, SIGNAL(mouseRelease(QMouseEvent*)), this, SLOT(deselectItem(QMouseEvent*)));
    connect(ui->plotterWidget, SIGNAL(itemDoubleClick(QCPAbstractItem*,QMouseEvent*)), this, SLOT(itemDoubleClick(QCPAbstractItem*,QMouseEvent*)));

    connect(ui->horizontalScrollBar, SIGNAL(valueChanged(int)), this, SLOT(on_horizontalScrollBar_valueChanged(int)));
    connect(ui->verticalScrollBar, SIGNAL(valueChanged(int)), this, SLOT(on_verticalScrollBar_valueChanged(int)));

    ui->plotterWidget->axisRect()->setRangeZoom(Qt::Horizontal);

    QFont m_mmTabFont1 = QFont("Courier New", 40, QFont::DemiBold);
    QFont m_mmTabFont2 = QFont("Courier New", 36, QFont::DemiBold);
    QFont m_mmTabFont3 = QFont("Courier New", 28, QFont::DemiBold);

#ifdef Q_OS_WIN
    int dpiX = QPaintDevice::logicalDpiX();
    int dpiY = QPaintDevice::logicalDpiY();

    if (dpiX == 144 && dpiY == 144) {
        ui->mainTabWidget->setMinimumWidth(600);
        ui->ch1Label->setMinimumHeight(40);
        ui->ch2Label->setMinimumHeight(40);
        ui->timeLabel->setMinimumHeight(40);
        ui->ch1ColorLabel->setMinimumWidth(20);
        ui->ch1ColorLabel->setMinimumHeight(20);
        ui->ch2ColorLabel->setMinimumWidth(20);
        ui->ch2ColorLabel->setMinimumHeight(20);
        m_mmTabFont1.setPixelSize(55);
        m_mmTabFont2.setPixelSize(50);
    } else if (dpiX == 192 && dpiY == 192) {
        ui->mainTabWidget->setMinimumWidth(700);
        ui->ch1Label->setMinimumHeight(50);
        ui->ch2Label->setMinimumHeight(50);
        ui->timeLabel->setMinimumHeight(50);
        ui->ch1ColorLabel->setMinimumWidth(25);
        ui->ch1ColorLabel->setMinimumHeight(25);
        ui->ch2ColorLabel->setMinimumWidth(25);
        ui->ch2ColorLabel->setMinimumHeight(25);
        m_mmTabFont1.setPixelSize(55);
        m_mmTabFont2.setPixelSize(50);
    }
#endif
    // Osciloscope
    ui->multimeterdisplayVppCh1->setValueFontSize(9);
    ui->multimeterdisplayVppCh1->setUnit("V");
    ui->multimeterdisplayVppCh1->setLabel("Vpp");
    ui->multimeterdisplayVppCh1->setPrecision(-2);

    ui->multimeterdisplayMaxCh1->setValueFontSize(9);
    ui->multimeterdisplayMaxCh1->setUnit("V");
    ui->multimeterdisplayMaxCh1->setLabel("Max");
    ui->multimeterdisplayMaxCh1->setPrecision(-2);

    ui->multimeterdisplayMinCh1->setValueFontSize(9);
    ui->multimeterdisplayMinCh1->setUnit("V");
    ui->multimeterdisplayMinCh1->setLabel("Min");
    ui->multimeterdisplayMinCh1->setPrecision(-2);

    ui->multimeterdisplayRMSCh1->setValueFontSize(9);
    ui->multimeterdisplayRMSCh1->setUnit("V");
    ui->multimeterdisplayRMSCh1->setLabel("RMS");
    ui->multimeterdisplayRMSCh1->setPrecision(-2);

    ui->multimeterdisplayAverCh1->setValueFontSize(9);
    ui->multimeterdisplayAverCh1->setUnit("V");
    ui->multimeterdisplayAverCh1->setLabel("Average");
    ui->multimeterdisplayAverCh1->setPrecision(-2);

    ui->multimeterdisplayFreqCh1->setValueFontSize(9);
    ui->multimeterdisplayFreqCh1->setUnit("Hz");
    ui->multimeterdisplayFreqCh1->setLabel("Frequency");
    ui->multimeterdisplayFreqCh1->setPrecision(-1);
    ui->multimeterdisplayFreqCh1->setUnitPower(3);

    ui->multimeterdisplayVppCh2->setValueFontSize(9);
    ui->multimeterdisplayVppCh2->setUnit("V");
    ui->multimeterdisplayVppCh2->setLabel("Vpp");
    ui->multimeterdisplayVppCh2->setPrecision(-2);

    ui->multimeterdisplayMaxCh2->setValueFontSize(9);
    ui->multimeterdisplayMaxCh2->setUnit("V");
    ui->multimeterdisplayMaxCh2->setLabel("Max");
    ui->multimeterdisplayMaxCh2->setPrecision(-2);

    ui->multimeterdisplayMinCh2->setValueFontSize(9);
    ui->multimeterdisplayMinCh2->setUnit("V");
    ui->multimeterdisplayMinCh2->setLabel("Min");
    ui->multimeterdisplayMinCh2->setPrecision(-2);

    ui->multimeterdisplayRMSCh2->setValueFontSize(9);
    ui->multimeterdisplayRMSCh2->setUnit("V");
    ui->multimeterdisplayRMSCh2->setLabel("RMS");
    ui->multimeterdisplayRMSCh2->setPrecision(-2);

    ui->multimeterdisplayAverCh2->setValueFontSize(9);
    ui->multimeterdisplayAverCh2->setUnit("V");
    ui->multimeterdisplayAverCh2->setLabel("Average");
    ui->multimeterdisplayAverCh2->setPrecision(-2);

    ui->multimeterdisplayFreqCh2->setValueFontSize(9);
    ui->multimeterdisplayFreqCh2->setUnit("Hz");
    ui->multimeterdisplayFreqCh2->setLabel("Frequency");
    ui->multimeterdisplayFreqCh2->setPrecision(-1);
    ui->multimeterdisplayFreqCh2->setUnitPower(3);

    // multimeter frequency
    ui->multimeterdisplayFrequencyValue->connectOptions(tomap_MultimeterDisplayOptions, this, SLOT(callSettings(QString)), SIGNAL(applyConfigurationRequest()));
    ui->multimeterdisplayFrequencyValue->setValueFontSize(20);
    ui->multimeterdisplayFrequencyValue->setLabelFontSize(12);
    ui->multimeterdisplayFrequencyValue->setUnit("Hz");
    ui->multimeterdisplayFrequencyValue->setLabel("Frequency");
    ui->multimeterdisplayFrequencyValue->setPrecision(-2);
    ui->multimeterdisplayFrequencyValue->enableContexMenu(true);
    ui->multimeterdisplayFrequencyValue->setButtonsVisible(true);
    ui->multimeterdisplayFrequencyValue->setAccuracy(1, "50ppm");
    ui->multimeterdisplayFrequencyValue->setAccuracyVisible(true);
    ui->multimeterdisplayFrequencyValue->enableExtreme(true);

    ui->multimeterdisplayDutyValue->connectOptions(tomap_MultimeterDisplayOptions, this, SLOT(callSettings(QString)), SIGNAL(applyConfigurationRequest()));
    ui->multimeterdisplayDutyValue->setValueFontSize(20);
    ui->multimeterdisplayDutyValue->setLabelFontSize(12);
    ui->multimeterdisplayDutyValue->setUnit("%");
    ui->multimeterdisplayDutyValue->setLabel("Duty");
    ui->multimeterdisplayDutyValue->setPrecision(-2);
    ui->multimeterdisplayDutyValue->enableContexMenu(true);
    ui->multimeterdisplayDutyValue->setButtonsVisible(true);
    ui->multimeterdisplayDutyValue->setAccuracy(0.000000125, "50ppm");
    ui->multimeterdisplayDutyValue->setAccuracyVisible(true);
    ui->multimeterdisplayDutyValue->setRange(0, 100);
    ui->multimeterdisplayDutyValue->enableExtreme(true);
    ui->multimeterdisplayDutyValue->enableBargraph(true);

    ui->multimeterdisplayDutyValue_s->connectOptions(tomap_MultimeterDisplayOptions, this, SLOT(callSettings(QString)), SIGNAL(applyConfigurationRequest()));
    ui->multimeterdisplayDutyValue_s->setValueFontSize(20);
    ui->multimeterdisplayDutyValue_s->setLabelFontSize(12);
    ui->multimeterdisplayDutyValue_s->setUnit("s");
    ui->multimeterdisplayDutyValue_s->setLabel("Duty");
    ui->multimeterdisplayDutyValue_s->setUnitPower(-6);
    ui->multimeterdisplayDutyValue_s->setPrecision(-8);
    ui->multimeterdisplayDutyValue_s->enableContexMenu(true);
    ui->multimeterdisplayDutyValue_s->setButtonsVisible(true);
    ui->multimeterdisplayDutyValue_s->setAccuracy(0.0000000625, "50ppm");
    ui->multimeterdisplayDutyValue_s->setAccuracyVisible(true);
    ui->multimeterdisplayDutyValue_s->enableExtreme(true);

    ui->multimeterdisplayPeriodValue->connectOptions(tomap_MultimeterDisplayOptions, this, SLOT(callSettings(QString)), SIGNAL(applyConfigurationRequest()));
    ui->multimeterdisplayPeriodValue->setValueFontSize(20);
    ui->multimeterdisplayPeriodValue->setLabelFontSize(12);
    ui->multimeterdisplayPeriodValue->setUnit("s");
    ui->multimeterdisplayPeriodValue->setLabel("Period");
    ui->multimeterdisplayPeriodValue->setUnitPower(-6);
    ui->multimeterdisplayPeriodValue->setPrecision(-8);
    ui->multimeterdisplayPeriodValue->enableContexMenu(true);
    ui->multimeterdisplayPeriodValue->setButtonsVisible(true);
    ui->multimeterdisplayPeriodValue->setAccuracy(0.0000000625, "50ppm");
    ui->multimeterdisplayPeriodValue->setAccuracyVisible(true);
    ui->multimeterdisplayPeriodValue->enableExtreme(true);

    ui->multimeterdisplayCounter->connectOptions(tomap_MultimeterDisplayOptions, this, SLOT(callSettings(QString)), SIGNAL(applyConfigurationRequest()));
    ui->multimeterdisplayCounter->setValueFontSize(20);
    ui->multimeterdisplayCounter->setLabelFontSize(12);
    ui->multimeterdisplayCounter->setUnit("");
    ui->multimeterdisplayCounter->setLabel("Counter");
    ui->multimeterdisplayCounter->setUnitPower(3);
    ui->multimeterdisplayCounter->enableContexMenu(true);
    ui->multimeterdisplayCounter->setButtonsVisible(true);

    // multimeter DC
    ui->multimeterdisplayCh1DC->connectOptions(tomap_MultimeterDisplayOptions, this, SLOT(callSettings(QString)), SIGNAL(applyConfigurationRequest()));
    ui->multimeterdisplayCh1DC->setValueFontSize(22);
    ui->multimeterdisplayCh1DC->setLabelFontSize(14);
    ui->multimeterdisplayCh1DC->setUnit("V");
    ui->multimeterdisplayCh1DC->setLabel("");
    ui->multimeterdisplayCh1DC->setPrecision(-3);
    ui->multimeterdisplayCh1DC->enableExtreme(true);
    ui->multimeterdisplayCh1DC->setRange(-24, 24);
    ui->multimeterdisplayCh1DC->enableCheckRange(true);
    ui->multimeterdisplayCh1DC->setButtonsVisible(true);
    ui->multimeterdisplayCh1DC->setAccuracy(0, "1%");
    ui->multimeterdisplayCh1DC->setAccuracyVisible(true);
    ui->multimeterdisplayCh1DC->enableContexMenu(true);
    ui->multimeterdisplayCh1DC->enableBargraph(true);


    ui->multimeterdisplayCh2DC->connectOptions(tomap_MultimeterDisplayOptions, this, SLOT(callSettings(QString)), SIGNAL(applyConfigurationRequest()));
    ui->multimeterdisplayCh2DC->setValueFontSize(22);
    ui->multimeterdisplayCh2DC->setLabelFontSize(14);
    ui->multimeterdisplayCh2DC->setUnit("V");
    ui->multimeterdisplayCh2DC->setLabel("");
    ui->multimeterdisplayCh2DC->setPrecision(-3);
    ui->multimeterdisplayCh2DC->enableExtreme(true);
    ui->multimeterdisplayCh2DC->setRange(-24, 24);
    ui->multimeterdisplayCh2DC->enableCheckRange(true);
    ui->multimeterdisplayCh2DC->setButtonsVisible(true);
    ui->multimeterdisplayCh2DC->setAccuracy(0, "1%");
    ui->multimeterdisplayCh2DC->setAccuracyVisible(true);
    ui->multimeterdisplayCh2DC->enableContexMenu(true);
    ui->multimeterdisplayCh2DC->enableBargraph(true);

    // multimeter AC CH1
    ui->multimeterdisplayCh1Vpp->connectOptions(tomap_MultimeterDisplayOptions, this, SLOT(callSettings(QString)), SIGNAL(applyConfigurationRequest()));
    ui->multimeterdisplayCh1Vpp->setValueFontSize(18);
    ui->multimeterdisplayCh1Vpp->setLabelFontSize(12);
    ui->multimeterdisplayCh1Vpp->enableExtreme(true);
    ui->multimeterdisplayCh1Vpp->setUnit("V");
    ui->multimeterdisplayCh1Vpp->setLabel("Vpp");
    ui->multimeterdisplayCh1Vpp->setPrecision(-2);
    ui->multimeterdisplayCh1Vpp->setRange(0, 48);
    ui->multimeterdisplayCh1Vpp->enableCheckRange(true);
    ui->multimeterdisplayCh1Vpp->setButtonsVisible(true);
    ui->multimeterdisplayCh1Vpp->setAccuracy(0, "1%");
    ui->multimeterdisplayCh1Vpp->setAccuracyVisible(true);
    ui->multimeterdisplayCh1Vpp->enableContexMenu(true);

    ui->multimeterdisplayCh1Max->connectOptions(tomap_MultimeterDisplayOptions, this, SLOT(callSettings(QString)), SIGNAL(applyConfigurationRequest()));
    ui->multimeterdisplayCh1Max->setValueFontSize(18);
    ui->multimeterdisplayCh1Max->setLabelFontSize(12);
    ui->multimeterdisplayCh1Max->enableExtreme(true);
    ui->multimeterdisplayCh1Max->setUnit("V");
    ui->multimeterdisplayCh1Max->setLabel("Max");
    ui->multimeterdisplayCh1Max->setPrecision(-2);
    ui->multimeterdisplayCh1Max->setRange(-24, 24);
    ui->multimeterdisplayCh1Max->enableCheckRange(true);
    ui->multimeterdisplayCh1Max->setButtonsVisible(true);
    ui->multimeterdisplayCh1Max->setAccuracy(0, "1%");
    ui->multimeterdisplayCh1Max->setAccuracyVisible(true);
    ui->multimeterdisplayCh1Max->enableContexMenu(true);

    ui->multimeterdisplayCh1Min->connectOptions(tomap_MultimeterDisplayOptions, this, SLOT(callSettings(QString)), SIGNAL(applyConfigurationRequest()));
    ui->multimeterdisplayCh1Min->setValueFontSize(18);
    ui->multimeterdisplayCh1Min->setLabelFontSize(12);
    ui->multimeterdisplayCh1Min->enableExtreme(true);
    ui->multimeterdisplayCh1Min->setUnit("V");
    ui->multimeterdisplayCh1Min->setLabel("Min");
    ui->multimeterdisplayCh1Min->setPrecision(-2);
    ui->multimeterdisplayCh1Min->setRange(-24, 24);
    ui->multimeterdisplayCh1Min->enableCheckRange(true);
    ui->multimeterdisplayCh1Min->setButtonsVisible(true);
    ui->multimeterdisplayCh1Min->setAccuracy(0, "1%");
    ui->multimeterdisplayCh1Min->setAccuracyVisible(true);
    ui->multimeterdisplayCh1Min->enableContexMenu(true);

    ui->multimeterdisplayCh1RMS->connectOptions(tomap_MultimeterDisplayOptions, this, SLOT(callSettings(QString)), SIGNAL(applyConfigurationRequest()));
    ui->multimeterdisplayCh1RMS->setValueFontSize(18);
    ui->multimeterdisplayCh1RMS->setLabelFontSize(12);
    ui->multimeterdisplayCh1RMS->setUnit("V");
    ui->multimeterdisplayCh1RMS->setLabel("RMS");
    ui->multimeterdisplayCh1RMS->setPrecision(-3);
    ui->multimeterdisplayCh1RMS->setRange(0, 24);
    ui->multimeterdisplayCh1RMS->setAccuracy(0, "1%");
    ui->multimeterdisplayCh1RMS->setAccuracyVisible(true);
    ui->multimeterdisplayCh1RMS->enableContexMenu(true);

    ui->multimeterdisplayCh1Average->connectOptions(tomap_MultimeterDisplayOptions, this, SLOT(callSettings(QString)), SIGNAL(applyConfigurationRequest()));
    ui->multimeterdisplayCh1Average->setValueFontSize(18);
    ui->multimeterdisplayCh1Average->setLabelFontSize(12);
    ui->multimeterdisplayCh1Average->setUnit("V");
    ui->multimeterdisplayCh1Average->setLabel("Average");
    ui->multimeterdisplayCh1Average->setPrecision(-3);
    ui->multimeterdisplayCh1Average->setRange(-24, 24);
    ui->multimeterdisplayCh1Average->setAccuracy(0, "1%");
    ui->multimeterdisplayCh1Average->setAccuracyVisible(true);
    ui->multimeterdisplayCh1Average->enableContexMenu(true);

    ui->multimeterdisplayCh1Frequency->connectOptions(tomap_MultimeterDisplayOptions, this, SLOT(callSettings(QString)), SIGNAL(applyConfigurationRequest()));
    ui->multimeterdisplayCh1Frequency->setValueFontSize(18);
    ui->multimeterdisplayCh1Frequency->setLabelFontSize(12);
    ui->multimeterdisplayCh1Frequency->enableExtreme(true);
    ui->multimeterdisplayCh1Frequency->setUnit("Hz");
    ui->multimeterdisplayCh1Frequency->setLabel("Frequency");
    ui->multimeterdisplayCh1Frequency->setUnitPower(3);
    ui->multimeterdisplayCh1Frequency->setRange(0, 100000);
    ui->multimeterdisplayCh1Frequency->enableCheckRange(true);
    ui->multimeterdisplayCh1Frequency->setButtonsVisible(true);
    ui->multimeterdisplayCh1Frequency->setAccuracy(1, "50ppm");
    ui->multimeterdisplayCh1Frequency->setAccuracyVisible(true);
    ui->multimeterdisplayCh1Frequency->enableContexMenu(true);

    // multimeter AC CH2
    ui->multimeterdisplayCh2Vpp->connectOptions(tomap_MultimeterDisplayOptions, this, SLOT(callSettings(QString)), SIGNAL(applyConfigurationRequest()));
    ui->multimeterdisplayCh2Vpp->setValueFontSize(18);
    ui->multimeterdisplayCh2Vpp->setLabelFontSize(12);
    ui->multimeterdisplayCh2Vpp->enableExtreme(true);
    ui->multimeterdisplayCh2Vpp->setUnit("V");
    ui->multimeterdisplayCh2Vpp->setLabel("Vpp");
    ui->multimeterdisplayCh2Vpp->setPrecision(-2);
    ui->multimeterdisplayCh2Vpp->setRange(0, 48);
    ui->multimeterdisplayCh2Vpp->enableCheckRange(true);
    ui->multimeterdisplayCh2Vpp->setButtonsVisible(true);
    ui->multimeterdisplayCh2Vpp->setAccuracy(0, "1%");
    ui->multimeterdisplayCh2Vpp->setAccuracyVisible(true);
    ui->multimeterdisplayCh2Vpp->enableContexMenu(true);

    ui->multimeterdisplayCh2Max->connectOptions(tomap_MultimeterDisplayOptions, this, SLOT(callSettings(QString)), SIGNAL(applyConfigurationRequest()));
    ui->multimeterdisplayCh2Max->setValueFontSize(18);
    ui->multimeterdisplayCh2Max->setLabelFontSize(12);
    ui->multimeterdisplayCh2Max->enableExtreme(true);
    ui->multimeterdisplayCh2Max->setUnit("V");
    ui->multimeterdisplayCh2Max->setLabel("Max");
    ui->multimeterdisplayCh2Max->setPrecision(-2);
    ui->multimeterdisplayCh2Max->setRange(-24, 24);
    ui->multimeterdisplayCh2Max->enableCheckRange(true);
    ui->multimeterdisplayCh2Max->setButtonsVisible(true);
    ui->multimeterdisplayCh2Max->setAccuracy(0, "1%");
    ui->multimeterdisplayCh2Max->setAccuracyVisible(true);
    ui->multimeterdisplayCh2Max->enableContexMenu(true);

    ui->multimeterdisplayCh2Min->connectOptions(tomap_MultimeterDisplayOptions, this, SLOT(callSettings(QString)), SIGNAL(applyConfigurationRequest()));
    ui->multimeterdisplayCh2Min->setValueFontSize(18);
    ui->multimeterdisplayCh2Min->setLabelFontSize(12);
    ui->multimeterdisplayCh2Min->enableExtreme(true);
    ui->multimeterdisplayCh2Min->setUnit("V");
    ui->multimeterdisplayCh2Min->setLabel("Min");
    ui->multimeterdisplayCh2Min->setPrecision(-2);
    ui->multimeterdisplayCh2Min->setRange(-24, 24);
    ui->multimeterdisplayCh2Min->enableCheckRange(true);
    ui->multimeterdisplayCh2Min->setButtonsVisible(true);
    ui->multimeterdisplayCh2Min->setAccuracy(0, "1%");
    ui->multimeterdisplayCh2Min->setAccuracyVisible(true);
    ui->multimeterdisplayCh2Min->enableContexMenu(true);

    ui->multimeterdisplayCh2RMS->connectOptions(tomap_MultimeterDisplayOptions, this, SLOT(callSettings(QString)), SIGNAL(applyConfigurationRequest()));
    ui->multimeterdisplayCh2RMS->setValueFontSize(18);
    ui->multimeterdisplayCh2RMS->setLabelFontSize(12);
    ui->multimeterdisplayCh2RMS->setUnit("V");
    ui->multimeterdisplayCh2RMS->setLabel("RMS");
    ui->multimeterdisplayCh2RMS->setPrecision(-3);
    ui->multimeterdisplayCh2RMS->setRange(0, 24);
    ui->multimeterdisplayCh2RMS->setAccuracy(0, "1%");
    ui->multimeterdisplayCh2RMS->setAccuracyVisible(true);
    ui->multimeterdisplayCh2RMS->enableContexMenu(true);

    ui->multimeterdisplayCh2Average->connectOptions(tomap_MultimeterDisplayOptions, this, SLOT(callSettings(QString)), SIGNAL(applyConfigurationRequest()));
    ui->multimeterdisplayCh2Average->setValueFontSize(18);
    ui->multimeterdisplayCh2Average->setLabelFontSize(12);
    ui->multimeterdisplayCh2Average->setUnit("V");
    ui->multimeterdisplayCh2Average->setLabel("Average");
    ui->multimeterdisplayCh2Average->setPrecision(-3);
    ui->multimeterdisplayCh2Average->setRange(-24, 24);
    ui->multimeterdisplayCh2Average->setAccuracy(0, "1%");
    ui->multimeterdisplayCh2Average->setAccuracyVisible(true);
    ui->multimeterdisplayCh2Average->enableContexMenu(true);

    ui->multimeterdisplayCh2Frequency->connectOptions(tomap_MultimeterDisplayOptions, this, SLOT(callSettings(QString)), SIGNAL(applyConfigurationRequest()));
    ui->multimeterdisplayCh2Frequency->setValueFontSize(18);
    ui->multimeterdisplayCh2Frequency->setLabelFontSize(12);
    ui->multimeterdisplayCh2Frequency->enableExtreme(true);
    ui->multimeterdisplayCh2Frequency->setUnit("Hz");
    ui->multimeterdisplayCh2Frequency->setLabel("Frequency");
    ui->multimeterdisplayCh2Frequency->setUnitPower(3);
    ui->multimeterdisplayCh2Frequency->setRange(0, 100000);
    ui->multimeterdisplayCh2Frequency->enableCheckRange(true);
    ui->multimeterdisplayCh2Frequency->setButtonsVisible(true);
    ui->multimeterdisplayCh2Frequency->setAccuracy(1, "50ppm");
    ui->multimeterdisplayCh2Frequency->setAccuracyVisible(true);
    ui->multimeterdisplayCh2Frequency->enableContexMenu(true);

    // AWG
    ui->multimeterdisplayAWGFrequency->setValueFontSize(11);
    ui->multimeterdisplayAWGFrequency->setUnit("Hz");
    ui->multimeterdisplayAWGFrequency->setLabel("Effective frequency");
    ui->multimeterdisplayAWGFrequency->setPrecision(-2);
    ui->multimeterdisplayAWGFrequency->setStretch(40);


    ui->progressBar->setValue(0);
    setHorizontalScrollBar(ui->plotterWidget->xAxis->range().lower, ui->plotterWidget->xAxis->range().upper);
    setVerticalScrollBar(ui->plotterWidget->yAxis->range().lower, ui->plotterWidget->yAxis->range().upper);
    clearBuffer(awgBuffer, 2 * AWG_WAVE_TEMPLATE_LENGTH);
    ui->toolButtonDACOffsetOK->setEnabled(false);
    ui->toolButtonDACGainOK->setEnabled(false);

    qApp->processEvents();

    QTimer::singleShot(0, this, SLOT(on_connectButton_clicked()));
    initializing = false;
    connect(&dataTimer, SIGNAL(timeout()), this, SLOT(dataShow()));
    connect(&m_mmTimer, SIGNAL(timeout()), this, SLOT(meterRequestData()));

    m_mainTimerDelay = 4;
    dataTimer.start(m_mainTimerDelay); // Interval 0 means to refresh as fast as possible

    triggerSetEnable(!initializing);
    writeBulkConfig = false;
    deviceSettingsLoad();
    deviceSettingsExec();
 //   connect(&usbDevice, SIGNAL(connectionLost()), this, SLOT(on_usbConnection_Lost()));
    connect(&usbDevice.serial, SIGNAL(connectionStatus(QString)), this, SLOT(setInfoText(QString)));
    connect(this, SIGNAL(applyConfigurationRequest()), this, SLOT(applyScopeConfiguration()));

    ui->collapsiblePanelAWGWaveform->setContentsMargins(2, 2, 2, 2);
    ui->collapsiblePanelAWGWaveform->update();
    ui->collapsiblePanelAWGFrequency->setContentsMargins(2, 2, 2, 2);
    ui->collapsiblePanelAWGFrequency->update();
    ui->collapsiblePanelAWGLevels->setContentsMargins(2, 2, 2, 2);
    ui->collapsiblePanelAWGLevels->update();
    ui->collapsiblePanelAWGSweep->setContentsMargins(2, 2, 2, 2);
    ui->collapsiblePanelAWGSweep->update();
    ui->collapsiblePanelScopeCursors->setContentsMargins(2, 2, 2, 2);
    ui->collapsiblePanelScopeCursors->update();
    ui->collapsiblePanelScopeDisplay->setContentsMargins(2, 2, 2, 2);
    ui->collapsiblePanelScopeDisplay->update();
    ui->collapsiblePanelScopeRecord->setContentsMargins(2, 2, 2, 2);
    ui->collapsiblePanelScopeRecord->update();

    ui->dockWidget_awg->setWidget(ui->scrollArea);
    ui->dockWidget_scope->setWidget(ui->mainTabWidget);

    // disabled
    ui->actionImport->setEnabled(false);
    ui->actionExport->setEnabled(false);
    ui->actionConnect->setEnabled(false);

    // after import config
    emit applyConfigurationRequest();
}

XlabInterface::~XlabInterface()
{
    if (recordWaveFile->isOpen()) {
        recordWaveFile->close();
        delete recordWaveFile;
        delete recordWaveTextStream;
    }
    fclose(stderr);
    delete ui;
}

void XlabInterface::closeEvent(QCloseEvent *event)
{
    closeApplication();
    event->accept();
}

void XlabInterface::closeApplication()
{
    logToFile("Application goes down", LOGGING_LEVEL_Warning);
    dataTimer.stop();
    m_mmTimer.stop();
    disconnectDevice(true);
    saveAppSettings();
}

void XlabInterface::itemDoubleClick(QCPAbstractItem *item, QMouseEvent *event)
{
    Q_UNUSED(event)
    if (item) { // only react if item was clicked (user could have clicked on border padding of legend where there is no item, then item is 0)
        QCPItemText *plItem = qobject_cast<QCPItemText*>(item);
        if (!plItem) return;
        bool ok = false;
        QString newName = QInputDialog::getText(this, "Rename", "New Label:", QLineEdit::Normal, plItem->text(), &ok);
        if (ok) plItem->setText(newName);
    }
}

bool XlabInterface::checkIfCorrectPort(QString name)
{
#ifdef Q_OS_WIN
    return name.toLower().startsWith("com");
#elif defined(Q_OS_MAC)
    return name.toLower().startsWith("bluetooth");
#else
    return name.toLower().startsWith("tty");
#endif
}

void XlabInterface::setupGrid(QCustomPlot *customPlot)
{
    setupGraphs(customPlot);
    setupItemLabels(customPlot);
    setupCursors(customPlot);
    setupTracers(customPlot);
    setTheme(tomap_ScopeOptions->theme, &customThemeDialog.customColors);

    QSharedPointer<QCPAxisTickerFixed> fixedTickerX(new QCPAxisTickerFixed);
    fixedTickerX->setTickStep(xmax / 8);
    fixedTickerX->setScaleStrategy(QCPAxisTickerFixed::ssNone);

    customPlot->xAxis->setTickLabels(false);
    customPlot->xAxis->setTicker(fixedTickerX);
    customPlot->xAxis->setRange(0, xmax);

    customPlot->xAxis2->setTickLabels(false);
    customPlot->xAxis2->setTicker(fixedTickerX);
    customPlot->xAxis2->setRange(0, xmax);

    QSharedPointer<QCPAxisTickerFixed> fixedTickerY(new QCPAxisTickerFixed);
    fixedTickerY->setTickStep(rangeMax / 8);
    fixedTickerY->setScaleStrategy(QCPAxisTickerFixed::ssNone);

    customPlot->yAxis->setTickLabels(false);
    customPlot->yAxis->setTicker(fixedTickerY);
    customPlot->yAxis->setRange(0, rangeMax);

    customPlot->yAxis2->setTickLabels(false);
    customPlot->yAxis2->setTicker(fixedTickerY);
    customPlot->yAxis2->setRange(0, rangeMax);

    customPlot->axisRect()->setupFullAxesBox();
    customPlot->axisRect()->setRangeDrag(Qt::Horizontal);

    // make left and bottom axis transfer their ranges to right and top axis:
    connect(customPlot->xAxis, SIGNAL(rangeChanged(QCPRange)), customPlot->xAxis2, SLOT(setRange(QCPRange)));
    connect(customPlot->yAxis, SIGNAL(rangeChanged(QCPRange)), customPlot->yAxis2, SLOT(setRange(QCPRange)));
    connect(customPlot->xAxis, SIGNAL(rangeChanged(QCPRange)), this, SLOT(xAxisChanged(QCPRange)));
    connect(customPlot->yAxis, SIGNAL(rangeChanged(QCPRange)), this, SLOT(yAxisChanged(QCPRange)));

    customPlot->setInteractions(QCP::iSelectPlottables);
    customPlot->setInteractions(QCP::iRangeZoom);
    customPlot->axisRect()->setRangeZoomFactor(0.8);
}

void XlabInterface::setTheme(int theme, CustomColors *customColors)
{
    QColor FillColor;

    if (theme == DISPLAY_THEME_Dark) {
        /************** Graph Pens **********/
        ch1Pen = QPen(QColor(75, 229, 28), 2);
        ch1Graph->setPen(ch1Pen);

        ch2Pen = QPen(QColor(150, 150, 150), 2);
        ch2Graph->setPen(ch2Pen);
        QString styleSheet;
        styleSheet = "QLabel { background-color : " + ch1Pen.color().name() + "; }";
        ui->ch1ColorLabel->setStyleSheet(styleSheet);
        styleSheet = "QLabel { background-color : " + ch2Pen.color().name() + "; }";
        ui->ch2ColorLabel->setStyleSheet(styleSheet);

        ch1RefPen = QPen(Qt::red, 2);
        ch1RefGraph->setPen(ch1RefPen);

        ch2RefPen = QPen(QColor("#4be51c"), 2);
        ch2RefGraph->setPen(ch2RefPen);

        for (int i = 0; i < TG; i++) {
            ch1PPen[i] = QPen(QColor(75, 229, 28), (TG - i) / TG);
            ch2PPen[i] = QPen(Qt::red, (TG - i) / TG);
            ch1PGraphs[i]->setPen(ch1PPen[i]);
            ch2PGraphs[i]->setPen(ch2PPen[i]);
        }

        for (int i = 0; i < 8; i++) {
            chdPen[i] = QPen(QColor(255 - i * 10, 4 + i * 10, 4), 1.5);
            FillColor = QColor(255 - i * 10, 4 + i * 10, 4, 35);
            chdGraph[i]->setPen(chdPen[i]);
            chd0Graph[i]->setPen(QPen(FillColor, 1.5));
            chdGraph[i]->setChannelFillGraph(chd0Graph[i]);
            chdGraph[i]->setBrush(QBrush(FillColor));
            chdRefPen[i] = QPen(Qt::blue, 1.5);
            chdRefGraph[i]->setPen(chdRefPen[i]);
        }

        ch1BarPen = QPen(QColor("#4be51c"), 2);
        ch1BarGraph->setPen(ch1BarPen);

        ch2BarPen = QPen(Qt::red, 2);
        ch2BarGraph->setPen(ch2BarPen);

        /************** Grid Pens **********/
        gridPenColor = QColor(140, 140, 140);
        gridPenStyle = Qt::DotLine;
        gridPen = QPen(gridPenColor, 1, gridPenStyle);
        gridBoxSet(gridPenColor, gridPenStyle);

        axesPen = QPen(Qt::white, 1);
        ui->plotterWidget->xAxis->setTickPen(axesPen);
        ui->plotterWidget->yAxis->setTickPen(axesPen);
        ui->plotterWidget->xAxis->setSubTickPen(axesPen);
        ui->plotterWidget->yAxis->setSubTickPen(axesPen);
        ui->plotterWidget->xAxis2->setTickPen(axesPen);
        ui->plotterWidget->yAxis2->setTickPen(axesPen);
        ui->plotterWidget->xAxis2->setSubTickPen(axesPen);
        ui->plotterWidget->yAxis2->setSubTickPen(axesPen);
        ui->plotterWidget->xAxis->setBasePen(axesPen);
        ui->plotterWidget->yAxis->setBasePen(axesPen);
        ui->plotterWidget->xAxis2->setBasePen(axesPen);
        ui->plotterWidget->yAxis2->setBasePen(axesPen);

        ui->plotterWidget->setBackground(QBrush(Qt::black));

        vCursorSetLabelColor("#4be51c");
        hCursorSetLabelColor("#4be51c");
    } else if (theme == DISPLAY_THEME_Light) {
        /************** Graph Pens **********/
        ch1Pen = QPen(QColor(75, 229, 28), 2);
        ch1Graph->setPen(ch1Pen);

        ch2Pen = QPen(Qt::red, 2);
        ch2Graph->setPen(ch2Pen);
        QString styleSheet;
        styleSheet = "QLabel { background-color : " + ch1Pen.color().name() + "; }";
        ui->ch1ColorLabel->setStyleSheet(styleSheet);
        styleSheet = "QLabel { background-color : " + ch2Pen.color().name() + "; }";
        ui->ch2ColorLabel->setStyleSheet(styleSheet);

        ch1RefPen = QPen(Qt::red, 2);
        ch1RefGraph->setPen(ch1RefPen);

        ch2RefPen = QPen(QColor("#4be51c"), 2);
        ch2RefGraph->setPen(ch2RefPen);

        for (int i = 0; i < TG; i++) {
            ch1PPen[i] = QPen(QColor("#2a50fd"), (TG - i) / TG);
            ch2PPen[i] = QPen(Qt::red, (TG - i) / TG);
            ch1PGraphs[i]->setPen(ch1PPen[i]);
            ch2PGraphs[i]->setPen(ch2PPen[i]);
        }

        for (int i = 0; i < 8; i++) {
            chdPen[i] = QPen(QColor(4 + i * 20, i, 255 - i * 5), 1.5);
            //chdPen[i] = QPen(Qt::red, 1.5);
            chdGraph[i]->setPen(chdPen[i]);
            chd0Graph[i]->setPen(chdPen[i]);

            chdRefPen[i] = QPen(Qt::blue, 1.5);
            chdRefGraph[i]->setPen(chdRefPen[i]);
        }

        ch1BarPen = QPen(QColor("#2a50fd"), 2);
        ch1BarGraph->setPen(ch1BarPen);

        ch2BarPen = QPen(Qt::red, 2);
        ch2BarGraph->setPen(ch2BarPen);

        /************** Grid Pens **********/
        gridPenColor = QColor("C8C8C8");
        gridPenStyle = Qt::DotLine;
        gridPen = QPen(gridPenColor, 1, gridPenStyle);
        gridBoxSet(gridPenColor, gridPenStyle);

        axesPen = QPen(Qt::black, 1);
        ui->plotterWidget->xAxis->setTickPen(axesPen);
        ui->plotterWidget->yAxis->setTickPen(axesPen);
        ui->plotterWidget->xAxis->setSubTickPen(axesPen);
        ui->plotterWidget->yAxis->setSubTickPen(axesPen);
        ui->plotterWidget->xAxis2->setTickPen(axesPen);
        ui->plotterWidget->yAxis2->setTickPen(axesPen);
        ui->plotterWidget->xAxis2->setSubTickPen(axesPen);
        ui->plotterWidget->yAxis2->setSubTickPen(axesPen);
        ui->plotterWidget->xAxis->setBasePen(axesPen);
        ui->plotterWidget->yAxis->setBasePen(axesPen);
        ui->plotterWidget->xAxis2->setBasePen(axesPen);
        ui->plotterWidget->yAxis2->setBasePen(axesPen);
        ui->plotterWidget->setBackground(QBrush(Qt::white));

        vCursorSetLabelColor(Qt::red);
        hCursorSetLabelColor(Qt::red);
    } else if (theme == DISPLAY_THEME_Custom) {
        /************** Graph Pens **********/
        ch1Pen = QPen(customColors->ch1, 2);
        ch1Graph->setPen(ch1Pen);

        ch2Pen = QPen(customColors->ch2, 2);
        ch2Graph->setPen(ch2Pen);
        QString styleSheet;
        styleSheet = "QLabel { background-color : " + ch1Pen.color().name() + "; }";
        ui->ch1ColorLabel->setStyleSheet(styleSheet);
        styleSheet = "QLabel { background-color : " + ch2Pen.color().name() + "; }";
        ui->ch2ColorLabel->setStyleSheet(styleSheet);

        ch1RefPen = QPen(customColors->ch1ref, 2);
        ch1RefGraph->setPen(ch1RefPen);

        ch2RefPen = QPen(customColors->ch2ref, 2);
        ch2RefGraph->setPen(ch2RefPen);

        for (int i = 0; i < TG; i++) {
            ch1PPen[i] = QPen(customColors->ch1, (TG - i) / TG);
            ch2PPen[i] = QPen(customColors->ch2, (TG - i) / TG);
            ch1PGraphs[i]->setPen(ch1PPen[i]);
            ch2PGraphs[i]->setPen(ch2PPen[i]);
        }

        for (int i = 0; i < 8; i++) {
            chdPen[i] = QPen(customColors->bit[i], 1.5);
            chdGraph[i]->setPen(chdPen[i]);
            chd0Graph[i]->setPen(chdPen[i]);

            chdRefPen[i] = QPen(customColors->bitref[i], 1.5);
            chdRefGraph[i]->setPen(chdRefPen[i]);
        }

        ch1BarPen = QPen(customColors->ch1fft, 2);
        ch1BarGraph->setPen(ch1BarPen);

        ch2BarPen = QPen(customColors->ch2fft, 2);
        ch2BarGraph->setPen(ch2BarPen);

        /************** Grid Pens **********/
        gridPenColor = customColors->grid;
        gridPenStyle = Qt::DotLine;
        gridPen = QPen(gridPenColor, 1, gridPenStyle);
        gridBoxSet(gridPenColor, gridPenStyle);

        axesPen = QPen(customColors->axes, 1);
        ui->plotterWidget->xAxis->setTickPen(axesPen);
        ui->plotterWidget->yAxis->setTickPen(axesPen);
        ui->plotterWidget->xAxis->setSubTickPen(axesPen);
        ui->plotterWidget->yAxis->setSubTickPen(axesPen);
        ui->plotterWidget->xAxis2->setTickPen(axesPen);
        ui->plotterWidget->yAxis2->setTickPen(axesPen);
        ui->plotterWidget->xAxis2->setSubTickPen(axesPen);
        ui->plotterWidget->yAxis2->setSubTickPen(axesPen);
        ui->plotterWidget->xAxis->setBasePen(axesPen);
        ui->plotterWidget->yAxis->setBasePen(axesPen);
        ui->plotterWidget->xAxis2->setBasePen(axesPen);
        ui->plotterWidget->yAxis2->setBasePen(axesPen);

        ui->plotterWidget->setBackground(customColors->background);

        vCursorSetLabelColor(customColors->label);
        hCursorSetLabelColor(customColors->label);
    }
    on_intensitySlider_valueChanged(0);
    ui->plotterWidget->replot();
}

void XlabInterface::setupGraphs(QCustomPlot *customPlot)
{
    ui->plotterWidget->setNoAntialiasingOnDrag(true);
    ch1Graph = customPlot->addGraph();
    ch1RefGraph = customPlot->addGraph();
    ch2Graph = customPlot->addGraph();
    ch2RefGraph = customPlot->addGraph();

    for (int i = 0; i < TG; i++) {
        ch1PGraphs[i] = customPlot->addGraph();
        ch2PGraphs[i] = customPlot->addGraph();
    }

    for (int i = 0; i < 8; i++) {
        chdGraph[i] = customPlot->addGraph();
        chdGraph[i]->setLineStyle(QCPGraph::lsStepCenter);
        chdGraph[i]->setSelectable(QCP::stNone);

        chd0Graph[i] = customPlot->addGraph();
        chd0Graph[i]->setLineStyle(QCPGraph::lsStepCenter);
        chd0Graph[i]->setSelectable(QCP::stNone);

        chdRefGraph[i] = customPlot->addGraph();
        chdRefGraph[i]->setLineStyle(QCPGraph::lsStepCenter);
    }
    ch1BarGraph = customPlot->addGraph();
    ch1BarGraph->setLineStyle(QCPGraph::lsImpulse);

    ch2BarGraph = customPlot->addGraph();
    ch2BarGraph->setLineStyle(QCPGraph::lsImpulse);
}

void XlabInterface::setupScatterStyles(bool lStyle)
{
    QCPScatterStyle sstyle;
    int lineStyle;

    if (lStyle && !ui->xyMode->isChecked()) {
        sstyle = QCPScatterStyle::ssNone;
        lineStyle = QCPGraph::lsLine;
    } else {
        sstyle = QCPScatterStyle::ssDot;
        lineStyle = QCPGraph::lsNone;
    }

    ch1Graph->setScatterStyle(sstyle);
    ch1Graph->setLineStyle((QCPGraph::LineStyle)lineStyle);

    ch1RefGraph->setScatterStyle(sstyle);
    ch1RefGraph->setLineStyle((QCPGraph::LineStyle)lineStyle);

    ch2Graph->setScatterStyle(sstyle);
    ch2Graph->setLineStyle((QCPGraph::LineStyle)lineStyle);

    ch2RefGraph->setScatterStyle(sstyle);
    ch2RefGraph->setLineStyle((QCPGraph::LineStyle)lineStyle);

    for (int i = 0; i < TG; i++) {
        ch1PGraphs[i]->setScatterStyle(sstyle);
        ch1PGraphs[i]->setLineStyle((QCPGraph::LineStyle)lineStyle);
        ch2PGraphs[i]->setScatterStyle(sstyle);
        ch2PGraphs[i]->setLineStyle((QCPGraph::LineStyle)lineStyle);
    }
}

void XlabInterface::setupCursors(QCustomPlot *customPlot)
{
    hCursorA = new QCPItemStraightLine(customPlot);
    hCursorA->setPen(QPen(QColor("#1692e5"), 1, Qt::DotLine));
    hCursorA->setSelectable(false);
    hCursorAHead = new QCPItemPixmap(customPlot);
    hCursorAHead->setPixmap(QPixmap(":/icons/HcursorCA.png"));
    hCursorAHead->topLeft->setPixelPosition(QPointF(14,(ui->plotterWidget->yAxis->coordToPixel(hCursorAPosCh1)) - 8));
    hCursorA->point1->setParentAnchor(hCursorAHead->right);
    hCursorA->point2->setParentAnchor(hCursorAHead->right);
    hCursorA->point1->setCoords(0, 0);
    hCursorA->point2->setCoords(10, 0);
    hCursorAHead->setVisible(false);

    hCursorB = new QCPItemStraightLine(customPlot);
    hCursorB->setPen(QPen(QColor("#1692e5"), 1, Qt::DotLine));
    hCursorB->setSelectable(false);
    hCursorBHead = new QCPItemPixmap(customPlot);
    hCursorBHead->setPixmap(QPixmap(":/icons/HcursorCB.png"));
    hCursorBHead->topLeft->setPixelPosition(QPointF(14,(ui->plotterWidget->yAxis->coordToPixel(hCursorBPosCh1)) - 8));
    hCursorB->point1->setParentAnchor(hCursorBHead->right);
    hCursorB->point2->setParentAnchor(hCursorBHead->right);
    hCursorB->point1->setCoords(0, 0);
    hCursorB->point2->setCoords(10, 0);
    hCursorBHead->setVisible(false);

    vCursorA = new QCPItemStraightLine(customPlot);
    vCursorA->setSelectable(false);
    vCursorAHead = new QCPItemPixmap(customPlot);
    vCursorAHead->setPixmap(QPixmap(":/icons/VcursorC.png"));
    vCursorAHead->topLeft->setPixelPosition(QPointF(150 - 9, 10));
    vCursorAHead->setVisible(false);
    vCursorAHead->setClipToAxisRect(false);
    vCursorAHead->topLeft->setType(QCPItemPosition::ptAbsolute);
    vCursorA->point1->setParentAnchor(vCursorAHead->bottom);
    vCursorA->point2->setParentAnchor(vCursorAHead->bottom);
    vCursorA->point1->setType(QCPItemPosition::ptAbsolute);
    vCursorA->point2->setType(QCPItemPosition::ptAbsolute);
    vCursorA->point1->setCoords(0, 0);
    vCursorA->point2->setCoords(0, 10);
    vCursorA->setPen(QPen(QColor("#e04e4e"), 1, Qt::DotLine));

    vCursorB = new QCPItemStraightLine(customPlot);
    vCursorB->setSelectable(false);
    vCursorBHead = new QCPItemPixmap(customPlot);
    vCursorBHead->setPixmap(QPixmap(":/icons/VcursorC.png"));
    vCursorBHead->topLeft->setPixelPosition(QPointF(100, 10));
    vCursorBHead->setVisible(false);
    vCursorBHead->setClipToAxisRect(false);
    vCursorBHead->topLeft->setType(QCPItemPosition::ptAbsolute);
    vCursorB->point1->setParentAnchor(vCursorBHead->bottom);
    vCursorB->point2->setParentAnchor(vCursorBHead->bottom);
    vCursorB->point1->setType(QCPItemPosition::ptAbsolute);
    vCursorB->point2->setType(QCPItemPosition::ptAbsolute);
    vCursorB->point1->setCoords(0, 0);
    vCursorB->point2->setCoords(0, 10);
    vCursorB->setPen(QPen(QColor("#e04e4e"), 1, Qt::DotLine));

    ch1Zero = new QCPItemStraightLine(customPlot);
    ch1Zero->setPen(QPen(QColor("#f9a94c"), 1, Qt::DotLine));
    ch1Zero->setSelectable(false);
    ch1ZeroHead = new QCPItemPixmap(customPlot);
    ch1ZeroHead->setPixmap(QPixmap(":/icons/chan1-zero.png"));
    ch1ZeroHead->topLeft->setPixelPosition(QPointF(2, hCursorAPos - 8));
    ch1Zero->point1->setParentAnchor(ch1ZeroHead->right);
    ch1Zero->point2->setParentAnchor(ch1ZeroHead->right);
    ch1ZeroHead->setVisible(true);
    ch1ZeroHead->setClipToAxisRect(false);
    ch1ZeroHead->topLeft->setType(QCPItemPosition::ptAbsolute);
    ch1Zero->point1->setPixelPosition(QPointF(1, hCursorAPos));
    ch1Zero->point2->setPixelPosition(QPointF(10, hCursorAPos));
    ch1Zero->point1->setType(QCPItemPosition::ptAbsolute);
    ch1Zero->point2->setType(QCPItemPosition::ptAbsolute);

    ch1vZero = new QCPItemStraightLine(customPlot);
    ch1vZero->setPen(QPen(QColor("#f9a94c"), 1, Qt::DotLine));
    ch1vZero->setSelectable(false);
    ch1vZeroHead = new QCPItemPixmap(customPlot);
    ch1vZeroHead->setPixmap(QPixmap(":/icons/chan1v-zero.png"));
    ch1vZeroHead->topLeft->setPixelPosition(QPointF(256, 1));
    ch1vZero->point1->setParentAnchor(ch1vZeroHead->bottom);
    ch1vZero->point2->setParentAnchor(ch1vZeroHead->bottom);
    ch1vZeroHead->setVisible(true);
    ch1vZeroHead->setClipToAxisRect(false);
    ch1vZeroHead->topLeft->setType(QCPItemPosition::ptAbsolute);
    ch1vZero->point1->setType(QCPItemPosition::ptAbsolute);
    ch1vZero->point2->setType(QCPItemPosition::ptAbsolute);
    ch1vZero->point1->setCoords(0, 0);
    ch1vZero->point2->setCoords(0, 10);

    ch2Zero = new QCPItemStraightLine(customPlot);
    ch2Zero->setPen(QPen(QColor("#f9a94c"), 1, Qt::DotLine));
    ch2Zero->setSelectable(false);
    ch2ZeroHead = new QCPItemPixmap(customPlot);
    ch2ZeroHead->setPixmap(QPixmap(":/icons/chan2-zero.png"));
    ch2ZeroHead->topLeft->setPixelPosition(QPointF(2, hCursorBPos - 8));
    ch2Zero->point1->setParentAnchor(ch2ZeroHead->right);
    ch2Zero->point2->setParentAnchor(ch2ZeroHead->right);
    ch2ZeroHead->setVisible(true);
    ch2ZeroHead->setClipToAxisRect(false);
    ch2ZeroHead->topLeft->setType(QCPItemPosition::ptAbsolute);
    ch2Zero->point1->setPixelPosition(QPointF(1, hCursorBPos));
    ch2Zero->point2->setPixelPosition(QPointF(10, hCursorBPos));
    ch2Zero->point1->setType(QCPItemPosition::ptAbsolute);
    ch2Zero->point2->setType(QCPItemPosition::ptAbsolute);

    chDZero = new QCPItemStraightLine(customPlot);
    chDZero->setPen(QPen(QColor("#f9a94c"), 1, Qt::DotLine));
    chDZero->setSelectable(false);
    chDZeroHead = new QCPItemPixmap(customPlot);
    chDZeroHead->setPixmap(QPixmap(":/icons/chanL-zero.png"));
    chDZeroHead->topLeft->setPixelPosition(QPointF(2, 20 - 8));
    chDZero->point1->setParentAnchor(chDZeroHead->right);
    chDZero->point2->setParentAnchor(chDZeroHead->right);
    chDZeroHead->setVisible(true);
    chDZeroHead->setClipToAxisRect(false);
    chDZeroHead->topLeft->setType(QCPItemPosition::ptAbsolute);

    triggerPixmap = new QCPItemPixmap(customPlot);
    triggerPixmap->setPixmap(QPixmap(":/icons/dualr.png"));
    triggerPixmap->topLeft->setCoords(150 - 9, 220 - 8);
    triggerPixmap->setVisible(false);

    triggerWin1Pixmap = new QCPItemPixmap(customPlot);
    triggerWin1Pixmap->setPixmap(QPixmap(":/icons/risingg.png"));
    triggerWin1Pixmap->topLeft->setCoords(150 - 9, 250 - 8);
    triggerWin1Pixmap->setVisible(false);

    triggerWin2Pixmap = new QCPItemPixmap(customPlot);
    triggerWin2Pixmap->setPixmap(QPixmap(":/icons/fallingg.png"));
    triggerWin2Pixmap->topLeft->setCoords(150 - 9, 250 - 8);
    triggerWin2Pixmap->setVisible(false);
}

void XlabInterface::setupTracers(QCustomPlot *customPlot)
{
    phaseTracerAA = new QCPItemTracer(customPlot);
    phaseTracerAA->setGraphKey(vCursor.AposX);
    phaseTracerAA->setInterpolating(true);
    phaseTracerAA->setStyle(QCPItemTracer::tsCircle);
    phaseTracerAA->setPen(QPen(QColor("#1692e5")));
    phaseTracerAA->setBrush(QColor("#1692e5"));
    phaseTracerAA->setSize(7);

    phaseTracerAB = new QCPItemTracer(customPlot);
    phaseTracerAB->setGraphKey(vCursor.BposX);
    phaseTracerAB->setInterpolating(true);
    phaseTracerAB->setStyle(QCPItemTracer::tsCircle);
    phaseTracerAB->setPen(QPen(QColor("#1692e5")));
    phaseTracerAB->setBrush(QColor("#1692e5"));
    phaseTracerAB->setSize(7);

    phaseTracerBA = new QCPItemTracer(customPlot);
    phaseTracerBA->setGraphKey(vCursor.AposX);
    phaseTracerBA->setInterpolating(true);
    phaseTracerBA->setStyle(QCPItemTracer::tsCircle);
    phaseTracerBA->setPen(QPen(QColor("#1692e5")));
    phaseTracerBA->setBrush(QColor("#1692e5"));
    phaseTracerBA->setSize(7);

    phaseTracerBB = new QCPItemTracer(customPlot);
    phaseTracerBB->setGraphKey(vCursor.AposX);
    phaseTracerBB->setInterpolating(true);
    phaseTracerBB->setStyle(QCPItemTracer::tsCircle);
    phaseTracerBB->setPen(QPen(QColor("#1692e5")));
    phaseTracerBB->setBrush(QColor("#1692e5"));
    phaseTracerBB->setSize(7);
}

void XlabInterface::setupItemLabels(QCustomPlot *customPlot)
{
    QFont m_monospaceFont = QFont("Courier New", 10, QFont::DemiBold);

#ifdef Q_OS_MAC
    m_monospaceFont.setPixelSize(13);
#endif
    for (int i = 0; i < 8; i++) {
        textLabelBit[i] = new QCPItemText(customPlot);
        textLabelBit[i]->setColor(Qt::red);
        textLabelBit[i]->position->setCoords(0, 0);
        textLabelBit[i]->setText("Bit "+QString::number(i));
        textLabelBit[i]->setFont(m_monospaceFont);
        textLabelBit[i]->setSelectable(false);
        textLabelBit[i]->setClipToAxisRect(false);
        textLabelBit[i]->setVisible(false);
    }

    textLabelDeltaTime = new QCPItemText(customPlot);
    textLabelDeltaTime->setPositionAlignment(Qt::AlignLeft);
    textLabelDeltaTime->setText(QString::fromUtf8("ΔT = 0 ms"));
    textLabelDeltaTime->setFont(m_monospaceFont);
    textLabelDeltaTime->setSelectable(false);
    textLabelDeltaTime->setVisible(false);

    textLabelFrequency = new QCPItemText(customPlot);
    textLabelFrequency->setPositionAlignment(Qt::AlignLeft);
    textLabelFrequency->setText(QString::fromUtf8(" 1/ΔT = 0 ms "));
    textLabelFrequency->setFont(m_monospaceFont);
    textLabelFrequency->setSelectable(false);
    textLabelFrequency->setVisible(false);

    textLabelDeltaVoltage = new QCPItemText(customPlot);
    textLabelDeltaVoltage->setPositionAlignment(Qt::AlignLeft);
    textLabelDeltaVoltage->setText("ΔV = 0 V");
    textLabelDeltaVoltage->setFont(m_monospaceFont);
    textLabelDeltaVoltage->setSelectable(false);
    textLabelDeltaVoltage->setVisible(false);

    textLabelVoltageB = new QCPItemText(customPlot);
    textLabelVoltageB->setPositionAlignment(Qt::AlignLeft);
    textLabelVoltageB->setText("Vb = 0 V");
    textLabelVoltageB->setFont(m_monospaceFont);
    textLabelVoltageB->setSelectable(false);
    textLabelVoltageB->setVisible(false);

    textLabelVoltageA = new QCPItemText(customPlot);
    textLabelVoltageA->setPositionAlignment(Qt::AlignLeft);
    textLabelVoltageA->setText("Va = 0 V");
    textLabelVoltageA->setFont(m_monospaceFont);
    textLabelVoltageA->setSelectable(false);
    textLabelVoltageA->setVisible(false);

    ui->stopButton->setText(tr(BUTTON_STOP_SCOPE_TEXT_START));
    ui->pauseSnifferButton->setText(tr("Pause"));
    updateLabelPosition();

    clearMMLabelsCH1AC();
    clearMMLabelsCH2AC();
}

void XlabInterface::updateLabelPosition()
{
    textLabelDeltaTime->position->setPixelPosition(QPointF(ui->plotterWidget->width() - 230, 25));
    textLabelFrequency->position->setPixelPosition(QPointF(ui->plotterWidget->width() - 230, 40));
    textLabelDeltaVoltage->position->setPixelPosition(QPointF(ui->plotterWidget->width() - 180, ui->plotterWidget->height() - 40));
    textLabelVoltageB->position->setPixelPosition(QPointF(ui->plotterWidget->width() - 180, ui->plotterWidget->height() - 55));
    textLabelVoltageA->position->setPixelPosition(QPointF(ui->plotterWidget->width() - 180, ui->plotterWidget->height() - 70));
}

void XlabInterface::setupValues()
{
    rateText << "8" + QString::fromStdWString(L"\u00b5") + "s/div" << "16" + QString::fromStdWString(L"\u00b5") + "s/div" << "32" + QString::fromStdWString(L"\u00b5") + "s/div"
             << "64" + QString::fromStdWString(L"\u00b5") + "s/div" << "128" + QString::fromStdWString(L"\u00b5") + "s/div" << "256"+ QString::fromStdWString(L"\u00b5") + "s/div"
             << "500" + QString::fromStdWString(L"\u00b5") + "s/div" << "1ms/div" << "2ms/div" << "5ms/div" << "10ms/div" << "20ms/div" << "50ms/div" << "0.1s/div" << "0.2s/div"
             << "0.5s/div" << "1s/div" << "2s/div" << "5s/div" << "10s/div" << "20s/div" << "50s/div";
    gainText << "8V/div" << "4V/div" << "2V/div" << "0.8V/div" << "0.4V/div" << "0.2V/div" << "0.1V/div" << "50mV/div" << "25mV/div" << "12.5mV/div" << "----";
    gainSliderText << "Gain (±24V)" << "Gain (±12V)" << "Gain (±6V)" << "Gain (±2.4V)" << "Gain (±1.2V)" << "Gain (±0.6V)" << "Gain (±300mV)" << "Gain (±150mV)"
                   << "Gain (±75mV)" << "Gain (±37mV)";
    freqValue[0] = 10;
    int temp = 2000000, i = 1;
    for (i = 1; i < 7; i++) { // Kilo Hertz
        freqValue[i] = temp;
        temp = temp / 2;
    }
    freqValue[7] = 32000;
    freqValue[8] = 16000000;
    freqValue[9] = 8000000;
    temp = 3200000;
    for (i = 10; i < 13; i++) { // Hertz
        freqValue[i] = temp;
        temp = temp / 2;
    }
    freqValue[13] = 320000;
    freqValue[14] = 160000;
    freqValue[15] = 80000;
    freqValue[16] = 32000;
    freqValue[17] = 16000;
    freqValue[18] = 8000;
    temp = 3200;
    for (i = 19; i < 22; i++) {  // Hertz
        freqValue[i] = temp;
        temp = temp / 2;
    }
    freqValue[22] = 320;
    QString tempP = ":/icons/";
    triggerIconPathsG << tempP + "risingg.png" << tempP + "fallingg.png" << tempP + "dualg.png" << tempP + "positiveg.png" << tempP + "negativeg.png"
                      << tempP + "window1g.png" << tempP + "window2g.png";
    triggerIconPathsR << tempP + "risingr.png" << tempP + "fallingr.png" << tempP + "dualr.png" << tempP + "positiver.png" << tempP + "negativer.png"
                      << tempP + "window1r.png" << tempP + "window2r.png";
}

void XlabInterface::setupAppSetting()
{
    initializing = true;
    QSettings settings;
    int w, h;

    triggerSetEnable(!initializing);
    w = settings.value ("size").toSize().width();
    h = settings.value ("size").toSize().height();
    if (!settings.value ("wasMaximized").toBool() && (w > 0) && (h > 0)) {
        this->setGeometry(0, 0, w, h);
        const QRect screen = QApplication::desktop()->screenGeometry();
        this->move( screen.center() - this->rect().center());
    }
    tomap_ScopeOptions->theme = settings.value ("theme").toInt();
    defaultDir = settings.value ("defaultDir").toString();
    if (defaultDir.isEmpty())
        defaultDir = QDir::homePath();
}

void XlabInterface::saveAppSettings()
{
    deviceSettingsSave();
    logToFile("Save application settings", LOGGING_LEVEL_Debug);
    QSettings settings;
    settings.setValue("size", this->size());
    settings.setValue("wasMaximized", this->isMaximized());
    settings.setValue("defaultDir", defaultDir);
    if (settings.value("loggingLevel", -1).toInt() < 0)
        settings.setValue("loggingLevel", 2);
}

void XlabInterface::clearBuffer(unsigned char *buffer, int length)
{
    for (int i = 0; i < length; i++)
        *(buffer++) = 0;
}

void XlabInterface::changeWindowTitle(QString status)
{
    this->setWindowTitle(QCoreApplication::applicationName() + " v." + QCoreApplication::applicationVersion() + "   " + status);
}

void XlabInterface::dataScopeHeaderGetData(frameconfig_t *head, uint8_t *data)
{
    head->type = *(uint8_t *)(data);
    head->id = *(uint16_t *)(data + 1);
    head->chenable = *(uint8_t *)(data + 3);
    head->chtype = *(uint8_t *)(data + 4);
    head->chlength = *(uint16_t *)(data + 5);
    head->sapmlingrate = *(uint8_t *)(data + 7);
    head->index = *(uint16_t *)(data + 8);
}

void XlabInterface::dataChannelMeterGetData(channela_t *meter, uint8_t *data)
{
    meter->control = *(uint8_t *)(data);
    meter->gain = *(uint8_t *)(data + 1);
    meter->offset = *(int8_t *)(data + 2);
    meter->max = *(int8_t *)(data + 3);
    meter->min = *(int8_t *)(data + 4);
    meter->rms = *(int16_t *)(data + 5);
    meter->aver = *(int16_t *)(data + 7);
    meter->frequency = *(uint32_t *)(data + 9);
    meter->cross1 = *(uint16_t *)(data + 13);
    meter->cross2 = *(uint16_t *)(data + 15);
    meter->samplingrate = *(uint8_t *)(data + 17);
}

void XlabInterface::dataInitChannelMeterBuffer(channela_t *meter)
{
    meter->control = 0;
    meter->gain = 0;
    meter->offset = 0;
    meter->max = 100;
    meter->min = -100;
    meter->rms = 0;
    meter->aver = 0;
    meter->frequency = 0;
    meter->cross1 = 64;
    meter->cross2 = 128;
    meter->samplingrate = 1;
}

void XlabInterface::dataChannelMeterGetSetitings(channela_t *meter, uint8_t *data)
{
    meter->control = *(uint8_t *)(data);
    meter->gain = *(uint8_t *)(data + 1);
    meter->offset = *(int8_t *)(data + 2);
    meter->samplingrate = *(uint8_t *)(data + 17);
}

void XlabInterface::dataSnifferHeaderGetData(sniffconfig_t *head, uint8_t *data)
{
    head->type = *(uint8_t *)(data);
    head->id = *(uint16_t *)(data + 1);
    head->protocol = *(uint8_t *)(data + 3);
    head->indexrx = *(uint16_t *)(data + 4);
    head->indextx = *(uint16_t *)(data + 6);
    head->bitrate = *(uint16_t *)(data + 8);
    head->option = *(uint8_t *)(data + 9);
}

void XlabInterface::dataShow()
{
    if (mode == MAIN_MODE_Sniff) sniffDataShow();
    else if (mode == MAIN_MODE_Scope) scopeDataShow();
}

void XlabInterface::scopeDataShow()
{
    int databuffersize = 0;
    bool dataavailable = false;
    frameconfig_t dataFrameHead;
    bool ch1trace = false, ch2trace = false, chdtrace = false;
    int i;

    if (!usbDevice.dataAvailable && !m_repaint)
        return;
    if (usbDevice.dataAvailable) {
        dataavailable = true;
        if (ui->checkBoxCH1Trace->isChecked()) databuffersize += DATA_BUFFER_BLOCK_SIZE * (ui->comboBoxADCBuffer->currentIndex() + 1);
        if (ui->checkBoxCH2Trace->isChecked()) databuffersize += DATA_BUFFER_BLOCK_SIZE * (ui->comboBoxADCBuffer->currentIndex() + 1);
        if (ui->checkBoxCHDTrace->isChecked()) databuffersize += DATA_BUFFER_BLOCK_SIZE * (ui->comboBoxADCBuffer->currentIndex() + 1);
        if (usbDevice.dataLength != (FRAME_HEADER_LENGTH + databuffersize + 2 * DATA_CHMETER_RESULT_LENGTH + DATA_CHDCONTROL_LENGTH)) {
            dataavailable = false;
            if (!m_repaint) return;
        }
        if (dataavailable) {
            dataScopeHeaderGetData(&dataFrameHead, usbDevice.chData);
            if (dataFrameHead.type != FRAME_HEADER_NAME_Scope) {
                dataavailable = false;
                logToFile("Scope data: Invalide header", LOGGING_LEVEL_Warning);
                if (!m_repaint) return;
            }
        }
        if (dataavailable) {
            for (i = 0; i < usbDevice.dataLength; i++) {
                *(datacache + i) = *(usbDevice.chData + i);
            }
        }
    }
    usbDevice.dataAvailable = false;
    m_repaint = false;
    dataScopeHeaderGetData(&dataFrameHead, datacache);
    if (dataFrameHead.chenable & (uint8_t)(1 << MAIN_CHANNEL_CH1)) ch1trace = true;
    if (dataFrameHead.chenable & (uint8_t)(1 << MAIN_CHANNEL_CH2)) ch2trace = true;
    if (dataFrameHead.chenable & (uint8_t)(1 << MAIN_CHANNEL_CHD)) chdtrace = true;
    if (bufferSize != dataFrameHead.chlength) {
        scopeSetSizeX(dataFrameHead.chlength / DATA_BUFFER_BLOCK_SIZE);
    }
    databuffersize = 0;
    if (ch1trace) databuffersize += bufferSize;
    if (ch2trace) databuffersize += bufferSize;
    if (chdtrace) databuffersize += bufferSize;
    showCH1ZeroVerticalIndicator(ch1trace);
    showCH2ZeroVerticalIndicator(ch2trace);
    showCHDZeroVerticalIndicator(chdtrace);
    hCursorSetChannelEnabled(MAIN_CHANNEL_CH1, ch1trace);
    hCursorSetChannelEnabled(MAIN_CHANNEL_CH2, ch2trace);
    hCursorSetChannelEnabled(MAIN_CHANNEL_CHD, chdtrace);
    ui->timeLabel->setText("Time = " + rateText[dataFrameHead.sapmlingrate]);

    QVector<double> key;
    double ch1, ch2;
    double minCh1 = 0, maxCh1 = 0, minCh2 = 0, maxCh2 = 0, ch1Int, ch2Int;
    QVector<double> ch1Buffer, ch2Buffer, ch1RefBuff, ch2RefBuff;
    QVector<double> bit[8], bit0[8], bitRefBuff[8];
    // QVector<double> fft1, fft2;
    // complex pSignal1[bufferSize], pSignal2[bufferSize];
    int step = 2, startpool;
    channela_t MMCH1buffer, MMCH2buffer;
    mmzeros_t idx;
    dataInitChannelMeterBuffer(&MMCH1buffer);
    dataInitChannelMeterBuffer(&MMCH2buffer);

    // Load channel data
    if ((ui->rollMode->isChecked()) && (ui->rollMode->isEnabled())) {
        i = (uint16_t)dataFrameHead.index;
    } else {
        i = 0;
    }
    for (int xtime = 0; xtime < step * bufferSize; xtime += step, i++) {
        if (i > (bufferSize - 1)) i = 0;
        if (ui->xyMode->isChecked()) {
            key.push_back(255 + (int8_t)datacache[FRAME_HEADER_LENGTH + i]);
        } else {
            key.push_back(xtime);
        }
        startpool = 0;
        // CH1
        if (ch1trace) {
            ch1Int = (int8_t)datacache[FRAME_HEADER_LENGTH + i];
            ch1 = ui->ch1PositionSlider->value() + 2 * ch1Int * GAIN_FACTOR + yAxisMax / 2;
            ch1Buffer.push_back(ch1);
            startpool += bufferSize;
            // Min, Max, VDC calculate
            if (xtime == 0) {
                MMCH1buffer.min = ch1Int;
                MMCH1buffer.max = ch1Int;
                minCh1 = ch1;
                maxCh1 = ch1;
            }
            if (MMCH1buffer.min > ch1Int) MMCH1buffer.min = ch1Int;
            else if (MMCH1buffer.max < ch1Int) MMCH1buffer.max = ch1Int;
            if (minCh1 > ch1) minCh1 = ch1;
            else if (maxCh1 < ch1) maxCh1 = ch1;
        }
        // CH2
        if (ch2trace) {
            ch2Int = (int8_t)datacache[FRAME_HEADER_LENGTH + i + startpool];
            ch2 = ui->ch2PositionSlider->value() + 2 * ch2Int * GAIN_FACTOR + yAxisMax / 2;
            ch2Buffer.push_back(ch2);
            startpool += bufferSize;
            // Min, Max, VDC calculate
            if (xtime == 0) {
                MMCH2buffer.min = ch2Int;
                MMCH2buffer.max = ch2Int;
                minCh2 = ch2;
                maxCh2 = ch2;
            }
            if (MMCH2buffer.min > ch2Int) MMCH2buffer.min = ch2Int;
            else if (MMCH2buffer.max < ch2Int) MMCH2buffer.max = ch2Int;
            if (minCh2 > ch2) minCh2 = ch2;
            else if (maxCh2 < ch2) maxCh2 = ch2;
        }
        // LOGIC
        if (chdtrace) {
            int pos0;
            for (int m = 0; m < 8; m++) {
                byte data = datacache[FRAME_HEADER_LENGTH + i + startpool];
                pos0 = logicBitPos0(m);
                if ((data & (byte)(1 << m)) != 0) {
                    bit[m].push_back(pos0 + 10 + ui->chdSizeSlider->value());
                } else {
                    bit[m].push_back(pos0);
                }
                bit0[m].push_back(pos0 - 1);
            }
        }
    }
    // Get meter from scope data
    if (ui->checkBoxmeterondev->isChecked()) {
        dataChannelMeterGetData(&MMCH1buffer, datacache + FRAME_HEADER_LENGTH + databuffersize);
        dataChannelMeterGetData(&MMCH2buffer, datacache + FRAME_HEADER_LENGTH + databuffersize + DATA_CHMETER_RESULT_LENGTH);
    } else {
        if (ch1trace) {
            MMCH1buffer.frequency = scopeGetMeasureFrequency((int8_t *)(datacache + FRAME_HEADER_LENGTH), bufferSize, MMCH1buffer.min, MMCH1buffer.max, &idx);
            MMCH1buffer.cross1 = idx.first;
            MMCH1buffer.cross2 = idx.second;
            scopeGetAverRMS((int8_t *)(datacache + FRAME_HEADER_LENGTH), &idx, &MMCH1buffer.aver, &MMCH1buffer.rms);
            dataChannelMeterGetSetitings(&MMCH1buffer, datacache + FRAME_HEADER_LENGTH + databuffersize);
        }
        if (ch2trace) {
            MMCH2buffer.frequency = scopeGetMeasureFrequency((int8_t *)(datacache + FRAME_HEADER_LENGTH + bufferSize), bufferSize, MMCH2buffer.min, MMCH2buffer.max, &idx);
            MMCH2buffer.cross1 = idx.first;
            MMCH2buffer.cross2 = idx.second;
            scopeGetAverRMS((int8_t *)(datacache + FRAME_HEADER_LENGTH + bufferSize), &idx, &MMCH2buffer.aver, &MMCH2buffer.rms);
            dataChannelMeterGetSetitings(&MMCH2buffer, datacache + FRAME_HEADER_LENGTH + databuffersize + DATA_CHMETER_RESULT_LENGTH);
        }
    }
    // Show meter
    if (ch1trace) {
        ui->multimeterdisplayVppCh1->setValue(int8ToVolt_f(MMCH1buffer.gain, MMCH1buffer.max, MMCH1buffer.min));
        ui->multimeterdisplayMaxCh1->setValue(int8ToVolt_f(MMCH1buffer.gain, MMCH1buffer.max));
        ui->multimeterdisplayMinCh1->setValue(int8ToVolt_f(MMCH1buffer.gain, MMCH1buffer.min));
        ui->multimeterdisplayRMSCh1->setValue(int16ToVolt_f(MMCH1buffer.gain, MMCH1buffer.rms));
        ui->multimeterdisplayAverCh1->setValue(int16ToVolt_f(MMCH1buffer.gain, MMCH1buffer.aver));
        ui->multimeterdisplayFreqCh1->setValue(samplex100ToFrequency_f(MMCH1buffer.frequency, MMCH1buffer.samplingrate));
    }
    if (ch2trace) {
        ui->multimeterdisplayVppCh2->setValue(int8ToVolt_f(MMCH2buffer.gain, MMCH2buffer.max, MMCH2buffer.min));
        ui->multimeterdisplayMaxCh2->setValue(int8ToVolt_f(MMCH2buffer.gain, MMCH2buffer.max));
        ui->multimeterdisplayMinCh2->setValue(int8ToVolt_f(MMCH2buffer.gain, MMCH2buffer.min));
        ui->multimeterdisplayRMSCh2->setValue(int16ToVolt_f(MMCH2buffer.gain, MMCH2buffer.rms));
        ui->multimeterdisplayAverCh2->setValue(int16ToVolt_f(MMCH2buffer.gain, MMCH2buffer.aver));
        ui->multimeterdisplayFreqCh2->setValue(samplex100ToFrequency_f(MMCH2buffer.frequency, MMCH2buffer.samplingrate));
    }
    // Set vertical cursor
    if (ui->comboBoxScopeCursorsHorizontalType->currentIndex() == 1) {
        uint16_t cross1, cross2;
        bool setauto = false;
        if ((ui->comboBoxScopeCursorsChannel->currentIndex() == 0) && ch1trace && (MMCH1buffer.frequency > 0)) {
            cross1 = MMCH1buffer.cross1;
            cross2 = MMCH1buffer.cross2;
            setauto = true;
        } else if ((ui->comboBoxScopeCursorsChannel->currentIndex() == 1) && ch2trace && (MMCH2buffer.frequency > 0)) {
            cross1 = MMCH2buffer.cross1;
            cross2 = MMCH2buffer.cross2;
            setauto = true;
        }
        if ((ui->rollMode->isChecked()) && (ui->rollMode->isEnabled()) && setauto) {
            cross1 = bufferSize - i + cross1;
            cross2 = bufferSize - i + cross2;
        }
        if (setauto) vCursorSetPosition(2 * cross1, 2 * cross2);
    }

    // Cursors
    if (ui->xyMode->isChecked()) vCursorSetVisible(false);

    hCursorcSetPositionAuto(ui->checkBoxScopeCursorsHorizontal->isChecked(), maxCh1, minCh1, maxCh2, minCh2);
    hCursorRefreshLabel(ui->checkBoxScopeCursorsHorizontal->isChecked(), MMCH1buffer.gain, MMCH2buffer.gain);
    vCursor.samplingrate = dataValueRangeTrunk(dataFrameHead.sapmlingrate, 0, 21);
    vCursorRefresh();
    zeroHeadRefresh();
    updateCh1Label(MMCH1buffer.control, MMCH2buffer.control);
    updateCh2Label(MMCH2buffer.control, MMCH1buffer.control);

    // Decay
    if (ui->decaySlider->value() > 0) {
        ch1PBuffer.push_back(ch1Buffer);
        while (ch1PBuffer.size() > ui->decaySlider->value())
            ch1PBuffer.pop_front();
        ch2PBuffer.push_back(ch2Buffer);
        while (ch2PBuffer.size() > ui->decaySlider->value())
            ch2PBuffer.pop_front();
        for (int k = ui->decaySlider->value(); k < TG; k++) {
            ch1PGraphs[k]->data().data()->clear();
            ch2PGraphs[k]->data().data()->clear();
        }
    } else {
        ch1PGraphs[0]->data().data()->clear();
        ch2PGraphs[0]->data().data()->clear();
        ch1PBuffer.clear();
        ch2PBuffer.clear();
    }

    // Captur buffer
    if (captureRef) {
        captureRef = false;
        ui->ch1CaptureSlider->setMaximum(yAxisMax - rangeMax / 4 - ch1ZeroPos);
        ui->ch1CaptureSlider->setMinimum(rangeMax / 4 - ch1ZeroPos);
        ui->ch1CaptureSlider->setValue(0);
        ui->ch2CaptureSlider->setMaximum(yAxisMax - rangeMax / 4 - ch2ZeroPos);
        ui->ch2CaptureSlider->setMinimum(rangeMax / 4 - ch2ZeroPos);
        ui->ch2CaptureSlider->setValue(0);
        ui->chdCaptureSlider->setMaximum(yAxisMax - rangeMax / 4 - 120 - chdZeroPos);
        ui->chdCaptureSlider->setMinimum(rangeMax / 4 - 120 - chdZeroPos);
        ui->chdCaptureSlider->setValue(0);
        ch1RefBuffer.clear();
        ch1RefBuffer = ch1Buffer;
        ch2RefBuffer.clear();
        ch2RefBuffer = ch2Buffer;
        for (int k = 0; k < 8; k++) {
            bitRef[k].clear();
            bitRef[k] = bit[k];
        }
    }
    // Show capture
    if (ui->refCH1->isChecked() && !ch1RefBuffer.isEmpty()) {
        for (int k = 0; k < ch1RefBuffer.size(); k++) {
            ch1RefBuff.push_back(ch1RefBuffer[k] + ui->ch1CaptureSlider->value());
        }
    }
    if (ui->refCH2->isChecked() && !ch2RefBuffer.isEmpty()) {
        for (int k = 0; k < ch2RefBuffer.size(); k++) {
            ch2RefBuff.push_back(ch2RefBuffer[k]+ui->ch2CaptureSlider->value());
        }
    }
    if (ui->refLogic->isChecked()) {
        for (int k = 0; k < 8; k++) {
            if (bitRef[k].isEmpty()) {
                continue;
            } else {
                for (int m = 0; m < bitRef[k].size(); m++) {
                    bitRefBuff[k].push_back(bitRef[k].at(m) + ui->chdCaptureSlider->value());
                }
            }
        }
    }
    // Show saved wave
    if (displayLoadedWave) {
        ch1Buffer.clear();
        ch1Buffer = ch1SaveBuffer;
        ch2Buffer.clear();
        ch2Buffer = ch2SaveBuffer;
        for (int k = 0; k < 8; k++) {
            bit[k].clear();
            bit[k] = bitSaveBuffer[k];
            bit0[k].clear();
        }
    }
    // Load recorded wave
    if (loadRecordedWave) {
        if (loadedWave.count() >= 10) {
            usbDevice.controlWriteTransfer(USB_SEND_CONFIGURATION, MAIN_CONFIGURATION_SamplingRate,(byte)loadedFrequency.at(0));
            ui->samplingSlider->setValue(loadedFrequency.at(0));
            usbDevice.serial.setSamplingValue(loadedFrequency.at(0));
            loadedFrequency.removeFirst();
            ch1Buffer.clear();
            ch1Buffer = loadedWave.at(0);
            loadedWave.removeFirst();
            ch2Buffer.clear();
            ch2Buffer = loadedWave.at(0);
            loadedWave.removeFirst();
            for (int k = 0; k < 8; k++) {
                bit[k].clear();
                bit[k] = loadedWave.at(0);
                bit0[k].clear();
                loadedWave.removeFirst();
            }
        } else {
            loadRecordedWave = false;
            ui->samplingSlider->setValue(freqBeforeLoaded);
            usbDevice.serial.setSamplingValue(freqBeforeLoaded);
            ui->label_13->setText("");
        }
    }
    // Save wave
    if (saveWave)     {
        saveWave = false;
        ch1SaveBuffer.clear();
        ch1SaveBuffer = ch1Buffer;
        ch2SaveBuffer.clear();
        ch2SaveBuffer = ch2Buffer;
        for (int k = 0; k < 8; k++) {
            bitSaveBuffer[k].clear();
            bitSaveBuffer[k] = bit[k];
        }
        saveWavetoFile();
    }

    // Widget presentation
    if (ui->xyMode->isChecked()) {
        ch1Graph->data().data()->clear();
        ch2Graph->data().data()->clear();
        for (int k = 0; k < TG; k++) {
            ch1PGraphs[k]->data().data()->clear();
            ch2PGraphs[k]->data().data()->clear();
        }
        ch1Graph->setData(key, ch2Buffer);
        ch1Graph->rescaleValueAxis(true);
    } else {
        if (!ui->checkBoxPersistence->isChecked()) {
            ch1Graph->data().data()->clear();
            for (int k = 0; k < TG; k++) {
                ch1PGraphs[k]->data().data()->clear();
            }
        }
        if (ch1trace) {
            if (!ui->checkBoxPersistence->isChecked()) {
                ch1Graph->setData(key, ch1Buffer);
                for (int k = 0; k < qMin(ch1PBuffer.size(), ui->decaySlider->value() + 1); k++) {
                    ch1PGraphs[k]->setData(key, ch1PBuffer[k]);
                }
            } else ch1Graph->addData(key, ch1Buffer);
        }
        if (ui->refCH1->isChecked()) ch1RefGraph->setData(key, ch1RefBuff);
        else ch1RefGraph->data().data()->clear();
        if (!ui->checkBoxPersistence->isChecked()) {
            ch2Graph->data().data()->clear();
            for (int k = 0; k < TG; k++) {
                ch2PGraphs[k]->data().data()->clear();
            }
        }
        if (ch2trace) {
            if (!ui->checkBoxPersistence->isChecked()) {
                ch2Graph->setData(key, ch2Buffer);
                for (int k = 0; k < qMin(ch2PBuffer.size(), ui->decaySlider->value() + 1); k++) {
                    ch2PGraphs[k]->setData(key, ch2PBuffer[k]);
                }
            } else ch2Graph->addData(key, ch2Buffer);
        }
        if ( ui->refCH2->isChecked()) ch2RefGraph->setData(key, ch2RefBuff);
        else ch2RefGraph->data().data()->clear();
        logicLabelsRefresh();
        for (int k = 0; k < 8; k++) {
            if (!ui->checkBoxPersistence->isChecked()) {
                chdGraph[k]->data().data()->clear();
                chd0Graph[k]->data().data()->clear();
            }
            textLabelBit[k]->setVisible(false);
            if (chdtrace && bitChecked[k]) {
                if ( !ui->checkBoxPersistence->isChecked()) {
                    chdGraph[k]->setData(key, bit[k]);
                    chd0Graph[k]->setData(key, bit0[k]);
                } else {
                    chdGraph[k]->addData(key, bit[k]);
                    chd0Graph[k]->addData(key, bit0[k]);
                }
                textLabelBit[k]->setVisible(true);
            }
            if (ui->refLogic->isChecked()&&bitChecked[k]) {
                chdRefGraph[k]->setData(key, bitRefBuff[k]);
                textLabelBit[k]->setVisible(true);
            } else chdRefGraph[k]->data().data()->clear();
            if (!ui->refLogic->isChecked()) chdRefGraph[k]->data().data()->clear();
        }
        if (!ui->checkBoxPersistence->isChecked()) {
            ch1BarGraph->data().data()->clear();
            ch2BarGraph->data().data()->clear();
        }
        // FFT
    /*    if (ui->checkBoxFFTTrace->isChecked()) {
            double magnitude, magnitude2;
            double real1, real2, imag;
            for (int j = 0; j < bufferSize; j++ ) {
                if (!ui->checkBoxIQFFT->isChecked()) {
                    real1 = (double)datacache[FRAME_HEADER_LENGTH + j] * fftWindow[j];
                    real2 = (double)datacache[FRAME_HEADER_LENGTH + j + bufferSize] * fftWindow[j];
                    pSignal1[j] = complex(real1, real1);
                    pSignal2[j] = complex(real2, real2);
                } else {
                    real1 = (double)datacache[FRAME_HEADER_LENGTH + j] * fftWindow[j];
                    imag = (double)datacache[FRAME_HEADER_LENGTH + j + bufferSize] * fftWindow[j];
                    pSignal1[j] = complex(real1, imag);
                }
            }
            if (!ui->checkBoxIQFFT->isChecked()) {
                if (ui->checkBoxFFTCH1->isChecked()) CFFT::Forward(pSignal1, 256);
                if (ui->checkBoxFFTCH2->isChecked()) CFFT::Forward(pSignal2, 256);
            } else CFFT::Forward(pSignal1, 256);
            for (int k = 0; k < 128; k++, i++) {
                if (i > 255) i = 0;
                if (!ui->checkBoxIQFFT->isChecked()) {
                    if (ui->checkBoxFFTCH1->isChecked())
                        magnitude = sqrt(pSignal1[k].re() * pSignal1[k].re() + pSignal1[k].im() * pSignal1[k].im());
                    if (ui->checkBoxFFTCH2->isChecked())
                        magnitude2 = sqrt(pSignal2[k].re() * pSignal2[k].re() + pSignal2[k].im() * pSignal2[k].im());
                    if (ui->checkBoxLogY->isChecked()) {
                        magnitude = 16 * log(magnitude) / log(2.0);
                        magnitude2 = 16 * log(magnitude2) / log(2.0);
                    } else {
                        magnitude = magnitude * 0.05;
                        magnitude2 = magnitude2 * 0.05;
                    }
                    if (ui->checkBoxFFTCH1->isChecked()) fft1.push_back(magnitude);
                    if (ui->checkBoxFFTCH2->isChecked()) fft2.push_back(magnitude2);
                } else {
                    magnitude = sqrt(pSignal1[i].re() * pSignal1[i].re() + pSignal1[i].im() * pSignal1[i].im());
                    if (ui->checkBoxLogY->isChecked()) magnitude = 16 * log(magnitude) / log(2.0);
                    else magnitude = magnitude * 0.05;
                    fft1.push_back(magnitude);
                }
            }
            if (!ui->checkBoxPersistence->isChecked()) {
                if (!ui->checkBoxIQFFT->isChecked()) {
                    ch1BarGraph->data().data()->clear();
                    if (ui->checkBoxFFTCH1->isChecked()) ch1BarGraph->setData(key, fft1);
                    ch2BarGraph->data().data()->clear();
                    if (ui->checkBoxFFTCH2->isChecked()) ch2BarGraph->setData(key, fft2);
                } else {
                    ch1BarGraph->data().data()->clear();
                    ch1BarGraph->setData(key, fft1);
                }
            } else {
                if (!ui->checkBoxIQFFT->isChecked()) {
                    if (ui->checkBoxFFTCH1->isChecked()) ch1BarGraph->addData(key, fft1);
                    if (ui->checkBoxFFTCH2->isChecked()) ch2BarGraph->addData(key, fft2);
                } else {
                    ch1BarGraph->addData(key, fft1);
                }
            }
        }  */
    }
    // Recording & logging
    QString tmp_wave;
    if (lastFrameScope != dataFrameHead.id) {
        lastFrameScope = dataFrameHead.id;
        if (logging || recordingWaves) {
            ch1SaveBuffer.clear();
            ch1SaveBuffer = ch1Buffer;
            ch2SaveBuffer.clear();
            ch2SaveBuffer = ch2Buffer;
            for (int k = 0; k < 8; k++) {
                bitSaveBuffer[k].clear();
                bitSaveBuffer[k] = bit[k];
            }
            tmp_wave = saveWavetoString();
        }
        if (recordingWaves) {
            dataToSave += QString::number(ui->samplingSlider->value()) + "<*>\n";
            dataToSave += tmp_wave;
        }
    }
    // Finish
    ui->plotterWidget->replot();
}

void XlabInterface::scopeInitDatacache()
{
    byte field = 0;

    *(uint8_t *)(datacache) = FRAME_HEADER_NAME_Scope;
    *(uint16_t *)(datacache + 1) = 0;
    if (ui->checkBoxCH1Trace->isChecked()) field += (1 << MAIN_CHANNEL_CH1);
    if (ui->checkBoxCH2Trace->isChecked()) field += (1 << MAIN_CHANNEL_CH2);
    if (ui->checkBoxCHDTrace->isChecked()) field += (1 << MAIN_CHANNEL_CHD);
    *(uint8_t *)(datacache + 3) = field;
    field = 0;
    if (ui->checkBoxCH1Trace->isChecked()) field += (1 << MAIN_CHANNEL_CH1);
    if (ui->checkBoxCH2Trace->isChecked()) field += (1 << MAIN_CHANNEL_CH2);
    *(uint8_t *)(datacache + 4) = field;
    *(uint16_t *)(datacache + 5) = DATA_BUFFER_BLOCK_SIZE * (ui->comboBoxADCBuffer->currentIndex() + 1);
    *(uint8_t *)(datacache + 7) = ui->samplingSlider->value();
    *(uint16_t *)(datacache + 8) = 0;
}

// Work on USB buffer channel data, return frequency as count of samples for 100 periods
uint32_t XlabInterface::scopeGetMeasureFrequency(int8_t* buffer, const uint16_t length, int16_t min, int16_t max, mmzeros_t* cross)
{
    #define BORDER	5
    uint8_t step = 0, crosses = 0;
    int8_t half;
    uint16_t i, idx[3], periods = 0;
    uint32_t frequency = 0;
    bool dirup = false;

    cross->first = 0;
    cross->second = 0;
    cross->last = length;
    if ((max - min) < 32) return 0;
    half = (int8_t)((min + max) / 2);
    min = (min * 10 + max * 6) / 16;
    max = (min * 6 + max * 10) / 16;
    for (i = BORDER; i < (length - BORDER); i++) {
        switch (step) {
            case 0:
                if (buffer[i] < min) {
                    step = 1;
                    dirup = true;
                } else if (buffer[i] > max) {
                    step = 1;
                    dirup = false;
                }
                break;
            case 1:
                if ((dirup && (buffer[i] > min)) || (!dirup && (buffer[i] < max))) {
                    step = 2;
                }
            case 2:
                if ((dirup && (buffer[i] > half)) || (!dirup && (buffer[i] < half))) {
                    step = 3;
                    idx[crosses] = i;
                    if (crosses < 2) crosses++;
                    periods++;
                }
                break;
            case 3:
                if ((dirup && (buffer[i] < min)) || (!dirup && (buffer[i] > max))) {
                    step = 1;
                }
                break;
        }
        if (periods > (length >> 2)) return 0;
    }
    periods--;
    if (periods < 1) return 0;
    else if (periods == 1) idx[2] = idx[1];
    cross->first = idx[0];
    cross->second = idx[1];
    cross->last = idx[2];
    frequency = ((100 * (int32_t)(idx[2] - idx[0])) / periods);
    return frequency;
}

void XlabInterface::scopeGetAverRMS(int8_t *buffer, mmzeros_t *cross, int16_t *aver, int16_t *rms)
{
    int32_t tmpaverage = 0, tmprms = 0;

    for (int i = cross->first; i < cross->last; i++) {
        tmprms += buffer[i] * buffer[i];
        tmpaverage += (buffer[i]);
    }
    // 128 measurements with a resolution of 8 bit
    *rms = (int16_t)(128 * sqrt(tmprms / (cross->last - cross->first)));
    // 128 measurements with a resolution of 8 bit
    *aver = (int16_t)(128 * tmpaverage / (cross->last - cross->first));
}

QString XlabInterface::get4Digits(qreal number)
{
    QString space = " ";
    int precision = 3;

    if (number >= 1000 || number <= -1000) precision = 0;
    else if (number >= 100  || number <= -100)  precision = 1;
    else if (number >= 10   || number <= -10)   precision = 2;
    if (number < 0) space = "";
    return space + QString::number(number, 'f', precision);
}

QString XlabInterface::get3Digits(qreal number)
{
    QString space = " ";
    int precision = 2;

    if (number >= 100 || number <= -100) precision = 0;
    else if (number >= 10  || number <= -10)  precision = 1;
    if (number < 0) space = "";
    return space + QString::number(number, 'f', precision);
}

QString XlabInterface::removeTrailingZeros(QString str)
{
    int n = str.size(), j = 0;

    for (int i = 0; i < (n - 1); i++) {
        if ((QString(str.at(i)) != "0") || (QString(str.at(i + 1)) == ".") || (QString(str.at(i + 1)) == ",")) {
            j = i;
            break;
        }
    }
    return str.right(n - j);
}

QString XlabInterface::getDigitsPrecision(qreal number, int prec)
{
    QString space = " ";
    if (number < 0) space = "";
    return space + QString::number(number, 'f', prec);
}

int XlabInterface::dataValueRangeTrunk(int value, int min, int max) {
    int oldvalue = value;
    if (oldvalue > max) oldvalue = max;
    else if (oldvalue < min) oldvalue = min;
    return oldvalue;
}

int XlabInterface::dataValueRangeTrunk64(qint64 value, qint64 min, qint64 max)
{
    qint64 oldvalue = value;
    if (oldvalue > max) oldvalue = max;
    else if (oldvalue < min) oldvalue = min;
    return oldvalue;
}

int XlabInterface::mathRoundTo(double value, int grad)
{
    return round(value / grad) * grad;
}

QString XlabInterface::getStringFromUnsignedChar( unsigned char *buffer, int length )
{
    QString result;

    for ( int i = 0; i < length; i++ ) {
        if (buffer[i] >= 0x20) result.append(buffer[i]);
    }
    return result.trimmed();
}

int XlabInterface::mapRange(int value, int oldMax, int oldMin, int newMax, int newMin)
{
    int newRange, oldRange;
    newRange = newMax - newMin;
    oldRange = oldMax - oldMin;
    return newMax - ((float)((value - oldMin) * newRange) / oldRange);
}

double XlabInterface::mapRangeF(double value, double oldMax, double oldMin, double newMax, double newMin)
{
    double newRange, oldRange;
    newRange = newMax - newMin;
    oldRange = oldMax - oldMin;
    return newMax - (((value - oldMin) * newRange) / oldRange);
}

QString XlabInterface::pixelValueToVolt(double value, QString unit, int precision)
{
    double dunit = 0;

    for (int i = 0; i < unit.length(); i++) {
        if (unit[i] == 'm' || unit[i] == 'V') {
            dunit = unit.left(i).toDouble();
            unit = unit.remove(unit.left(i));
            break;
        }
    }
    unit = unit.replace("/div", "");
    if (precision <= 3) return get3Digits(value * dunit / 64.0) + " " + unit;
    if (precision >= 4) return get4Digits(value * dunit / 64.0) + " " + unit;
    return "? " + unit;
}

double XlabInterface::rawValueToVolt(double value, QString unit)
{
    double dunit = 0;

    for (int i = 0; i < unit.length(); i++) {
        if (unit[i] == 'V') {
            dunit = unit.left(i).toDouble();
            break;
        }
        else if (unit[i] == 'm') {
            dunit = unit.left(i).toDouble() / 1000.0;
            break;
        }
    }
    return (value * dunit / 64.0);
}

void XlabInterface::vCursorInit()
{
    vCursor.AposX = 64;
    vCursor.BposX = 192;
    vCursor.visible = false;
    vCursor.samplingrate = 1;
}

void XlabInterface::vCursorSetRange(int min, int max)
{
    min = dataValueRangeTrunk(min, 0, xAxisMax - 1);
    vCursor.posXmin = min;
    max = dataValueRangeTrunk(max, min + 1, xAxisMax);
    vCursor.posXmax = max;
}

void XlabInterface::vCursorSetPosition(int a, int b)
{
    vCursor.AposX = a;
    vCursor.BposX = b;
}

void XlabInterface::vCursorRefreshLabel()
{
    // Vertical cursor & delta time
    if (vCursor.visible) {
        double deltaTime, freq;
        QString unit;

        deltaTime = vCursorDeltaTime(vCursor.AposX, vCursor.BposX, &unit, 1, vCursor.samplingrate);
        textLabelDeltaTime->setText(("ΔT = " + get4Digits(deltaTime) + " " + unit));
        freq = vCursorFrequency(deltaTime, &unit, vCursor.samplingrate);
        textLabelFrequency->setText(("1/ΔT = " + get4Digits(freq) + " " + unit));
    }
}

void XlabInterface::vCursorSetLabelVisible(bool enabled)
{
    textLabelDeltaTime->setVisible(enabled);
    textLabelFrequency->setVisible(enabled);
}

void XlabInterface::vCursorRefresh()
{
    vCursorRefreshA();
    vCursorRefreshB();
    ch1vZeroRefresh();
    vCursorRefreshLabel();
}

void XlabInterface::vCursorRefreshA()
{
    if (vCursor.visible) {
        vCursor.AposX = dataValueRangeTrunk(vCursor.AposX, vCursor.posXmin, vCursor.posXmax);
        vCursorAHead->topLeft->setCoords(ui->plotterWidget->xAxis->coordToPixel(vCursor.AposX) - 9, 12);
    }
}

void XlabInterface::vCursorRefreshB()
{
    if (vCursor.visible) {
        vCursor.BposX = dataValueRangeTrunk(vCursor.BposX, vCursor.posXmin, vCursor.posXmax);
        vCursorBHead->topLeft->setCoords(ui->plotterWidget->xAxis->coordToPixel(vCursor.BposX) - 9, 12);
    }
}

void XlabInterface::ch1vZeroRefresh()
{
    if (ui->xyMode->isChecked()) {
        ch1vZeroHead->topLeft->setCoords(ui->plotterWidget->xAxis->coordToPixel(256) - 9, 1);
    }
}

void XlabInterface::vCursorSetVisible(bool enabled) {
    vCursor.visible = enabled;
    vCursorAHead->setVisible(enabled);
    vCursorBHead->setVisible(enabled);
    vCursorA->setVisible(enabled);
    vCursorB->setVisible(enabled);
    vCursorSetLabelVisible(vCursor.visible);
}

void XlabInterface::vCursorSetLabelColor(QColor color)
{
    vCursor.color = color;
    textLabelDeltaTime->setColor(color);
    textLabelFrequency->setColor(color);
}

double XlabInterface::vCursorDeltaTime(int apos, int bpos, QString *unit, int div, int samplingrate)
{
    QString u;
    double width, result, value = 0;

    samplingrate = dataValueRangeTrunk(samplingrate, 0, SAMPLING_RATE_RangeOut - 1);
    u = (rateText[samplingrate]);
    for (int i = 0; i < u.length(); i++) {
        if (u[i] == 'm' || u[i] == 's' || u[i] == QChar(0xB5)) {
            value = u.left(i).toDouble();
            u = u.remove(u.left(i));
            break;
        }
    }
    u = u.replace("/div", "");
    u = u.replace('u', QChar(0xb5));
    *unit = u;
    width = 8 * value / 128.0;
    int a = apos * 128.0 / xmax;
    int b = bpos * 128.0 / xmax;
    result = get4Digits(((b - a) * width) / div).toDouble();
    if (result < 0) result = -result;
    return result;
}

double XlabInterface::vCursorFrequency(double deltatime, QString *unit, int samplingrate)
{
    samplingrate = dataValueRangeTrunk(samplingrate, 0, SAMPLING_RATE_RangeOut - 1);
    if (samplingrate <= 6) {
        *unit = "kHz";
    } else if (samplingrate <= 12) {
        *unit = "Hz";
    } else {
        *unit = "mHz";
    }
    return 1000 / deltatime;
}

double XlabInterface::deltaToFrequency(double deltatime, int samplingrate)
{
    double out = 1;
    samplingrate = dataValueRangeTrunk(samplingrate, 0, SAMPLING_RATE_RangeOut - 1);
    if (samplingrate <= 6) {
        out = 1000000;
    } else if (samplingrate <= 12) {
        out = 1000;
    } else {
        out = 1;
    }
    return out / deltatime;
}

void XlabInterface::hCursorcSetPositionAuto(bool enabled, double minCh1, double maxCh1, double minCh2, double maxCh2)
{
    double aTrack, bTrack;

    if (enabled) {
        if (ui->comboBoxScopeCursorsHorizontalType->currentIndex() == 1) {
            if (ui->comboBoxScopeCursorsChannel->currentIndex() == 0 && hCursor.coord[MAIN_CHANNEL_CH1].enabled) {
                hCursorAPosCh1 = minCh1;
                hCursorBPosCh1 = maxCh1;
                hCursorAHead->topLeft->setPixelPosition(QPointF(14,(ui->plotterWidget->yAxis->coordToPixel(hCursorAPosCh1)) - 8));
                hCursorBHead->topLeft->setPixelPosition(QPointF(14,(ui->plotterWidget->yAxis->coordToPixel(hCursorBPosCh1)) - 8));
                hCursorAPosCh1 -= ch1ZeroPos;
                hCursorBPosCh1 -= ch1ZeroPos;
            } else if (ui->comboBoxScopeCursorsChannel->currentIndex() == 1 && hCursor.coord[MAIN_CHANNEL_CH2].enabled) {
                hCursorAPosCh2 = minCh2;
                hCursorBPosCh2 = maxCh2;
                hCursorAHead->topLeft->setPixelPosition(QPointF(14,(ui->plotterWidget->yAxis->coordToPixel(hCursorAPosCh2)) - 8));
                hCursorBHead->topLeft->setPixelPosition(QPointF(14,(ui->plotterWidget->yAxis->coordToPixel(hCursorBPosCh2)) - 8));
                hCursorAPosCh2 -= ch2ZeroPos;
                hCursorBPosCh2 -= ch2ZeroPos;
            }
        } else if (ui->comboBoxScopeCursorsHorizontalType->currentIndex() == 0) {
            if (ui->comboBoxScopeCursorsChannel->currentIndex() == 0 && hCursor.coord[MAIN_CHANNEL_CH1].enabled) {
                phaseTracerAA->setGraphKey(ui->plotterWidget->xAxis->pixelToCoord(vCursorAHead->bottom->pixelPosition().rx()));
                phaseTracerAB->setGraphKey(ui->plotterWidget->xAxis->pixelToCoord(vCursorBHead->bottom->pixelPosition().rx()));
                phaseTracerAA->setGraph(ch1Graph);
                phaseTracerAB->setGraph(ch1Graph);
                phaseTracerAA->updatePosition();
                phaseTracerAB->updatePosition();
                aTrack = phaseTracerAA->position->value();
                bTrack = phaseTracerAB->position->value();
                hCursorAPosCh1 = aTrack;
                hCursorBPosCh1 = bTrack;
                hCursorAHead->topLeft->setPixelPosition(QPointF(14,(ui->plotterWidget->yAxis->coordToPixel(hCursorAPosCh1)) - 8));
                hCursorBHead->topLeft->setPixelPosition(QPointF(14,(ui->plotterWidget->yAxis->coordToPixel(hCursorBPosCh1)) - 8));
                hCursorAPosCh1 += -ch1ZeroPos;
                hCursorBPosCh1 += -ch1ZeroPos;
            } else if (ui->comboBoxScopeCursorsChannel->currentIndex() == 1 && hCursor.coord[MAIN_CHANNEL_CH2].enabled) {
                phaseTracerBA->setGraphKey(ui->plotterWidget->xAxis->pixelToCoord(vCursorAHead->bottom->pixelPosition().rx()));
                phaseTracerBB->setGraphKey(ui->plotterWidget->xAxis->pixelToCoord(vCursorBHead->bottom->pixelPosition().rx()));
                phaseTracerBA->setGraph(ch2Graph);
                phaseTracerBB->setGraph(ch2Graph);
                phaseTracerBA->updatePosition();
                phaseTracerBB->updatePosition();
                aTrack = phaseTracerBA->position->value();
                bTrack = phaseTracerBB->position->value();
                hCursorAPosCh2 = aTrack;
                hCursorBPosCh2 = bTrack;
                hCursorAHead->topLeft->setPixelPosition(QPointF(14,(ui->plotterWidget->yAxis->coordToPixel(hCursorAPosCh2)) - 8));
                hCursorBHead->topLeft->setPixelPosition(QPointF(14,(ui->plotterWidget->yAxis->coordToPixel(hCursorBPosCh2)) - 8));
                hCursorAPosCh2 += - ch2ZeroPos;
                hCursorBPosCh2 += - ch2ZeroPos;
            }
        }
    }
}

void XlabInterface::hCursorRefreshA()
{
    hCursorAPos = dataValueRangeTrunk(hCursorAPos, 8, yAxisMax - 8);
    if (ui->comboBoxScopeCursorsChannel->currentIndex() == 0) {
        hCursorAPos = dataValueRangeTrunk(hCursorAPos, ch1ZeroPos - CURSOR_DISTANCE_MAX, ch1ZeroPos + CURSOR_DISTANCE_MAX);
        hCursorAPosCh1 = hCursorAPos - ch1ZeroPos;
    } else if (ui->comboBoxScopeCursorsChannel->currentIndex() == 1) {
        hCursorAPos = dataValueRangeTrunk(hCursorAPos, ch2ZeroPos - CURSOR_DISTANCE_MAX, ch2ZeroPos + CURSOR_DISTANCE_MAX);
        hCursorAPosCh2 = hCursorAPos - ch2ZeroPos;
    }
    hCursorAHead->topLeft->setPixelPosition(QPointF(14, (ui->plotterWidget->yAxis->coordToPixel(hCursorAPos)) - 8));
}

void XlabInterface::hCursorRefreshB()
{
    hCursorBPos = dataValueRangeTrunk(hCursorBPos, 8, yAxisMax - 8);
    if (ui->comboBoxScopeCursorsChannel->currentIndex() == 0) {
        hCursorBPos = dataValueRangeTrunk(hCursorBPos, ch1ZeroPos - CURSOR_DISTANCE_MAX, ch1ZeroPos + CURSOR_DISTANCE_MAX);
        hCursorBPosCh1 = hCursorBPos - ch1ZeroPos;
    } else if (ui->comboBoxScopeCursorsChannel->currentIndex() == 1) {
        hCursorBPos = dataValueRangeTrunk(hCursorBPos, ch2ZeroPos - CURSOR_DISTANCE_MAX, ch2ZeroPos + CURSOR_DISTANCE_MAX);
        hCursorBPosCh2 = hCursorBPos - ch2ZeroPos;
    }
    hCursorBHead->topLeft->setPixelPosition(QPointF(14, (ui->plotterWidget->yAxis->coordToPixel(hCursorBPos)) - 8));
}

void XlabInterface::hCursorRefreshLabel(bool enabled, int ch1gain, int ch2gain)
{
    if (enabled) {
        if (ui->comboBoxScopeCursorsChannel->currentIndex() == 0) {
            ch1gain = dataValueRangeTrunk(ch1gain, 0, MAIN_CHANNEL_GAIN_RangeOut - 1);
            textLabelVoltageA->setText("Va = " + pixelValueToVolt(hCursorAPosCh1, gainText[ch1gain], 3));
            textLabelVoltageB->setText("Vb = " + pixelValueToVolt(hCursorBPosCh1, gainText[ch1gain], 3));
            textLabelDeltaVoltage->setText("ΔV = " + pixelValueToVolt(hCursorAPosCh1 - hCursorBPosCh1, gainText[ch1gain], 3));
        } else if (ui->comboBoxScopeCursorsChannel->currentIndex() == 1) {
            ch2gain = dataValueRangeTrunk(ch2gain, 0, MAIN_CHANNEL_GAIN_RangeOut - 1);
            textLabelVoltageA->setText("Va = " + pixelValueToVolt(hCursorAPosCh2, gainText[ch2gain], 3));
            textLabelVoltageB->setText("Vb = " + pixelValueToVolt(hCursorBPosCh2, gainText[ch2gain], 3));
            textLabelDeltaVoltage->setText("ΔV = " + pixelValueToVolt(hCursorAPosCh2 - hCursorBPosCh2, gainText[ch2gain], 3));
        }
    }
}

void XlabInterface::hCursorSetTracerVisible(bool enabled)
{
    if (ui->comboBoxScopeCursorsChannel->currentIndex() == 0) {
        phaseTracerAA->setVisible(enabled);
        phaseTracerAB->setVisible(enabled);
    } else {
        phaseTracerAA->setVisible(false);
        phaseTracerAB->setVisible(false);
    }
    if (ui->comboBoxScopeCursorsChannel->currentIndex() == 1) {
        phaseTracerBA->setVisible(enabled);
        phaseTracerBB->setVisible(enabled);
    } else {
        phaseTracerBA->setVisible(false);
        phaseTracerBB->setVisible(false);
    }
}

void XlabInterface::hCursorInit()
{
    hCursor.visible = false;
    hCursor.maxDistance = CURSOR_DISTANCE_MAX;
    hCursor.channel = MAIN_CHANNEL_CH1;
    hCursor.track = false;
    hCursor.coord[MAIN_CHANNEL_CH1].enabled = false;
    hCursor.coord[MAIN_CHANNEL_CH1].headA = 100;
    hCursor.coord[MAIN_CHANNEL_CH1].headB = -100;
    hCursor.coord[MAIN_CHANNEL_CH2].enabled = false;
    hCursor.coord[MAIN_CHANNEL_CH2].headA = 100;
    hCursor.coord[MAIN_CHANNEL_CH2].headB = -100;
}

void XlabInterface::hCursorSetVisible(bool enabled)
{
    hCursor.visible = enabled;
    hCursorRefreshVisible();
}

void XlabInterface::hCursorSetChannelEnabled(int channel, bool enabled)
{
    hCursor.coord[channel].enabled = enabled;
    hCursorRefreshVisible();
}

void XlabInterface::hCursorRefreshVisible()
{
    if (hCursor.coord[hCursor.channel].enabled) {
        hCursorAHead->setVisible(hCursor.visible);
        hCursorBHead->setVisible(hCursor.visible);
        hCursorA->setVisible(hCursor.visible);
        hCursorB->setVisible(hCursor.visible);
        hCursorSetLabelVisible(hCursor.visible);
    } else {
        hCursorAHead->setVisible(false);
        hCursorBHead->setVisible(false);
        hCursorA->setVisible(false);
        hCursorB->setVisible(false);
        hCursorSetLabelVisible(false);
    }
}

void XlabInterface::hCursorSetLabelVisible(bool enabled)
{
    textLabelDeltaVoltage->setVisible(enabled);
    textLabelVoltageA->setVisible(enabled);
    textLabelVoltageB->setVisible(enabled);
}

void XlabInterface::hCursorSetLabelColor(QColor color)
{
    hCursor.color = color;
    textLabelVoltageA->setColor(color);
    textLabelVoltageB->setColor(color);
    textLabelDeltaVoltage->setColor(color);
}

int XlabInterface::logicBitPos0(int m)
{
    return  m * ( 20 + ui->chdSizeSlider->value()) + ui->chdPositionSlider->value() + yAxisMax / 2;
}

void XlabInterface::logicLabelPositionSet(int idx)
{
    if (idx == 0) logicLabelPosition = ui->plotterWidget->visibleRegion().boundingRect().left() + 46;
    else if (idx == 1) logicLabelPosition = ui->plotterWidget->visibleRegion().boundingRect().right() - 50;
}

void XlabInterface::logicLabelsRefresh()
{
    int pos0;

    logicLabelPositionSet(tomap_ScopeOptions->logicLabelPosition);
    for (int m = 0; m < 8; m++) {
        pos0 = 5 + ui->chdSizeSlider->value() / 6 + logicBitPos0(m);
        textLabelBit[m]->position->setPixelPosition(QPointF(logicLabelPosition, ui->plotterWidget->yAxis->coordToPixel(pos0)));
    }
}

void XlabInterface::zeroHeadRefresh()
{
    ch1ZeroHead->topLeft->setPixelPosition(QPointF(2, ui->plotterWidget->yAxis->coordToPixel(ch1ZeroPos)-8));
    ch2ZeroHead->topLeft->setPixelPosition(QPointF(2, ui->plotterWidget->yAxis->coordToPixel(ch2ZeroPos)-8));
    chDZeroHead->topLeft->setPixelPosition(QPointF(2, ui->plotterWidget->yAxis->coordToPixel(chdZeroPos)-8));
}

void XlabInterface::showCH1ZeroVerticalIndicator(bool show)
{
    if (ui->xyMode->isChecked()) {
        ch1ZeroHead->setVisible(false);
        ch1Zero->setVisible(false);
        ch1vZeroHead->setVisible(show);
        ch1vZero->setVisible(show);
    } else {
        ch1ZeroHead->setVisible(show);
        ch1Zero->setVisible(show);
        ch1vZeroHead->setVisible(false);
        ch1vZero->setVisible(false);
    }
}

void XlabInterface::showCH2ZeroVerticalIndicator(bool show)
{
    ch2ZeroHead->setVisible(show);
    ch2Zero->setVisible(show);
}

void XlabInterface::showCHDZeroVerticalIndicator(bool show)
{
    chDZeroHead->setVisible(show);
    chDZero->setVisible(show);
}

void XlabInterface::moveCursor(QMouseEvent *event)
{
    if (!(event->buttons() & Qt::LeftButton) || currentSelected == isNone)
        return;
    int curPos;
    if (event->type()== QMouseEvent::MouseMove) {
        if (currentSelected == isHCursorAHead) {
            hCursorAPos = ui->plotterWidget->yAxis->pixelToCoord(event->y());
            hCursorRefreshA();
        } else if (currentSelected == isHCursorBHead) {
            hCursorBPos = ui->plotterWidget->yAxis->pixelToCoord(event->y());
            hCursorRefreshB();
        } else if (currentSelected == isVCursorAHead) {
            vCursor.AposX = ui->plotterWidget->xAxis->pixelToCoord(event->x());
            vCursorRefreshA();
        } else if (currentSelected == isVCursorBHead) {
            vCursor.BposX = ui->plotterWidget->xAxis->pixelToCoord(event->x());
            vCursorRefreshB();
        } else if (currentSelected == isCH1Zero) {
            int value, maxp, minp;
            maxp = yAxisMax - (rangeMax / 4);
            minp = rangeMax / 4;
            curPos = event->y();
            curPos = dataValueRangeTrunk(curPos, ui->plotterWidget->yAxis->coordToPixel(maxp), ui->plotterWidget->yAxis->coordToPixel(minp));
            ch1ZeroHead->topLeft->setPixelPosition(QPointF(2, curPos-8));
            value = ui->plotterWidget->yAxis->pixelToCoord(curPos);
            ui->ch1PositionSlider->setValue(-mapRange(value, maxp, minp, ui->ch1PositionSlider->maximum(), ui->ch1PositionSlider->minimum()));
        } else if (currentSelected == isCH2Zero) {
            int value, maxp, minp;
            maxp = yAxisMax - (rangeMax / 4);
            minp = rangeMax / 4;
            curPos = event->y();
            curPos = dataValueRangeTrunk(curPos, ui->plotterWidget->yAxis->coordToPixel(maxp), ui->plotterWidget->yAxis->coordToPixel(minp));
            ch2ZeroHead->topLeft->setPixelPosition(QPointF(2, curPos-8));
            value = ui->plotterWidget->yAxis->pixelToCoord(curPos);
            ui->ch2PositionSlider->setValue(-mapRange(value, maxp, minp, ui->ch2PositionSlider->maximum(), ui->ch2PositionSlider->minimum()));
        } else if (currentSelected == isCHDZero) {
            int value,maxp,minp;
            maxp = yAxisMax - (rangeMax / 4) - 120;
            minp = rangeMax / 4 - 120;
            curPos = event->y();
            curPos = dataValueRangeTrunk(curPos, ui->plotterWidget->yAxis->coordToPixel(maxp), ui->plotterWidget->yAxis->coordToPixel(minp));
            chDZeroHead->topLeft->setPixelPosition(QPointF(2, curPos-8));
            value = ui->plotterWidget->yAxis->pixelToCoord(curPos) - 240;
            ui->chdPositionSlider->setValue(-mapRange(value, maxp, minp, ui->chdPositionSlider->maximum(), ui->chdPositionSlider->minimum()));
        } else if (currentSelected == isTriggerPixmap) {
            triggerSetPositionCoord(ui->plotterWidget->xAxis->pixelToCoord(event->x()), ui->plotterWidget->yAxis->pixelToCoord(event->y()));
            if (!ui->radioButtonFree->isChecked()) {
                if (ui->comboBoxTrigSource->currentIndex() < 2) sendTriggerLevel(trigger.coord[trigger.channel].posY);
                sendTriggerPost(trigger.coord[trigger.channel].posX);
            }
        } else if (currentSelected == isTriggerWin1Pixmap) {
            triggerSetPositionWin1Coord(ui->plotterWidget->xAxis->pixelToCoord(event->x()), ui->plotterWidget->yAxis->pixelToCoord(event->y()));
            if (!ui->radioButtonFree->isChecked()) {
                if (ui->comboBoxTrigSource->currentIndex() < 2) sendTriggerWin1Level(trigger.coord[trigger.channel].posY1);
                sendTriggerPost(trigger.coord[trigger.channel].posX);
            }
        } else if (currentSelected == isTriggerWin2Pixmap) {
            triggerSetPositionWin2Coord(ui->plotterWidget->xAxis->pixelToCoord(event->x()), ui->plotterWidget->yAxis->pixelToCoord(event->y()));
            if (!ui->radioButtonFree->isChecked()) {
                if (ui->comboBoxTrigSource->currentIndex() < 2) sendTriggerWin2Level(trigger.coord[trigger.channel].posY2);
                sendTriggerPost(trigger.coord[trigger.channel].posX);
            }
        }
        m_repaint = true;
    }
}

void XlabInterface::triggerSetEnable(bool enable)
{
    trigger.enable = enable;
}

void XlabInterface::triggerInit()
{
    trigger.enable = false;
    trigger.visible = false;
    trigger.mode = 0;
    trigger.source = 0;
    trigger.channel = 0;
}

void XlabInterface::triggerSetChannelSize(int ch, int size)
{
    trigger.coord[ch].size = size;
}

void XlabInterface::triggerSetChannelBit(int ch, int bit)
{
    bit = dataValueRangeTrunk(bit, 0, 7);
    trigger.coord[ch].bit = bit;
}

void XlabInterface::triggerSetChannelOffset(int ch, int offset)
{
    trigger.coord[ch].offset = offset;
}

void XlabInterface::triggerSetMode(int mode)
{
    mode = dataValueRangeTrunk(mode, 0, 5);
    trigger.mode = mode;
    triggerSetIcon();
}

void XlabInterface::triggerSetPosition(int posx, int posy)
{
    trigger.coord[trigger.channel].posX = dataValueRangeTrunk(posx, trigger.posXmin, trigger.posXmax);
    trigger.coord[trigger.channel].posY = dataValueRangeTrunk(posy, trigger.posYmin, trigger.posYmax);
    triggerMoveTrigger(triggerPosXIntToPixel(trigger.coord[trigger.channel].posX), triggerPosYIntToPixel(trigger.coord[trigger.channel].posY));
}

void XlabInterface::triggerSetPositionCoord(double posx, double posy) {
    triggerSetPosition((int)(posx / 2), triggerPosYPixelToInt(posy));
}

void XlabInterface::triggerSetPositionWin1(int posx, int posy)
{
    trigger.coord[trigger.channel].posX = dataValueRangeTrunk(posx, trigger.posXmin, trigger.posXmax);
    trigger.coord[trigger.channel].posY1 = dataValueRangeTrunk(posy, trigger.posYWmin, trigger.posYWmax - trigger.posYDist);
    if (trigger.coord[trigger.channel].posY1 > (trigger.coord[trigger.channel].posY2 - trigger.posYDist))
        trigger.coord[trigger.channel].posY1 = trigger.coord[trigger.channel].posY2 - trigger.posYDist;
    triggerMoveWinTrigger(triggerPosXIntToPixel(trigger.coord[trigger.channel].posX), triggerPosYIntToPixel(trigger.coord[trigger.channel].posY1), triggerPosYIntToPixel(trigger.coord[trigger.channel].posY2));
}

void XlabInterface::triggerSetPositionWin1Coord(int posx, int posy) {
    triggerSetPositionWin1((int)(posx / 2), triggerPosYPixelToInt(posy));
}

void XlabInterface::triggerSetPositionWin2(int posx, int posy)
{
    trigger.coord[trigger.channel].posX = dataValueRangeTrunk(posx, trigger.posXmin, trigger.posXmax);
    trigger.coord[trigger.channel].posY2 = dataValueRangeTrunk(posy, trigger.posYWmin + trigger.posYDist, trigger.posYWmax);
    if (trigger.coord[trigger.channel].posY2 < (trigger.coord[trigger.channel].posY1 + trigger.posYDist))
        trigger.coord[trigger.channel].posY2 = trigger.coord[trigger.channel].posY1 + trigger.posYDist;
    triggerMoveWinTrigger(triggerPosXIntToPixel(trigger.coord[trigger.channel].posX), triggerPosYIntToPixel(trigger.coord[trigger.channel].posY1), triggerPosYIntToPixel(trigger.coord[trigger.channel].posY2));
}

void XlabInterface::triggerSetPositionWin2Coord(int posx, int posy)
{
    triggerSetPositionWin2((int)(posx / 2), triggerPosYPixelToInt(posy));
}

void XlabInterface::triggerSetChannel(int channel)
{
    channel = dataValueRangeTrunk(channel, 0, 21);
    trigger.channel = channel;
}

void XlabInterface::triggerSetRange(int xmin, int xmax, int ymax, int ywmax, int ydist)
{
    if (xmin < 0) xmin = 0;
    trigger.posXmin = xmin;
    if (xmax > TRIGGER_POST_MAX) xmax = TRIGGER_POST_MAX;
    trigger.posXmax = xmax;
    if (ymax > TRIGGRT_LEVEL_VALUE_MAX) ymax = TRIGGRT_LEVEL_VALUE_MAX;
    trigger.posYmin = -ymax;
    trigger.posYmax = ymax;
    if (ywmax > TRIGGRT_WINDOW_VALUE_MAX) ywmax = TRIGGRT_WINDOW_VALUE_MAX;
    trigger.posYWmin = -ywmax;
    trigger.posYWmax = ywmax;
    if (ydist < TRIGGRT_WINDOW_DISTANCE_MIN) ydist = TRIGGRT_WINDOW_DISTANCE_MIN;
    trigger.posYDist = ydist;
}

void XlabInterface::triggerRefreshPos()
{
    if (trigger.mode == TRIGGER_MODE_Window) {
        triggerSetPositionWin1(trigger.coord[trigger.channel].posX, trigger.coord[trigger.channel].posY1);
        triggerSetPositionWin2(trigger.coord[trigger.channel].posX, trigger.coord[trigger.channel].posY2);
    } else {
        triggerSetPosition(trigger.coord[trigger.channel].posX, trigger.coord[trigger.channel].posY);
    }
}

int XlabInterface::triggerPosYPixelToInt(int posYpx)
{
    return (posYpx - (yAxisMax / 2) - trigger.coord[trigger.channel].offset) / (2 * GAIN_FACTOR);
}

int XlabInterface::triggerPosYIntToPixel(int posY)
{
    return ui->plotterWidget->yAxis->coordToPixel((2 * GAIN_FACTOR * posY + trigger.coord[trigger.channel].offset + (yAxisMax / 2)));
}

double XlabInterface::triggerPosYBit(int bit)
{
    return (20 + trigger.coord[trigger.channel].size) * bit + (trigger.coord[trigger.channel].size / 2) + 2 + trigger.coord[trigger.channel].offset + (yAxisMax / 2);
}

void XlabInterface::triggerMoveTrigger(int curPosX, int curPosY, bool force)
{
    if (!force)
        if (!trigger.enable) return;
    if (trigger.ssRate) curPosX = TRIGGER_POSX_SSR;
    if (trigger.channel == MAIN_CHANNEL_CHD)
        curPosY = ui->plotterWidget->yAxis->coordToPixel(triggerPosYBit(trigger.coord[trigger.channel].bit));
    triggerPixmap->topLeft->setPixelPosition(QPointF(curPosX - 9, curPosY - 8));
}

void XlabInterface::triggerMoveWinTrigger(int curPosX, int curPosY1, int curPosY2, bool force)
{
    if (!force)
        if (!trigger.enable) return;
    if (trigger.ssRate) curPosX = TRIGGER_POSX_SSR;
    triggerWin1Pixmap->topLeft->setPixelPosition(QPointF(curPosX - 9, curPosY1 - 8));
    triggerWin2Pixmap->topLeft->setPixelPosition(QPointF(curPosX - 9, curPosY2 - 8));
}

int XlabInterface::triggerPosXIntToPixel(int posX)
{
    return ui->plotterWidget->xAxis->coordToPixel(2 * posX);
}

void XlabInterface::triggerSetVisible(bool visible)
{
    trigger.visible = visible;
    if (!trigger.enable) visible = false;
    if (trigger.mode == TRIGGER_MODE_Window) {
        triggerWin1Pixmap->setVisible(visible);
        triggerWin2Pixmap->setVisible(visible);
        triggerPixmap->setVisible(false);
    } else {
        triggerWin1Pixmap->setVisible(false);
        triggerWin2Pixmap->setVisible(false);
        triggerPixmap->setVisible(visible);
    }
}

void XlabInterface::triggerSetSlowSamplingRate(bool value)
{
    trigger.ssRate = value;
}

void XlabInterface::triggerSetSource(int index)
{
    index = dataValueRangeTrunk(index, 0, 11);
    trigger.source = index;
    triggerSetIcon();
}

void XlabInterface::triggerSetIcon()
{
    if (!trigger.enable) return;
    triggerSetVisible(trigger.visible);
    if (trigger.mode == TRIGGER_MODE_Window) {
        if (trigger.source == MAIN_CHANNEL_CH1) {
            triggerWin1Pixmap->setPixmap(QPixmap(triggerIconPathsG[trigger.mode + 1]));
            triggerWin2Pixmap->setPixmap(QPixmap(triggerIconPathsG[trigger.mode]));
        } else if (trigger.source == MAIN_CHANNEL_CH2) {
            triggerWin1Pixmap->setPixmap(QPixmap(triggerIconPathsR[trigger.mode + 1]));
            triggerWin2Pixmap->setPixmap(QPixmap(triggerIconPathsR[trigger.mode]));
        }
    } else {
        if (trigger.source == MAIN_CHANNEL_CH1)
            triggerPixmap->setPixmap(QPixmap(triggerIconPathsG[trigger.mode]));
        else
            triggerPixmap->setPixmap(QPixmap(triggerIconPathsR[trigger.mode]));
    }
}

void XlabInterface::scopeSetTriggerType()
{
    if(ui->radioButtonFree->isChecked()) {
        ui->forceButton->setEnabled(false);
        triggerSetVisible(false);
    } else {
        ui->forceButton->setEnabled(true);
        if (ui->xyMode->isChecked() && ui->comboBoxTrigSource->currentIndex() == 0) triggerSetVisible(false);
        else triggerSetVisible(true);
    }
    sendTriggerControls();
    if (!(ui->radioButtonSingle->isChecked())) scopeStartForce();
    m_repaint = true;
}

void XlabInterface::scopeSetTriggerMode()
{
    if (ui->radioButtonRisingEdge->isChecked()) triggerSetMode(TRIGGER_MODE_Rising);
    else if (ui->radioButtonFallingEdge->isChecked()) triggerSetMode(TRIGGER_MODE_Falling);
    else if (ui->radioButtonDual->isChecked()) triggerSetMode(TRIGGER_MODE_Dual);
    else if (ui->radioButtonPositiveSlope->isChecked()) triggerSetMode(TRIGGER_MODE_Positive);
    else if (ui->radioButtonNegativeSlope->isChecked()) triggerSetMode(TRIGGER_MODE_Negative);
    else if (ui->radioButtonWindow->isChecked()) triggerSetMode(TRIGGER_MODE_Window);
    triggerRefreshPos();
    triggerSetIcon();
    sendTriggerControls();
}

void XlabInterface::scopeSetTriggerModeButtonVisible(bool visible)
{
    ui->radioButtonWindow->setEnabled(visible);
    ui->radioButtonPositiveSlope->setEnabled(visible);
    ui->radioButtonNegativeSlope->setEnabled(visible);
}

void XlabInterface::scopeSetTriggerAutoTimer(double value)
{
    byte data;

    data = (byte)((value / 0.04096) - 1);
    sendAutoTrigger(data);
}

void XlabInterface::scopeTriggerSourceChanged(int index)
{
    scopeSetTriggerType();
    if (index == 0) triggerSetChannel(MAIN_CHANNEL_CH1);
    else if (index == 1) triggerSetChannel(MAIN_CHANNEL_CH2);
    else triggerSetChannel(MAIN_CHANNEL_CHD);
    if ((index > 1) && (index < 10)) triggerSetChannelBit(MAIN_CHANNEL_CHD, index - 2);
    triggerSetSource(index);
    sendTriggerSource(index);
    if (index > 1) {
        if (ui->radioButtonWindow->isChecked()||ui->radioButtonNegativeSlope->isChecked()||ui->radioButtonPositiveSlope->isChecked()) {
            ui->radioButtonRisingEdge->setChecked(true);
            scopeSetTriggerMode();
        }
        scopeSetTriggerModeButtonVisible(false);
    } else {
        scopeSetTriggerModeButtonVisible(true);
    }
    triggerRefreshPos();
    sendTriggerPost(trigger.coord[trigger.channel].posX);
    sendTriggerLevel(trigger.coord[trigger.channel].posY);
    sendTriggerWin1Level(trigger.coord[trigger.channel].posY1);
    sendTriggerWin2Level(trigger.coord[trigger.channel].posY2);
}

void XlabInterface::keyPressEvent(QKeyEvent *event)
{
    if (event->key() == Qt::Key_Control) {
        ui->plotterWidget->setInteractions(QCP::iRangeDrag );
    }
    if (event->key() == Qt::Key_Shift) {
        ui->plotterWidget->axisRect()->setRangeZoom(Qt::Vertical);
        // ui->plotterWidget->setInteractions(QCP::iRangeDrag );
    }
}

void XlabInterface::keyReleaseEvent(QKeyEvent *event)
{
    if (event->key() == Qt::Key_Control) {
        ui->plotterWidget->setInteractions(QCP::iRangeZoom);
    }
    if (event->key() == Qt::Key_Shift) {
        ui->plotterWidget->axisRect()->setRangeZoom(Qt::Horizontal);
    }
}

void XlabInterface::selectItem(QMouseEvent *event)
{
    if (!(event->buttons() & Qt::LeftButton)) return;
    if (hCursorAHead->selectTest(event->pos(), false) >= 0 && hCursorAHead->selectTest(event->pos(), false) < 8.0) {
        currentSelected = isHCursorAHead;
        hCursorAHead->setPen(QPen(QColor("#1692e5")));
    } else if (hCursorBHead->selectTest(event->pos(), false) >= 0 && hCursorBHead->selectTest(event->pos(), false) < 8.0) {
        currentSelected = isHCursorBHead;
        hCursorBHead->setPen(QPen(QColor("#1692e5")));
    } else if (vCursorAHead->selectTest(event->pos(), false) >= 0 && vCursorAHead->selectTest(event->pos(), false) < 8.0) {
        currentSelected = isVCursorAHead;
        vCursorAHead->setPen(QPen(QColor("#e04e4e")));
    } else if (vCursorBHead->selectTest(event->pos(), false) >= 0 && vCursorBHead->selectTest(event->pos(), false) < 8.0) {
        currentSelected = isVCursorBHead;
        vCursorBHead->setPen(QPen(QColor("#e04e4e")));
    } else if (ch1ZeroHead->selectTest(event->pos(), false) >= 0 && ch1ZeroHead->selectTest(event->pos(), false) < 8.0) {
        currentSelected = isCH1Zero;
        ch1ZeroHead->setPen(QPen(QColor("#f9a94c")));
    } else if (ch2ZeroHead->selectTest(event->pos(), false) >= 0 && ch2ZeroHead->selectTest(event->pos(), false) < 8.0) {
        currentSelected = isCH2Zero;
        ch2ZeroHead->setPen(QPen(QColor("#f9a94c")));
    } else if (chDZeroHead->selectTest(event->pos(), false) >= 0 && chDZeroHead->selectTest(event->pos(), false) < 8.0) {
        currentSelected = isCHDZero;
        chDZeroHead->setPen(QPen(QColor("#f9a94c")));
    } else if (triggerPixmap->selectTest(event->pos(),false) >= 0 && triggerPixmap->selectTest(event->pos(), false) < 8.0) {
        currentSelected = isTriggerPixmap;
        if (ui->comboBoxTrigSource->currentIndex() == 0)
            triggerPixmap->setPen(QPen(QColor(75, 229, 28)));
        else
            triggerPixmap->setPen(QPen(QColor("#ff0000")));
    } else if (triggerWin1Pixmap->selectTest(event->pos(), false) >= 0 && triggerWin1Pixmap->selectTest(event->pos(), false) < 8.0) {
        currentSelected = isTriggerWin1Pixmap;
        if (ui->comboBoxTrigSource->currentIndex() == 0)
            triggerWin1Pixmap->setPen(QPen(QColor(75, 229, 28)));
        else
            triggerWin1Pixmap->setPen(QPen(QColor("#ff0000")));
    }
    else if (triggerWin2Pixmap->selectTest(event->pos(), false) >= 0 && triggerWin2Pixmap->selectTest(event->pos(), false) < 8.0) {
        currentSelected = isTriggerWin2Pixmap;
        if (ui->comboBoxTrigSource->currentIndex() == 0)
            triggerWin2Pixmap->setPen(QPen(QColor(75, 229, 28)));
        else
            triggerWin2Pixmap->setPen(QPen(QColor("#ff0000")));
    } else currentSelected = isNone;
}

void XlabInterface::deselectItem(QMouseEvent*)
{
    currentSelected = isNone;
    hCursorAHead->setPen(QPen(Qt::NoPen));
    hCursorBHead->setPen(QPen(Qt::NoPen));
    vCursorAHead->setPen(QPen(Qt::NoPen));
    vCursorBHead->setPen(QPen(Qt::NoPen));
    triggerPixmap->setPen(QPen(Qt::NoPen));
    triggerWin1Pixmap->setPen(QPen(Qt::NoPen));
    triggerWin2Pixmap->setPen(QPen(Qt::NoPen));
    ch1ZeroHead->setPen(QPen(Qt::NoPen));
    ch2ZeroHead->setPen(QPen(Qt::NoPen));
    chDZeroHead->setPen(QPen(Qt::NoPen));
}

void XlabInterface::xAxisChanged(QCPRange range)
{
    int newRangeLower = range.lower;
    int newRangeUpper = range.upper;

    if (range.lower < 0) {
        ui->plotterWidget->xAxis->setRangeLower(0);
        newRangeLower = 0;
    }
    if (range.lower > xAxisMax - xAxisMin) {
        ui->plotterWidget->xAxis->setRangeLower(xAxisMax - xAxisMin);
        newRangeLower = xAxisMax - xAxisMin;
    }
    if (range.upper > xAxisMax) {
        ui->plotterWidget->xAxis->setRangeUpper(xAxisMax);
        newRangeUpper = xAxisMax;
    }
    if (range.upper < (newRangeLower + xAxisMin)) {
        ui->plotterWidget->xAxis->setRangeUpper(newRangeLower + xAxisMin);
        newRangeUpper = newRangeLower + xAxisMin;
    }
    if (ui->comboBoxScopeCursorsChannel->currentIndex() == 0) {
        hCursorAHead->topLeft->setPixelPosition(QPointF(14, (ui->plotterWidget->yAxis->coordToPixel(hCursorAPosCh1 + ch1ZeroPos)) - 8));
        hCursorBHead->topLeft->setPixelPosition(QPointF(14, (ui->plotterWidget->yAxis->coordToPixel(hCursorBPosCh1 + ch1ZeroPos)) - 8));
    }
    if (ui->comboBoxScopeCursorsChannel->currentIndex() == 1) {
        hCursorAHead->topLeft->setPixelPosition(QPointF(14, (ui->plotterWidget->yAxis->coordToPixel(hCursorAPosCh2 + ch2ZeroPos)) - 8));
        hCursorBHead->topLeft->setPixelPosition(QPointF(14, (ui->plotterWidget->yAxis->coordToPixel(hCursorBPosCh2 + ch2ZeroPos)) - 8));
    }
    if (ui->checkBoxCHDTrace->isChecked()) logicLabelsRefresh();

    updateLabelPosition();
    vCursorRefresh();
    setHorizontalScrollBar(newRangeLower, newRangeUpper);
}

void XlabInterface::setHorizontalScrollBar(int lo, int up)
{
    ui->horizontalScrollBar->setPageStep(up - lo);
    ui->horizontalScrollBar->setMaximum(xAxisMax - up + lo);
    ui->horizontalScrollBar->setValue(lo);
}

void XlabInterface::yAxisChanged(QCPRange range)
{
    int newRangeLower = range.lower;
    int newRangeUpper = range.upper;

    if (range.lower < 0) {
        ui->plotterWidget->yAxis->setRangeLower(0);
        newRangeLower = 0;
    }
    if (range.lower > yAxisMax - yAxisMin) {
        ui->plotterWidget->yAxis->setRangeLower(yAxisMax - yAxisMin);
        newRangeLower = yAxisMax-yAxisMin;
    }
    if (range.upper > yAxisMax) {
        ui->plotterWidget->yAxis->setRangeUpper(yAxisMax);
        newRangeUpper = yAxisMax;
    }
    if (range.upper < (newRangeLower + yAxisMin)) {
        ui->plotterWidget->yAxis->setRangeUpper(newRangeLower + yAxisMin);
        newRangeUpper = newRangeLower + yAxisMin;
    }
    if (ui->checkBoxCHDTrace->isChecked()) logicLabelsRefresh();

    updateLabelPosition();
    zeroHeadRefresh();
    setVerticalScrollBar(newRangeLower, newRangeUpper);
}

void XlabInterface::setVerticalScrollBar(int lo, int up)
{
    ui->verticalScrollBar->setPageStep(up - lo);
    ui->verticalScrollBar->setMaximum(yAxisMax - up + lo);
    ui->verticalScrollBar->setValue(yAxisMax - up);
}

void XlabInterface::updateCh1Label(int ch1ctrl, int ch2ctrl)
{
    QString tmp_text = "";

    if (ch1ctrl & (uint8_t)(1 << MAIN_CHANNEL_CONTROL_Invert)) tmp_text += "-CH1";
    else tmp_text += " CH1";
    if (ch1ctrl & (uint8_t)(1 << MAIN_CHANNEL_CONTROL_Math)) {
        if (ch1ctrl & (uint8_t)(1 << MAIN_CHANNEL_CONTROL_Subtract)) {
            if (ch2ctrl & (uint8_t)(1 << MAIN_CHANNEL_CONTROL_Invert)) tmp_text += " + CH2";
            else tmp_text += " - CH2";
        } else if (ch1ctrl & (uint8_t)(1 << MAIN_CHANNEL_CONTROL_Multiply)) {
            if (ch2ctrl & (uint8_t)(1 << MAIN_CHANNEL_CONTROL_Invert)) tmp_text += " * -CH2";
            else tmp_text += " * CH2";
        }
    }
    tmp_text += " = " + gainText[ui->ch1GainSlider->value()];
    ui->ch1Label->setText(tmp_text);
}

void XlabInterface::updateCh2Label(int ch2ctrl, int ch1ctrl)
{
    QString tmp_text = "";

    if (ch2ctrl & (uint8_t)(1 << MAIN_CHANNEL_CONTROL_Invert)) tmp_text += "-CH2";
    else tmp_text += " CH2";
    if (ch2ctrl & (uint8_t)(1 << MAIN_CHANNEL_CONTROL_Math)) {
        if (ch2ctrl & (uint8_t)(1 << MAIN_CHANNEL_CONTROL_Subtract)) {
            if (ch1ctrl & (uint8_t)(1 << MAIN_CHANNEL_CONTROL_Invert)) tmp_text += " + CH1";
            else tmp_text += " - CH1";
        } else if (ch2ctrl & (uint8_t)(1 << MAIN_CHANNEL_CONTROL_Multiply)) {
            if (ch1ctrl & (uint8_t)(1 << MAIN_CHANNEL_CONTROL_Invert)) tmp_text += " * -CH1";
            else tmp_text += " * CH1";
        }
    }
    tmp_text += " = " + gainText[ui->ch2GainSlider->value()];
    ui->ch2Label->setText(tmp_text);
}

void XlabInterface::setInfoText(QString text, bool mode)
{
    ui->infoLabel->setText(text);
    ui->centralWidget->setDisabled(mode);
    ui->centralWidget->repaint();
}

void XlabInterface::displayControlsSet()
{
    double width = 2;
    ch1Pen.setWidthF(width);
    ch1Graph->setPen(ch1Pen);
    ch2Pen.setWidthF(width);
    ch2Graph->setPen(ch2Pen);
}

void XlabInterface::gridBoxSet(QColor color, Qt::PenStyle style)
{
    gridPen = QPen(color, 1, style);
    ui->plotterWidget->xAxis->grid()->setPen(gridPen);
    ui->plotterWidget->yAxis->grid()->setPen(gridPen);
}

void XlabInterface::counterButtonEnable(bool state)
{
    ui->pushButtonCounterStart->setEnabled(state);
    ui->pushButtonCounteStop->setEnabled(false);
}

void XlabInterface::logToFile(QString s, int level)
{
    switch (level) {
        case LOGGING_LEVEL_Debug:
            qDebug() << s.toUtf8().constData() << "\n";
            break;
        #if (QT_VERSION >= QT_VERSION_CHECK(5, 5, 0))
        case LOGGING_LEVEL_Info:
            qInfo() << s.toUtf8().constData() << "\n";
            break;
        #endif
        case LOGGING_LEVEL_Warning:
            qWarning() << s.toUtf8().constData() << "\n";
            break;
        case LOGGING_LEVEL_Critical:
            qCritical() << s.toUtf8().constData() << "\n";
            break;
        case LOGGING_LEVEL_Fatal:
            //QTBUG-38992
            //qFatal() << s.toUtf8().constData() << "\n";
            qCritical() << s.toUtf8().constData() << "\n";
            break;
    }
}

void XlabInterface::clearMMLabelsFreq()
{
    ui->multimeterdisplayFrequencyValue->setValue(0);
    ui->multimeterdisplayDutyValue->setValue(0);
    ui->multimeterdisplayDutyValue_s->setValue(0);
    ui->multimeterdisplayPeriodValue->setValue(0);
    ui->multimeterdisplayCounter->setValue(0);
}

void XlabInterface::clearMMLabelsCH1AC()
{
    ui->multimeterdisplayCh1Vpp->setValue(0.0);
    ui->multimeterdisplayCh1Min->setValue(0.0);
    ui->multimeterdisplayCh1Max->setValue(0.0);
    ui->multimeterdisplayCh1RMS->setValue(0.0);
    ui->multimeterdisplayCh1Average->setValue(0.0);
    ui->multimeterdisplayCh1Frequency->setValue(0.0);
}

void XlabInterface::clearMMLabelsCH2AC()
{
    ui->multimeterdisplayCh2Vpp->setValue(0.0);
    ui->multimeterdisplayCh2Min->setValue(0.0);
    ui->multimeterdisplayCh2Max->setValue(0.0);
    ui->multimeterdisplayCh2RMS->setValue(0.0);
    ui->multimeterdisplayCh2Average->setValue(0.0);
    ui->multimeterdisplayCh2Frequency->setValue(0.0);
}

double XlabInterface::int8ToVolt_f(uint8_t gain, int8_t v0, int8_t v1)
{
    return rawValueToVolt((2 * (v0 - v1) * GAIN_FACTOR), gainText[gain]);
}

double XlabInterface::int8ToVolt_f(uint8_t gain, int8_t v0)
{
    return rawValueToVolt((2 * v0 * GAIN_FACTOR), gainText[gain]);
}

double XlabInterface::int16ToVolt_f(uint8_t gain, int16_t v0)
{
    return rawValueToVolt(((double)(v0) / 64) * GAIN_FACTOR, gainText[gain]);
}

double XlabInterface::samplex100ToFrequency_f(int32_t fpix, uint8_t samplingrate)
{
    double deltaTime, freq;
    QString unit;

    deltaTime = vCursorDeltaTime(0, fpix, &unit, 50, samplingrate);
    freq = deltaToFrequency(deltaTime, samplingrate);
    return freq;
}

void XlabInterface::meterRequestData()
{
    unsigned char tmp_buff[2 * DATA_CHMETER_RESULT_LENGTH];
    channela_t MMCH1buffer, MMCH2buffer;
    int size;

    if (!usbDevice.isDeviceConnected) return;
    clearBuffer(tmp_buff, 2 * DATA_CHMETER_RESULT_LENGTH);
    size = usbDevice.controlReadTransfer(USB_REQUEST_DATA, tmp_buff, 2 * DATA_CHMETER_RESULT_LENGTH, MAIN_DATA_REQUEST_Measure);
    if (size != 2 * DATA_CHMETER_RESULT_LENGTH) {
        logToFile("Meter, data too short", LOGGING_LEVEL_Warning);
        return;
    }
    dataChannelMeterGetData(&MMCH1buffer, tmp_buff);
    dataChannelMeterGetData(&MMCH2buffer, tmp_buff + DATA_CHMETER_RESULT_LENGTH);

    if (ui->mmTabWidget->currentIndex() == 0) {          // AC, 8-bit
        if (ui->groupBoxCH1AC->isChecked()) {
            ui->multimeterdisplayCh1Vpp->setValue(int8ToVolt_f(MMCH1buffer.gain, MMCH1buffer.max, MMCH1buffer.min));
            ui->multimeterdisplayCh1Max->setValue(int8ToVolt_f(MMCH1buffer.gain, MMCH1buffer.max));
            ui->multimeterdisplayCh1Min->setValue(int8ToVolt_f(MMCH1buffer.gain, MMCH1buffer.min));
            ui->multimeterdisplayCh1RMS->setValue(int16ToVolt_f(MMCH1buffer.gain, MMCH1buffer.rms));
            ui->multimeterdisplayCh1Average->setValue(int16ToVolt_f(MMCH1buffer.gain, MMCH1buffer.aver));
            ui->multimeterdisplayCh1Frequency->setValue(samplex100ToFrequency_f(MMCH1buffer.frequency, MMCH1buffer.samplingrate));
        } else clearMMLabelsCH1AC();
        if (ui->groupBoxCH2AC->isChecked()) {
            ui->multimeterdisplayCh2Vpp->setValue(int8ToVolt_f(MMCH2buffer.gain, MMCH2buffer.max, MMCH2buffer.min));
            ui->multimeterdisplayCh2Max->setValue(int8ToVolt_f(MMCH2buffer.gain, MMCH2buffer.max));
            ui->multimeterdisplayCh2Min->setValue(int8ToVolt_f(MMCH2buffer.gain, MMCH2buffer.min));
            ui->multimeterdisplayCh2RMS->setValue(int16ToVolt_f(MMCH2buffer.gain, MMCH2buffer.rms));
            ui->multimeterdisplayCh2Average->setValue(int16ToVolt_f(MMCH2buffer.gain, MMCH2buffer.aver));
            ui->multimeterdisplayCh2Frequency->setValue(samplex100ToFrequency_f(MMCH2buffer.frequency, MMCH2buffer.samplingrate));
        } else clearMMLabelsCH2AC();
    } else if (ui->mmTabWidget->currentIndex() == 1) {   // DC, 12-bit
        if (ui->groupBoxCH1DC->isChecked()) {
            ui->multimeterdisplayCh1DC->setValue(int16ToVolt_f(MMCH1buffer.gain, MMCH1buffer.aver));
        } else ui->multimeterdisplayCh1DC->setValue(0.0);
        if (ui->groupBoxCH2DC->isChecked()) {
            ui->multimeterdisplayCh2DC->setValue(int16ToVolt_f(MMCH2buffer.gain, MMCH2buffer.aver));
        } else ui->multimeterdisplayCh2DC->setValue(0.0);
    } else if (ui->mmTabWidget->currentIndex() == 2) {   // f, count, pulse, gate
        if (ui->radioPulse->isChecked()) {
            ui->multimeterdisplayCounter->setValue(MMCH1buffer.frequency);
        } else if (ui->radioPeriod->isChecked()) {
            ui->multimeterdisplayPeriodValue->setValue((double)MMCH1buffer.frequency / 32000000);
            if ((double)MMCH1buffer.frequency != 0) {
                ui->multimeterdisplayDutyValue->setValue(100 * ((double)MMCH2buffer.frequency) / ((double)MMCH1buffer.frequency));
                ui->multimeterdisplayDutyValue_s->setValue((double)MMCH2buffer.frequency / 32000000);
            } else {
                ui->multimeterdisplayDutyValue->setNaN();
                ui->multimeterdisplayDutyValue_s->setNaN();
            }
            if ((double)MMCH1buffer.frequency != 0) {
                ui->multimeterdisplayFrequencyValue->setValue((double)(32000000) / MMCH1buffer.frequency);
            } else {
                ui->multimeterdisplayFrequencyValue->setNaN();
            }
        } else if (ui->radioFreq->isChecked()) {
            ui->multimeterdisplayFrequencyValue->setValue(MMCH1buffer.frequency);
            if ((double)MMCH1buffer.frequency != 0) {
                ui->multimeterdisplayPeriodValue->setValue((double)1 / MMCH1buffer.frequency);
            } else {
                ui->multimeterdisplayPeriodValue->setNaN();
            }
        } else if (ui->radioExtGate->isChecked()) {
            ui->multimeterdisplayCounter->setValue(MMCH1buffer.frequency);
        }
    }
}

void XlabInterface::fftWindowFunctionSet()
{
    int type;

    if (ui->radioButtonRect->isChecked()) type = FFT_FUNCTION_Rectangular;
    else if (ui->radioButtonHamming->isChecked()) type = FFT_FUNCTION_Hamming;
    else if (ui->radioButtonHann->isChecked()) type = FFT_FUNCTION_Hann;
    else if (ui->radioButtonBlackman->isChecked()) type = FFT_FUNCTION_Blackman;

    delete[]fftWindow;
    fftWindow = new double[bufferSize];
    for (int i = 0; i < bufferSize; ++i) {
        double x = 1.0;

        switch (type) {
        case FFT_FUNCTION_Rectangular: {
                x = 1.0;
                break;
            }
        case FFT_FUNCTION_Hamming: {
                x = 0.53836 - 0.46164 * qCos((2 * M_PI * i) / (bufferSize - 1));
                break;
            }
        case FFT_FUNCTION_Hann: {
                x = 0.5 * (1 - qCos((2 * M_PI * i) / (bufferSize - 1)));
                break;
            }
        case FFT_FUNCTION_Blackman: {
                x = 0.42 - 0.5 * qCos((2 * M_PI * i) / (bufferSize - 1)) + 0.08 * qCos(( 4 *M_PI * i) / (bufferSize - 1));
                break;
            }
        }
        fftWindow[i] = x;
    }
}

/*
void XlabInterface::on_usbConnection_Lost()
{
    logToFile("Signal >Connection Lost<", LOGGING_LEVEL_Warning);
    if (usbDevice.usbConnectionLost){
        logToFile("USB Connection Lost, try to reconnect", LOGGING_LEVEL_Warning);
        if (usbDevice.isDeviceConnected)
            on_connectButton_clicked();
    }
}
*/

void XlabInterface::disconnectDevice(bool mode)
{
    usbDevice.controlWriteTransfer(USB_DISCONNECT, 0, 0);
    if (usbDevice.usbConnectionLost)
        changeWindowTitle("Interface - Connection lost... wait");
    else
        changeWindowTitle("Interface - Disconnecting... wait");
    if (mode)
        usbDevice.closeDevice();
    usbDevice.reset();
    ui->connectLabel->setText(tr("Not Connected"));
    changeWindowTitle("Interface - Not Connected");
    ui->connectIcon->setPixmap(QPixmap(":/icons/led-off.png"));
}

int XlabInterface::deviceInitialize()
{
    int result = 0;
    unsigned char tmp_buff[DATA_DEVICE_INFO_LENGTH];
    QString version;

    clearBuffer(tmp_buff, DATA_DEVICE_INFO_LENGTH);
    int size = usbDevice.controlReadTransfer(USB_REQUEST_DATA, tmp_buff, DATA_DEVICE_INFO_LENGTH);
    if (size != DATA_DEVICE_INFO_LENGTH){
        logToFile("Device initializing: unknow device, disconnecting ...", LOGGING_LEVEL_Critical);
        disconnectDevice(false);
        return 1;
    }
    version = getStringFromUnsignedChar(tmp_buff, size);
    logToFile("Device connected: " + version, LOGGING_LEVEL_Info);
    ui->textBrowserDeviceInfo->append("Device connected: " + version);
    result = deviceProfileGet(version);
    if (result == 0) {
        result = deviceSettingsInitialize();
        if (result == 0) {
            ui->connectLabel->setText(tr("USB Connected"));
            changeWindowTitle("Interface - USB Connected");
            ui->connectIcon->setPixmap(QPixmap(":/icons/led-on.png"));
            setInfoText("USB Connected");
        }
    }
    return result;
}

int XlabInterface::deviceSettingsInitialize()
{
    int result = 0;

    usbDevice.controlWriteTransfer(USB_SEND_CONTROL, MAIN_CONTROL_SCOPE_Stop);
    writeBulkConfig = true;
    deviceSettingsSendBulk();
    writeBulkConfig = false;
    //Sleep(300);
    usbDevice.controlWriteTransfer(USB_SEND_DATA_BULK, MAIN_BULK_DATA_Config);
    logToFile("Initialize bulk read transfer", LOGGING_LEVEL_Debug);
    usbDevice.asyncBulkReadTransfer();
    //Sleep(100);
    deviceSettingsSendInteractive();
    QTimer::singleShot(2000, this, SLOT(scopeStartForce()));
    return result;
}

void XlabInterface::deviceSettingsSendBulk()
{
    deviceSettingsExec();
    logToFile("Finish prepare configuration, now send buffer as AWG", LOGGING_LEVEL_Debug);
    for (int i=0; i < MAIN_CONFIGURATION_RangeOut; i++)
        logToFile(">> buffer: index[" + QString::number(i) + "] value:" + QString::number(awgBuffer[4 * i]), LOGGING_LEVEL_Debug);
    usbDevice.bulkWriteTransfer(awgBuffer, USB_ENDPOINT1_AUXDATA_LENGTH);
}

void XlabInterface::deviceSettingsSendInteractive()
{
    awgDACBufferChanged(ui->comboBoxDACBuffer->currentIndex());
    scopeADCBufferChanged(ui->comboBoxADCBuffer->currentIndex());
    frequencyDigitsValueChanged();
    channel1TraceChanged(ui->checkBoxCH1Trace->isChecked());
    channel2TraceChanged(ui->checkBoxCH2Trace->isChecked());
    channelLogicTraceChanged();
    awgEnableChanged();
}

int XlabInterface::deviceProfileGet(QString version) {
    int result = 0;
    int hardLo, hardHi, softLo, softHi;
    QString configFile;

    version = version.split(' ').first();
    QXmlGet xmlGet;
    xmlGet.load("hardware.xml");
    if (xmlGet.docType() == "XLabHardware") {
        if (xmlGet.find(version.split('.').first())) {
            hardLo = xmlGet.getAttributeInt("hardLo");
            hardHi = xmlGet.getAttributeInt("hardHi");
            softLo = xmlGet.getAttributeInt("softLo");
            softHi = xmlGet.getAttributeInt("softHi");
            configFile = xmlGet.getAttributeString("file");
            if ((version.split('.').at(1).toInt() < hardLo) || (version.split('.').at(1).toInt() > hardHi)) {
                QMessageBox::critical(this, "Hardware", "Hardware version (" + version + ") not supported");
                result = 1;
            }
            if ((version.split('.').at(2).toInt() < softLo) || (version.split('.').at(2).toInt() > softHi)) {
                QMessageBox::critical(this, "Software", "Software version (" + version + ") not supported");
                result = 2;
            }
            QFile hardwareConfig(configFile);
            if (!hardwareConfig.exists()) {
                QMessageBox::critical(this, "Software", "Can't read file: " + configFile);
                result = 3;
            }
            else deviceInfoRead(configFile);
        } else {
            QMessageBox::critical(this, "Hardware", "hardware (" + version + ") not suported");
            result = 4;
        }
    } else {
        QMessageBox::critical(this, "Hardware", "Wrong doctype in hardware.xml!");
        result = 5;
    }
    return result;
}

void XlabInterface::fftControlsPlotSet()
{
    if (ui->checkBoxIQFFT->isChecked()) {
        ui->groupBoxPlot->setEnabled(false);
    } else {
        ui->groupBoxPlot->setEnabled(true);
    }
}

void XlabInterface::deviceSettingsExec()
{
    // CH1
    {
        bool ch1CursorUpdate = false;
        channel1GainSliderChanged(ui->ch1GainSlider->value(), false);
        if (ui->comboBoxScopeCursorsChannel->currentIndex() == 0) ch1CursorUpdate = true;
        channelPositionSliderChanged(MAIN_CHANNEL_CH1, ui->ch1PositionSlider->value(), &ch1ZeroPos, ch1CursorUpdate);
        channel1ACDCChanged(false);
        sendCH1Controls();
    }
    // CH2
    {
        bool ch2CursorUpdate = false;
        channel2GainSliderChanged(ui->ch2GainSlider->value(), false);
        if (ui->comboBoxScopeCursorsChannel->currentIndex() == 1) ch2CursorUpdate = true;
        channelPositionSliderChanged(MAIN_CHANNEL_CH2, ui->ch2PositionSlider->value(), &ch2ZeroPos, ch2CursorUpdate);
        channel2ACDCChanged(false);
        sendCH2Controls();
    }
    // LOGIC
    {
        channelLogicSizeSliderChanged();
        channelLogicPositionSliderChanged(ui->chdPositionSlider->value());
        sendCHDBitControls();
    }
    // FFT
    {
        fftControlsPlotSet();
        fftWindowFunctionSet();
        on_checkBoxFFTTrace_clicked();
    }
    // Display
    {
        displayIntensitySliderChanged(ui->intensitySlider->value());
        displayVectorsChanged(ui->checkBoxVectors->isChecked());
        displayControlsSet();
        applyScopeConfiguration();
    }
    // Cursors
    {
        cursorsVerticalChanged(ui->checkBoxCursorVertical->isChecked());
        cursorsHorizontalChanged(ui->checkBoxScopeCursorsHorizontal->isChecked());
        cursorsHorizontalTypeSet(ui->comboBoxScopeCursorsHorizontalType->currentIndex());
        cursorsSetChannelActive(ui->comboBoxScopeCursorsChannel->currentIndex());
    }
    // Trigger
    {
        scopeTriggerSourceChanged(ui->comboBoxTrigSource->currentIndex());
        scopeSetTriggerAutoTimer(ui->doubleSpinBoxTrigAuto->value());
        sendTriggerHold(ui->doubleSpinBoxTrigHold->value());
        sendTriggerLevel(trigger.coord[trigger.channel].posY);
        sendTriggerWin1Level(trigger.coord[trigger.channel].posY1);
        sendTriggerWin2Level(trigger.coord[trigger.channel].posY2);
        sendTriggerPost(trigger.coord[trigger.channel].posX);
        scopeSetTriggerType();
        scopeSetTriggerMode();
    }
    // Sampling
    {
        scopeRollModeChanged();
        scopeElasticModeChanged();
        scopeSamplingSliderChanged(ui->samplingSlider->value());
    }
    // MeterFC
    {
        sendTriggerSource();
        sendGateSource();
        meterFrequencyModeChanged();
    }
    // Meter
    {
        meterTabChanged(ui->mmTabWidget->currentIndex());
    }
    // AWG
    {
        sendWaveType();
        awgAmplitudeSliderChanged(ui->amplitudeSlider->value());
        awgDutyCycleSliderChanged(ui->dutyCycleSlider->value());
        awgOffsetSliderChanged(ui->offsetSlider->value());
        frequencyDigitsValueChanged();
    }
    // AWGSweep
    {
        sweepSpeedSliderChanged(ui->sweepSpeedSlider->value());
        sendSweepTypeControls();
        sweepStartFreqChanged(ui->sweepStartFreqSlider, ui->sweepEndFreqSlider);
        sweepEndFreqChanged(ui->sweepStartFreqSlider, ui->sweepEndFreqSlider);
        sweepTypeChanged();
    }
    // Sniffer
    {
        sendMixedControls();
        sniffProtocolTabChanged(ui->protocolTabWidget->currentIndex());
    }
    // SnifferSPI
    {
        sendSnifferOptionControl();
    }
    // SnifferUART
    {
        sendSnifferOptionControl();
        sendSnifferUSARTControl();
    }
    // SnifferParallel
    {
        sendSnifferOptionControl();
        sendSnifferStrobeEdge(ui->comboBoxSniffParallelStrobe->currentIndex());
    }
    mainTabChanged(ui->mainTabWidget->currentIndex());
}

void XlabInterface::deviceSettingsLoad()
{

    tomap_PythonConsoleOptions->readFromSettings(tomap_PythonConsoleOptions->getName());
    tomap_ScopeOptions->readFromSettings(tomap_ScopeOptions->getName());
    tomap_MultimeterDisplayOptions->readFromSettings(tomap_MultimeterDisplayOptions->getName());
    // ScopeOptions->readFromSettins("ScopeOptions"->comonname);     // <-------------
    QSettings settings;
    QString val;
    logToFile("Read saved device settings", LOGGING_LEVEL_Debug);
    settings.beginGroup("CH1");
    {
        ui->ch1GainSlider->setValue(settings.value("Gain", 0).toInt());
        ui->ch1PositionSlider->setValue(settings.value("Position", 128).toInt());
        ui->checkBoxCH1Trace->setChecked(settings.value("Trace", true).toBool());
        ui->checkBoxCH1ACDC->setChecked(settings.value("AC", false).toBool());
        ui->checkBoxCH1Average->setChecked(settings.value("Average", false).toBool());
        ui->checkBoxCH1Invert->setChecked(settings.value("Invert", false).toBool());
        ui->checkBoxCH1Math->setChecked(settings.value("MathEn", false).toBool());
        val = settings.value("Math", "Multiply").toString();
        if (val == "Sub") ui->radioButtonCH1Sub->setChecked(true);
        else if (val == "Multiply") ui->radioButtonCH1Multiply->setChecked(true);
        settings.endGroup();
    }
    settings.beginGroup("CH2");
    {
        ui->ch2GainSlider->setValue(settings.value("Gain", 0).toInt());
        ui->ch2PositionSlider->setValue(settings.value("Position", 32).toInt());
        ui->checkBoxCH2Trace->setChecked(settings.value("Trace", true).toBool());
        ui->checkBoxCH2ACDC->setChecked(settings.value("AC", false).toBool());
        ui->checkBoxCH2Average->setChecked(settings.value("Average", false).toBool());
        ui->checkBoxCH2Invert->setChecked(settings.value("Invert", false).toBool());
        ui->checkBoxCH2Math->setChecked(settings.value("MathEn", false).toBool());
        val = settings.value("Math", "Multiply").toString();
        if (val == "Sub") ui->radioButtonCH2Sub->setChecked(true);
        else if (val == "Multiply") ui->radioButtonCH2Multiply->setChecked(true);
        settings.endGroup();
    }
    settings.beginGroup("LOGIC");
    {
        ui->chdSizeSlider->setValue(settings.value("Size", 0).toInt());
        ui->chdPositionSlider->setValue(settings.value("Position", -96).toInt());
        ui->checkBoxCHDTrace->setChecked(settings.value("Trace", false).toBool());
        ui->checkBoxCHDInvert->setChecked(settings.value("Invert", false).toBool());
        ui->checkBoxCHD0->setChecked(settings.value("CHD0En", true).toBool());
        ui->checkBoxCHD1->setChecked(settings.value("CHD1En", true).toBool());
        ui->checkBoxCHD2->setChecked(settings.value("CHD2En", true).toBool());
        ui->checkBoxCHD3->setChecked(settings.value("CHD3En", true).toBool());
        ui->checkBoxCHD4->setChecked(settings.value("CHD4En", true).toBool());
        ui->checkBoxCHD5->setChecked(settings.value("CHD5En", true).toBool());
        ui->checkBoxCHD6->setChecked(settings.value("CHD6En", true).toBool());
        ui->checkBoxCHD7->setChecked(settings.value("CHD7En", true).toBool());
        settings.endGroup();
    }
    settings.beginGroup("FFT");
    {
        ui->verticalSliderFFTPosition->setValue(settings.value("Position", 0).toInt());
        ui->checkBoxFFTTrace->setChecked(settings.value("Trace", false).toBool());
        ui->checkBoxIQFFT->setChecked(settings.value("IQFFT", false).toBool());
        ui->checkBoxLogY->setChecked(settings.value("LogY", false).toBool());
        ui->checkBoxFFTCH1->setChecked(settings.value("CH1", true).toBool());
        ui->checkBoxFFTCH2->setChecked(settings.value("CH2", false).toBool());
        val = settings.value("Window", "Rect").toString();
        if (val == "Rect") ui->radioButtonRect->setChecked(true);
        else if (val == "Hamming") ui->radioButtonHamming->setChecked(true);
        else if (val == "Hann") ui->radioButtonHann->setChecked(true);
        else if (val == "Blackman") ui->radioButtonBlackman->setChecked(true);
        settings.endGroup();
    }
    settings.beginGroup("Display");
    {
        ui->intensitySlider->setValue(settings.value("Intensity", 0).toInt());
        ui->decaySlider->setValue(settings.value("Decay", 0).toInt());
        ui->checkBoxVectors->setChecked(settings.value("Vectors", true).toBool());
        ui->checkBoxPersistence->setChecked(settings.value("Persistence", false).toBool());
        ui->checkBoxmeterondev->setChecked(settings.value("MeterOn", true).toBool());
        settings.endGroup();
    }
    settings.beginGroup("Cursors");
    {
        ui->checkBoxCursorVertical->setChecked(settings.value("Vertical", true).toBool());
        ui->checkBoxScopeCursorsHorizontal->setChecked(settings.value("Horizontal", true).toBool());
        val = settings.value("HType", "Automatic").toString();
        if (val == "Manual") ui->comboBoxScopeCursorsHorizontalType->setCurrentIndex(2);
        else if (val == "Automatic") ui->comboBoxScopeCursorsHorizontalType->setCurrentIndex(1);
        else if (val == "Track") ui->comboBoxScopeCursorsHorizontalType->setCurrentIndex(0);
        val = settings.value("HChannel", "CH1").toString();
        if (val == "CH1") ui->comboBoxScopeCursorsChannel->setCurrentIndex(0);
        else if (val == "CH2") ui->comboBoxScopeCursorsChannel->setCurrentIndex(1);
        settings.endGroup();
    }
    settings.beginGroup("Trigger");
    {
        ui->comboBoxTrigSource->setCurrentIndex(settings.value("Source", 0).toInt());
        ui->doubleSpinBoxTrigAuto->setValue(settings.value("TAuto", 0.163840).toDouble());
        ui->doubleSpinBoxTrigHold->setValue(settings.value("THold", 10).toDouble());
        val = settings.value("Type", "Free").toString();
        if (val == "Free") ui->radioButtonFree->setChecked(true);
        else if (val == "Triger") ui->radioButtonNormal->setChecked(true);
        else if (val == "Auto") ui->radioButtonAuto->setChecked(true);
        else if (val == "Single") ui->radioButtonSingle->setChecked(true);
        val=settings.value("Mode", "Rising").toString();
        if (val == "Rising") ui->radioButtonRisingEdge->setChecked(true);
        else if (val == "Falling") ui->radioButtonFallingEdge->setChecked(true);
        else if (val == "Dual") ui->radioButtonDual->setChecked(true);
        else if (val == "Positive") ui->radioButtonPositiveSlope->setChecked(true);
        else if (val == "Negative") ui->radioButtonNegativeSlope->setChecked(true);
        else if (val == "Window") ui->radioButtonWindow->setChecked(true);
        settings.beginGroup("CH0Coord");
        {
            trigger.coord[0].posY = settings.value("TLevel", -4).toInt();
            trigger.coord[0].posY1 = settings.value("TWin1Level", 96).toInt();
            trigger.coord[0].posY2 = settings.value("TWin2Level", -96).toInt();
            trigger.coord[0].posX = settings.value("TPost", 64).toInt();
            settings.endGroup();
        }
        settings.beginGroup("CH1Coord");
        {
            trigger.coord[1].posY = settings.value("TLevel", -4).toInt();
            trigger.coord[1].posY1 = settings.value("TWin1Level", 96).toInt();
            trigger.coord[1].posY2 = settings.value("TWin2Level", -96).toInt();
            trigger.coord[1].posX = settings.value("TPost", 64).toInt();
            settings.endGroup();
        }
        settings.beginGroup("CH2Coord");
        {
            trigger.coord[2].posY = settings.value("TLevel", -4).toInt();
            trigger.coord[2].posY1 = settings.value("TWin1Level", 96).toInt();
            trigger.coord[2].posY2 = settings.value("TWin2Level", -96).toInt();
            trigger.coord[2].posX = settings.value("TPost", 64).toInt();
            settings.endGroup();
        }
        settings.endGroup();
    }
    settings.beginGroup("Sampling");
    {
        ui->samplingSlider->setValue(settings.value("Value", 6).toInt());
        ui->xyMode->setChecked(settings.value("XY", false).toBool());
        ui->rollMode->setChecked(settings.value("Roll", false).toBool());
        ui->elasticMode->setChecked(settings.value("Elastic", false).toBool());
        settings.endGroup();
    }
    settings.beginGroup("Meter");
    {
        ui->mmTabWidget->setCurrentIndex(settings.value("Mode", 0).toInt());
        settings.endGroup();
    }
    settings.beginGroup("MeterFC");
    {
        val = settings.value("Source", "EXT").toString();
        if (val == "Schmitt") ui->fcR11->setChecked(true);
        else if (val == "EXT") ui->fcR10->setChecked(true);
        else if (val == "7") ui->fcR7->setChecked(true);
        else if (val == "6") ui->fcR6->setChecked(true);
        else if (val == "5") ui->fcR5->setChecked(true);
        else if (val == "4") ui->fcR4->setChecked(true);
        else if (val == "3") ui->fcR3->setChecked(true);
        else if (val == "2") ui->fcR2->setChecked(true);
        else if (val == "1") ui->fcR1->setChecked(true);
        else if (val == "0") ui->fcR0->setChecked(true);
        val = settings.value("Gate", "EXT").toString();
        if (val == "Schmitt") ui->fcR11_2->setChecked(true);
        else if (val == "EXT") ui->fcR10_2->setChecked(true);
        else if (val == "7") ui->fcR7_2->setChecked(true);
        else if (val == "6") ui->fcR6_2->setChecked(true);
        else if (val == "5") ui->fcR5_2->setChecked(true);
        else if (val == "4") ui->fcR4_2->setChecked(true);
        else if (val == "3") ui->fcR3_2->setChecked(true);
        else if (val == "2") ui->fcR2_2->setChecked(true);
        else if (val == "1") ui->fcR1_2->setChecked(true);
        else if (val == "0") ui->fcR0_2->setChecked(true);
        val = settings.value("Type", "Frequency").toString();
        if (val == "Frequency") ui->radioFreq->setChecked(true);
        else if (val == "CountPulses") ui->radioPulse->setChecked(true);
        else if (val == "Reriod") ui->radioPeriod->setChecked(true);
        else if (val == "ExternalGate") ui->radioExtGate->setChecked(true);
        ui->checkBoxFreqFastMode->setChecked(settings.value("FastMode", false).toBool());
        val = settings.value("GateType", "Pulse").toString();
        if (val == "Period") ui->radioFrequencyGatePeriod->setChecked(true);
        else if (val == "Pulse") ui->radioFrequencyGatePulse->setChecked(true);
        settings.endGroup();
    }
    settings.beginGroup("AWG");
    {
        ui->checkBoxAWBEnable->setChecked(settings.value("AWGEnable", true).toBool());
        val = settings.value("Waveform", "Sine").toString();
        if (val == "Sine") ui->radioButtonSine->setChecked(true);
        else if (val == "Triangle") ui->radioButtonTriangle->setChecked(true);
        else if (val == "Expo") ui->radioButtonExpo->setChecked(true);
        else if (val == "Square") ui->radioButtonSquare->setChecked(true);
        else if (val == "WhiteNoise") ui->radioButtonWhiteNoise->setChecked(true);
        else if (val == "PinkNoise") ui->radioButtonPinkNoise->setChecked(true);
        else if (val == "Constant") ui->radioButtonConstant->setChecked(true);
        else if (val == "Custom") ui->radioButtonCustom->setChecked(true);
        ui->spinBoxConstant->setValue(settings.value("ConstantValue", 0).toDouble());
        ui->labelFrequencyDigit100000->setText(settings.value("FDigit100000", "0").toString());
        ui->labelFrequencyDigit10000->setText(settings.value("FDigit10000", "0").toString());
        ui->labelFrequencyDigit1000->setText(settings.value("FDigit1000", "1").toString());
        ui->labelFrequencyDigit100->setText(settings.value("FDigit100", "0").toString());
        ui->labelFrequencyDigit10->setText(settings.value("FDigit10", "0").toString());
        ui->labelFrequencyDigit1->setText(settings.value("FDigit1", "0").toString());
        ui->labelFrequencyDigit01->setText(settings.value("FDigit01", "0").toString());
        ui->labelFrequencyDigit001->setText(settings.value("FDigit001", "0").toString());
        ui->doubleSpinBoxAmp->setValue(settings.value("Amplitude", 96).toDouble());
        ui->doubleSpinBoxDuty->setValue(settings.value("Duty", 128).toDouble());
        ui->doubleSpinBoxOffset->setValue(settings.value("Offset", 0).toDouble());
        settings.endGroup();
    }
    settings.beginGroup("AWGSweep");
    {
        ui->sweepSpeedSlider->setValue(settings.value("Speed", 64).toInt());
        ui->collapsiblePanelAWGSweep->setChecked(settings.value("Active", false).toBool());
        val = settings.value("Type", "Duty").toString();
        if (val == "Frequency") ui->radioButtonSweepFrequency->setChecked(true);
        else if (val == "Amplitude") ui->radioButtonSweepAmplitude->setChecked(true);
        else if (val == "Duty") ui->radioButtonSweepDutyCycle->setChecked(true);
        else if (val == "Offset") ui->radioButtonSweepOffset->setChecked(true);
        val = settings.value("Direction", "PingPong").toString();
        if (val == "Rising") ui->radioButtonSweepRising->setChecked(true);
        else if (val == "PingPong") ui->radioButtonSweepPingPong->setChecked(true);
        else if (val == "Falling") ui->radioButtonSweepFalling->setChecked(true);
        ui->sweepStartFreqSlider->setValue(settings.value("Start", 64).toInt());
        ui->sweepEndFreqSlider->setValue(settings.value("End", 192).toInt());
        settings.endGroup();
    }
    settings.beginGroup("Sniffer");
    {
        val = settings.value("Mode", "Single").toString();
        if (val == "Normal") ui->radioButtonSniffNormal->setChecked(true);
        else if (val == "Single") ui->radioButtonSniffSingle->setChecked(true);
        ui->checkBoxCircular->setChecked(settings.value("Circular", false).toBool());
        ui->checkBoxASCII->setChecked(settings.value("ASCII", false).toBool());
        ui->protocolTabWidget->setCurrentIndex(settings.value("Protocol", 0).toInt());
        settings.endGroup();
    }
    settings.beginGroup("SnifferSPI");
    {
        ui->comboBoxCPOL->setCurrentIndex(settings.value("CPOL", 0).toInt());
        ui->comboBoxCPHA->setCurrentIndex(settings.value("CPHA", 0).toInt());
        ui->comboBoxSniffSSINV->setCurrentIndex(settings.value("SSINV", 0).toInt());
        settings.endGroup();
    }
    settings.beginGroup("SnifferUART");
    {
        ui->comboBoxBaud->setCurrentIndex(settings.value("Baud", 3).toInt());
        ui->comboBoxParity->setCurrentIndex(settings.value("Parity", 0).toInt());
        ui->comboBoxStopBits->setCurrentIndex(settings.value("Stop", 0).toInt());
        ui->comboBoxUARTData->setCurrentIndex(settings.value("Data", 3).toInt());
        ui->checkBoxSynchro->setChecked(settings.value("Synchro", false).toBool());
        settings.endGroup();
    }
    settings.beginGroup("SnifferParallel");
    {
        ui->comboBoxSniffParallelStrobe->setCurrentIndex(settings.value("Strobe", 0).toInt());
        ui->checkBoxSniffParallelInputInvert->setChecked(settings.value("Invert", false).toBool());
        settings.endGroup();
    }
    settings.beginGroup("Buffers");
    {
        ui->comboBoxDACBuffer->setCurrentIndex(settings.value("DAC", 0).toInt());
        ui->comboBoxADCBuffer->setCurrentIndex(settings.value("ADC", 0).toInt());
        settings.endGroup();
    }
    ui->tabWidget1->setCurrentIndex(settings.value("CurrentScopeTab",0).toInt());
    scopeInitDatacache();
}

void XlabInterface::deviceSettingsSave()
{
    tomap_PythonConsoleOptions->writeToSettings(tomap_PythonConsoleOptions->getName());
    tomap_ScopeOptions->writeToSettings(tomap_ScopeOptions->getName());
    tomap_MultimeterDisplayOptions->writeToSettings(tomap_MultimeterDisplayOptions->getName());
    QSettings settings;
    logToFile("Save device settings", LOGGING_LEVEL_Debug);
    settings.beginGroup("CH1");
    {
        settings.setValue("Gain", ui->ch1GainSlider->value());
        settings.setValue("Position", ui->ch1PositionSlider->value());
        settings.setValue("Trace", ui->checkBoxCH1Trace->isChecked());
        settings.setValue("AC", ui->checkBoxCH1ACDC->isChecked());
        settings.setValue("Average", ui->checkBoxCH1Average->isChecked());
        settings.setValue("Invert", ui->checkBoxCH1Invert->isChecked());
        settings.setValue("MathEn", ui->checkBoxCH1Math->isChecked());
        if (ui->radioButtonCH1Sub->isChecked()) settings.setValue("Math", "Sub");
        else if (ui->radioButtonCH1Multiply->isChecked()) settings.setValue("Math", "Multiply");
        settings.endGroup();
    }
    settings.beginGroup("CH2");
    {
        settings.setValue("Gain", ui->ch2GainSlider->value());
        settings.setValue("Position", ui->ch2PositionSlider->value());
        settings.setValue("Trace", ui->checkBoxCH2Trace->isChecked());
        settings.setValue("AC", ui->checkBoxCH2ACDC->isChecked());
        settings.setValue("Average", ui->checkBoxCH2Average->isChecked());
        settings.setValue("Invert", ui->checkBoxCH2Invert->isChecked());
        settings.setValue("MathEn", ui->checkBoxCH2Math->isChecked());
        if (ui->radioButtonCH2Sub->isChecked()) settings.setValue("Math", "Sub");
        else if (ui->radioButtonCH2Multiply->isChecked()) settings.setValue("Math", "Multiply");
        settings.endGroup();
    }
    settings.beginGroup("LOGIC");
    {
        settings.setValue("Size", ui->chdSizeSlider->value());
        settings.setValue("Position", ui->chdPositionSlider->value());
        settings.setValue("Trace", ui->checkBoxCHDTrace->isChecked());
        settings.setValue("Invert", ui->checkBoxCHDInvert->isChecked());
        settings.setValue("CHD0En", ui->checkBoxCHD0->isChecked());
        settings.setValue("CHD1En", ui->checkBoxCHD1->isChecked());
        settings.setValue("CHD2En", ui->checkBoxCHD2->isChecked());
        settings.setValue("CHD3En", ui->checkBoxCHD3->isChecked());
        settings.setValue("CHD4En", ui->checkBoxCHD4->isChecked());
        settings.setValue("CHD5En", ui->checkBoxCHD5->isChecked());
        settings.setValue("CHD6En", ui->checkBoxCHD6->isChecked());
        settings.setValue("CHD7En", ui->checkBoxCHD7->isChecked());
        settings.endGroup();
    }
    settings.beginGroup("FFT");
    {
        settings.setValue("Position", ui->verticalSliderFFTPosition->value());
        settings.setValue("Trace", ui->checkBoxFFTTrace->isChecked());
        settings.setValue("IQFFT", ui->checkBoxIQFFT->isChecked());
        settings.setValue("LogY", ui->checkBoxLogY->isChecked());
        settings.setValue("CH1", ui->checkBoxFFTCH1->isChecked());
        settings.setValue("CH2", ui->checkBoxFFTCH2->isChecked());
        if (ui->radioButtonRect->isChecked()) settings.setValue("Window", "Rect");
        else if (ui->radioButtonHamming->isChecked()) settings.setValue("Window", "Hamming");
        else if (ui->radioButtonHann->isChecked()) settings.setValue("Window", "Hann");
        else if (ui->radioButtonBlackman->isChecked()) settings.setValue("Window", "Blackman");
        settings.endGroup();
    }
    settings.beginGroup("Display");
    {
        settings.setValue("Intensity", ui->intensitySlider->value());
        settings.setValue("Decay", ui->decaySlider->value());
        settings.setValue("Vectors", ui->checkBoxVectors->isChecked());
        settings.setValue("Persistence", ui->checkBoxPersistence->isChecked());
        settings.setValue("MeterOn", ui->checkBoxmeterondev->isChecked());
        settings.endGroup();
    }
    settings.beginGroup("Cursors");
    {
        settings.setValue("Vertical", ui->checkBoxCursorVertical->isChecked());
        settings.setValue("Horizontal", ui->checkBoxScopeCursorsHorizontal->isChecked());
        if (ui->comboBoxScopeCursorsHorizontalType->currentIndex() == 2) settings.setValue("HType", "Manual");
        else if (ui->comboBoxScopeCursorsHorizontalType->currentIndex() == 1) settings.setValue("HType", "Automatic");
        else if (ui->comboBoxScopeCursorsHorizontalType->currentIndex() == 0) settings.setValue("HType", "Track");
        if (ui->comboBoxScopeCursorsChannel->currentIndex() == 0) settings.setValue("HChannel", "CH1");
        else if (ui->comboBoxScopeCursorsChannel->currentIndex() == 1) settings.setValue("HChannel", "CH2");
        settings.endGroup();
    }
    settings.beginGroup("Trigger");
    {
        settings.setValue("Source", ui->comboBoxTrigSource->currentIndex());
        settings.setValue("TAuto", ui->doubleSpinBoxTrigAuto->value());
        settings.setValue("THold", ui->doubleSpinBoxTrigHold->value());
        if (ui->radioButtonFree->isChecked()) settings.setValue("Type", "Free");
        else if (ui->radioButtonNormal->isChecked()) settings.setValue("Type", "Triger");
        else if (ui->radioButtonAuto->isChecked()) settings.setValue("Type", "Auto");
        else if (ui->radioButtonSingle->isChecked()) settings.setValue("Type", "Single");
        if (ui->radioButtonRisingEdge->isChecked()) settings.setValue("Mode", "Rising");
        else if (ui->radioButtonFallingEdge->isChecked()) settings.setValue("Mode", "Falling");
        else if (ui->radioButtonDual->isChecked()) settings.setValue("Mode", "Dual");
        else if (ui->radioButtonPositiveSlope->isChecked()) settings.setValue("Mode", "Positive");
        else if (ui->radioButtonNegativeSlope->isChecked()) settings.setValue("Mode", "Negative");
        else if (ui->radioButtonWindow->isChecked()) settings.setValue("Mode", "Window");
        settings.beginGroup("CH0Coord");
        {
            settings.setValue("TLevel", trigger.coord[0].posY);
            settings.setValue("TWin1Level", trigger.coord[0].posY1);
            settings.setValue("TWin2Level", trigger.coord[0].posY2);
            settings.setValue("TPost", trigger.coord[0].posX);
            settings.endGroup();
        }
        settings.beginGroup("CH1Coord");
        {
            settings.setValue("TLevel", trigger.coord[1].posY);
            settings.setValue("TWin1Level", trigger.coord[1].posY1);
            settings.setValue("TWin2Level", trigger.coord[1].posY2);
            settings.setValue("TPost", trigger.coord[1].posX);
            settings.endGroup();
        }
        settings.beginGroup("CH2Coord");
        {
            settings.setValue("TLevel", trigger.coord[2].posY);
            settings.setValue("TWin1Level", trigger.coord[2].posY1);
            settings.setValue("TWin2Level", trigger.coord[2].posY2);
            settings.setValue("TPost", trigger.coord[2].posX);
            settings.endGroup();
        }
        settings.endGroup();
    }
    settings.beginGroup("Sampling");
    {
        settings.setValue("Value", ui->samplingSlider->value());
        settings.setValue("XY", ui->xyMode->isChecked());
        settings.setValue("Roll", ui->rollMode->isChecked());
        settings.setValue("Elastic", ui->elasticMode->isChecked());
        settings.endGroup();
    }
    settings.beginGroup("Meter");
    {
        settings.setValue("Mode", ui->mmTabWidget->currentIndex());
        settings.endGroup();
    }
    settings.beginGroup("MeterFC");
    {
        if (ui->fcR11->isChecked()) settings.setValue("Source", "Schmitt");
        else if (ui->fcR10->isChecked()) settings.setValue("Source", "EXT");
        else if (ui->fcR7->isChecked()) settings.setValue("Source", "7");
        else if (ui->fcR6->isChecked()) settings.setValue("Source", "6");
        else if (ui->fcR5->isChecked()) settings.setValue("Source", "5");
        else if (ui->fcR4->isChecked()) settings.setValue("Source", "4");
        else if (ui->fcR3->isChecked()) settings.setValue("Source", "3");
        else if (ui->fcR2->isChecked()) settings.setValue("Source", "2");
        else if (ui->fcR1->isChecked()) settings.setValue("Source", "1");
        else if (ui->fcR0->isChecked()) settings.setValue("Source", "0");
        if (ui->fcR11_2->isChecked()) settings.setValue("Gate", "Schmitt");
        else if (ui->fcR10_2->isChecked()) settings.setValue("Gate", "EXT");
        else if (ui->fcR7_2->isChecked()) settings.setValue("Gate", "7");
        else if (ui->fcR6_2->isChecked()) settings.setValue("Gate", "6");
        else if (ui->fcR5_2->isChecked()) settings.setValue("Gate", "5");
        else if (ui->fcR4_2->isChecked()) settings.setValue("Gate", "4");
        else if (ui->fcR3_2->isChecked()) settings.setValue("Gate", "3");
        else if (ui->fcR2_2->isChecked()) settings.setValue("Gate", "2");
        else if (ui->fcR1_2->isChecked()) settings.setValue("Gate", "1");
        else if (ui->fcR0_2->isChecked()) settings.setValue("Gate", "0");
        if (ui->radioFreq->isChecked()) settings.setValue("Type", "Frequency");
        else if (ui->radioPulse->isChecked()) settings.setValue("Type", "CountPulses");
        else if (ui->radioPeriod->isChecked()) settings.setValue("Type", "Reriod");
        else if (ui->radioExtGate->isChecked()) settings.setValue("Type", "ExternalGate");
        settings.setValue("FastMode", ui->checkBoxFreqFastMode->isChecked());
        if (ui->radioFrequencyGatePeriod->isChecked()) settings.setValue("GateType", "Period");
        else if (ui->radioFrequencyGatePulse->isChecked()) settings.setValue("GateType", "Pulse");
        settings.endGroup();
    }
    settings.beginGroup("AWG");
    {
        settings.setValue("AWGEnable", ui->checkBoxAWBEnable->isChecked());
        if (ui->radioButtonSine->isChecked()) settings.setValue("Waveform", "Sine");
        else if (ui->radioButtonTriangle->isChecked()) settings.setValue("Waveform", "Triangle");
        else if (ui->radioButtonExpo->isChecked()) settings.setValue("Waveform", "Expo");
        else if (ui->radioButtonSquare->isChecked()) settings.setValue("Waveform", "Square");
        else if (ui->radioButtonWhiteNoise->isChecked()) settings.setValue("Waveform", "WhiteNoise");
        else if (ui->radioButtonPinkNoise->isChecked()) settings.setValue("Waveform", "PinkNoise");
        else if (ui->radioButtonConstant->isChecked()) settings.setValue("Waveform", "Constant");
        else if (ui->radioButtonCustom->isChecked()) settings.setValue("Waveform", "Custom");
        settings.setValue("ConstantValue", ui->spinBoxConstant->value());
        settings.setValue("FDigit100000", ui->labelFrequencyDigit100000->text());
        settings.setValue("FDigit10000", ui->labelFrequencyDigit10000->text());
        settings.setValue("FDigit1000", ui->labelFrequencyDigit1000->text());
        settings.setValue("FDigit100", ui->labelFrequencyDigit100->text());
        settings.setValue("FDigit10", ui->labelFrequencyDigit10->text());
        settings.setValue("FDigit1", ui->labelFrequencyDigit1->text());
        settings.setValue("FDigit01", ui->labelFrequencyDigit01->text());
        settings.setValue("FDigit001", ui->labelFrequencyDigit001->text());
        settings.setValue("Amplitude", ui->doubleSpinBoxAmp->value());
        settings.setValue("Duty", ui->doubleSpinBoxDuty->value());
        settings.setValue("Offset", ui->doubleSpinBoxOffset->value());
        settings.endGroup();
    }
    settings.beginGroup("AWGSweep");
    {
        settings.setValue("Speed", ui->sweepSpeedSlider->value());
        settings.setValue("Active", ui->collapsiblePanelAWGSweep->isChecked());
        if (ui->radioButtonSweepFrequency->isChecked()) settings.setValue("Type", "Frequency");
        else if (ui->radioButtonSweepAmplitude->isChecked()) settings.setValue("Type", "Amplitude");
        else if (ui->radioButtonSweepDutyCycle->isChecked()) settings.setValue("Type", "Duty");
        else if (ui->radioButtonSweepOffset->isChecked()) settings.setValue("Type", "Offset");
        if (ui->radioButtonSweepRising->isChecked()) settings.setValue("Direction", "Rising");
        else if (ui->radioButtonSweepPingPong->isChecked()) settings.setValue("Direction", "PingPong");
        else if (ui->radioButtonSweepFalling->isChecked()) settings.setValue("Direction", "Falling");
        settings.setValue("Start", ui->sweepStartFreqSlider->value());
        settings.setValue("End", ui->sweepEndFreqSlider->value());
        settings.endGroup();
    }
    settings.beginGroup("Sniffer");
    {
        if (ui->radioButtonSniffNormal->isChecked()) settings.setValue("Mode", "Normal");
        else if (ui->radioButtonSniffSingle->isChecked()) settings.setValue("Mode", "Single");
        settings.setValue("Circular", ui->checkBoxCircular->isChecked());
        settings.setValue("ASCII", ui->checkBoxASCII->isChecked());
        settings.setValue("Protocol", ui->protocolTabWidget->currentIndex());
        settings.endGroup();
    }
    settings.beginGroup("SnifferSPI");
    {
        settings.setValue("CPOL", ui->comboBoxCPOL->currentIndex());
        settings.setValue("CPHA", ui->comboBoxCPHA->currentIndex());
        settings.setValue("SSINV", ui->comboBoxSniffSSINV->currentIndex());
        settings.endGroup();
    }
    settings.beginGroup("SnifferUART");
    {
        settings.setValue("Baud", ui->comboBoxBaud->currentIndex());
        settings.setValue("Parity", ui->comboBoxParity->currentIndex());
        settings.setValue("Stop", ui->comboBoxStopBits->currentIndex());
        settings.setValue("Data", ui->comboBoxUARTData->currentIndex());
        settings.setValue("Synchro", ui->checkBoxSynchro->isChecked());
        settings.endGroup();
    }
    settings.beginGroup("SnifferParallel");
    {
        settings.setValue("Strobe", ui->comboBoxSniffParallelStrobe->currentIndex());
        settings.setValue("Invert", ui->checkBoxSniffParallelInputInvert->isChecked());
        settings.endGroup();
    }
    settings.beginGroup("Buffers");
    {
        settings.setValue("DAC", ui->comboBoxDACBuffer->currentIndex());
        settings.setValue("ADC", ui->comboBoxADCBuffer->currentIndex());
        settings.endGroup();
    }
    settings.setValue("CurrentScopeTab", ui->tabWidget1->currentIndex());
    settings.sync();
}

QString XlabInterface::saveWavetoString()
{
    QString out;
    out += "CH1,";
    for (int i = 0; i < ch1SaveBuffer.size(); i++) {
        out += QString::number(ch1SaveBuffer[i]);
        out += ",";
    }
    out += "<*>\n";
    out += "CH2,";
    for (int i = 0; i < ch2SaveBuffer.size(); i++) {
        out += QString::number(ch2SaveBuffer[i]);
        out += ",";
    }
    out += "<*>\n";
    out += "Bit0,";
    for (int i = 0; i < bitSaveBuffer[0].size(); i++) {
        out += QString::number(bitSaveBuffer[0].at(i));
        out += ",";
    }
    out += "<*>\n";
    out += "Bit1,";
    for (int i = 0; i < bitSaveBuffer[1].size(); i++) {
        out += QString::number(bitSaveBuffer[1].at(i));
        out += ",";
    }
    out += "<*>\n";
    out += "Bit2,";
    for (int i = 0; i < bitSaveBuffer[2].size(); i++) {
        out += QString::number(bitSaveBuffer[2].at(i));
        out += ",";
    }
    out += "<*>\n";
    out += "Bit3,";
    for (int i = 0; i < bitSaveBuffer[3].size(); i++)
    {
        out += QString::number(bitSaveBuffer[3].at(i));
        out += ",";
    }
    out += "<*>\n";
    out += "Bit4,";
    for (int i = 0; i < bitSaveBuffer[4].size(); i++) {
        out += QString::number(bitSaveBuffer[4].at(i));
        out += ",";
    }
    out += "<*>\n";
    out += "Bit5,";
    for (int i = 0; i < bitSaveBuffer[5].size(); i++) {
        out += QString::number(bitSaveBuffer[5].at(i));
        out += ",";
    }
    out += "<*>\n";
    out += "Bit6,";
    for (int i = 0; i < bitSaveBuffer[6].size(); i++) {
        out += QString::number(bitSaveBuffer[6].at(i));
        out += ",";
    }
    out += "<*>\n";
    out += "Bit7,";
    for (int i = 0; i < bitSaveBuffer[7].size(); i++) {
        out += QString::number(bitSaveBuffer[7].at(i));
        out += ",";
    }

    ch1SaveBuffer.clear();
    ch2SaveBuffer.clear();
    for (int k = 0; k < 8; k++) {
        bitSaveBuffer[k].clear();
    }
    out+="----------";
    return out;
}

void XlabInterface::saveWavetoFile()
{
    QString fileName = "waveFile" + QTime::currentTime().toString() + ".xsp";
    fileName.replace(":", "");
    if (filePath.isEmpty()) {
        filePath=QFileDialog::getExistingDirectory(this, tr("Choose Destination Folder"), defaultDir);
    }
    if (filePath.isEmpty())
        return;
    defaultDir = filePath;
    QFile waveData(filePath + QDir::separator() + fileName);
    if (waveData.open(QFile::WriteOnly | QFile::Truncate | QIODevice::Text)) {
        QTextStream out(&waveData);
        out << "CH1,";
        for (int i = 0; i < ch1SaveBuffer.size(); i++) {
            out << ch1SaveBuffer[i] << ",";
        }
        out << "<*>" << endl;
        out << "CH2,";
        for (int i=0; i < ch2SaveBuffer.size(); i++) {
            out << ch2SaveBuffer[i] << ",";
        }
        out << "<*>" << endl;
        out << "Bit0,";
        for (int i = 0; i < bitSaveBuffer[0].size(); i++) {
            out << bitSaveBuffer[0].at(i) << ",";
        }
        out << "<*>" << endl;
        out << "Bit1,";
        for (int i = 0; i < bitSaveBuffer[1].size(); i++) {
            out << bitSaveBuffer[1].at(i) << ",";
        }
        out << "<*>" << endl;
        out << "Bit2,";
        for (int i = 0; i < bitSaveBuffer[2].size(); i++) {
            out << bitSaveBuffer[2].at(i) << ",";
        }
        out << "<*>" << endl;
        out << "Bit3,";
        for (int i = 0; i < bitSaveBuffer[3].size(); i++) {
            out << bitSaveBuffer[3].at(i) << ",";
        }
        out << "<*>" << endl;
        out << "Bit4,";
        for (int i = 0; i < bitSaveBuffer[4].size(); i++) {
            out << bitSaveBuffer[4].at(i) << ",";
        }
        out << "<*>" << endl;
        out << "Bit5,";
        for (int i = 0; i < bitSaveBuffer[5].size(); i++) {
            out << bitSaveBuffer[5].at(i) << ",";
        }
        out << "<*>" << endl;
        out << "Bit6,";
        for (int i = 0; i < bitSaveBuffer[6].size(); i++) {
            out << bitSaveBuffer[6].at(i) << ",";
        }
        out << "<*>" << endl;
        out << "Bit7,";
        for (int i = 0; i < bitSaveBuffer[7].size(); i++) {
            out << bitSaveBuffer[7].at(i) << ",";
        }

    }
    waveData.close();
    ch1SaveBuffer.clear();
    ch2SaveBuffer.clear();
    for (int k = 0; k < 8; k++) {
        bitSaveBuffer[k].clear();
    }
    ui->notifLabel->setText("File " + fileName + " created!");
}

void XlabInterface::wavePlay(QString file)
{
    loadedWave.clear();
    loadedFrequency.clear();
    QString buffer;
    QFile waveData(file);
    if (waveData.open(QFile::ReadOnly)) {
        freqBeforeLoaded=ui->samplingSlider->value();
        buffer = waveData.readAll();
        while(!buffer.isEmpty()) {
            QStringList rawParsedData = buffer.split("<*>");
            rawParsedData.removeAll("\n");
            rawParsedData.removeAll("\r");
            int freq=rawParsedData[0].toInt();
            loadedFrequency.push_back(freq);
            rawParsedData.removeFirst();
            QStringList parsedData = rawParsedData[0].split(",");
            QVector<double> tmp;
            for (int i = 1; i < parsedData.length() - 1; i++) {
                tmp.push_back(parsedData[i].toDouble());
            }
            loadedWave.push_back(tmp);
            parsedData.clear();
            parsedData = rawParsedData[1].split(",");
            QVector<double> tmp2;
            for (int i = 1; i < parsedData.length() - 1; i++) {
                tmp2.push_back(parsedData[i].toDouble());
            }
            loadedWave.push_back(tmp2);
            QVector<double> tmp3[8];
            for (int k = 0; k < 8; k++)
            {
                parsedData.clear();
                parsedData = rawParsedData[k + 2].split(",");
                for (int i = 1; i < parsedData.length() - 1; i++) {
                    tmp3[k].push_back(parsedData[i].toDouble());
                }
                loadedWave.push_back(tmp3[k]);
            }
            buffer = buffer.mid(buffer.indexOf("----------") + 10);
        }
    }
    ui->label_13->setText("Load recorded wave");
    loadRecordedWave = true;
}

void XlabInterface::deviceInfoRead(QString configFile)
{
    QXmlGet xmlHardware;

    unsigned char health[1];
    clearBuffer(health, 1);
    int size = usbDevice.controlReadTransfer(USB_REQUEST_DATA, health, 1, MAIN_DATA_REQUEST_DeviceStatus);
    if (size != 1) return;

    xmlHardware.load(configFile);
    if (xmlHardware.docType() == "XLabHardware") {
        while (xmlHardware.findNext("info")) {
            xmlHardware.descend();
            while (xmlHardware.findNext("description"))
                ui->textBrowserDeviceInfo->append("    "+xmlHardware.getString());
        }
        xmlHardware.rise();
        while (xmlHardware.findNext("health")) {
            xmlHardware.descend();
            ui->textBrowserDeviceInfo->append("Health status");
            while (xmlHardware.findNext("status")) {
                if ((*health & (byte)(1 << (xmlHardware.getAttributeInt("code")) )) == 0)
                    ui->textBrowserDeviceInfo->append("    "+xmlHardware.getString()+": Problem!");
                else
                    ui->textBrowserDeviceInfo->append("    "+xmlHardware.getString()+": OK");
            }
        }
        xmlHardware.rise();
    }
}

void XlabInterface::scopeStartForce()
{
    sendScopeStart();
}

void XlabInterface::parseCSV(QString csvString, byte *buffer)
{
    int i = 0;
    int max;

    max = csvString.length();
    while(i < max) {
        if (csvString[i] == '\n' || csvString[i] == '\r') csvString[i] = ' ';
        i++;
    }
    QStringList values = csvString.split(',');
    values.removeAll("");
    max = values.length();
    i = 0;
    if (max > AWG_WAVE_TEMPLATE_LENGTH) max = AWG_WAVE_TEMPLATE_LENGTH;
    while(i < max) {
        *(int16_t *)(buffer) = (int16_t)(values[i].trimmed().toInt());
        buffer++;
        buffer++;
        i++;
    }
}

void XlabInterface::sniffDataShow()
{
    sniffconfig_t sniffLogicCfg;
    sniffdata_t *sniffLogicData;

    if (usbDevice.dataLength < (SNIFFER_BUFFER_LENGTH + SNIFFER_CONFIG_LENGTH)) {
        return;
    }
    dataSnifferHeaderGetData(&sniffLogicCfg, usbDevice.chData);
    if (sniffLogicCfg.type != FRAME_HEADER_NAME_Sniffer) {
        logToFile("Sniffer, frame header not valid", LOGGING_LEVEL_Warning);
        return;
    }
    sniffLogicData = (sniffdata_t *)(usbDevice.chData + SNIFFER_CONFIG_LENGTH);
    if (lastFrameSniff == sniffLogicCfg.id) {
        logToFile("Sniffer, same data, nothing to do", LOGGING_LEVEL_Info);
        return;
    }
    lastFrameSniff = sniffLogicCfg.id;
    if (ui->pauseSnifferButton->isChecked()) return;
    if (ui->checkBoxCircular->isChecked()) {
        if (sniffbufferfulfillrx <= sniffLogicCfg.indexrx) {
            sniffbufferfulfillrx = sniffLogicCfg.indexrx;
        } else {
            sniffbufferfulfillrx = ui->progressBarSnifferBufferFill->maximum();
        }
        if (sniffbufferfulfilltx <= sniffLogicCfg.indextx) {
            sniffbufferfulfilltx = sniffLogicCfg.indextx;
        } else {
            sniffbufferfulfilltx = ui->progressBarSnifferBufferFill->maximum();
        }
        if (sniffbufferfulfillrx > sniffLogicCfg.indexrx) sniffbufferfirstrxloop = false;
        if (sniffbufferfulfilltx > sniffLogicCfg.indextx) sniffbufferfirsttxloop = false;
        if (sniffbufferfulfillrx > sniffbufferfulfilltx) {
            sniffbufferfulfill = sniffbufferfulfillrx;
        } else {
            sniffbufferfulfill = sniffbufferfulfilltx;
        }
    } else {
        sniffbufferfirstrxloop = true;
        sniffbufferfirsttxloop = true;
        if (sniffLogicCfg.indexrx > sniffLogicCfg.indextx) {
            sniffbufferfulfill = sniffLogicCfg.indexrx;
        } else {
            sniffbufferfulfill = sniffLogicCfg.indextx;
        }
    }
    if (sniffLogicCfg.protocol == SNIFFER_PROTOCOL_USART) {
        ui->progressBarSnifferBufferFill->setMaximum(SNIFFER_USART_BUFFER_LENGTH - 1);
        sniffProtocolSerialPrint(sniffLogicData->serial.Rx, sniffLogicData->serial.Dx, SNIFFER_VALID_Rx_DATA_Bit, sniffLogicCfg.indexrx, ui->rxTextEdit, sniffbufferfirstrxloop);
        sniffProtocolSerialPrint(sniffLogicData->serial.Tx, sniffLogicData->serial.Dx, SNIFFER_VALID_Tx_DATA_Bit, sniffLogicCfg.indextx, ui->txTextEdit, sniffbufferfirsttxloop);
    } else if (sniffLogicCfg.protocol == SNIFFER_PROTOCOL_SPI) {
        ui->progressBarSnifferBufferFill->setMaximum(SNIFFER_SPI_BUFFER_LENGTH - 1);
        sniffProtocolSPIPrint(sniffLogicData->spi.MOSI, sniffLogicCfg.indexrx, ui->mosiTextEdit, sniffbufferfirstrxloop);
        sniffProtocolSPIPrint(sniffLogicData->spi.MISO, sniffLogicCfg.indextx, ui->misoTextEdit, sniffbufferfirsttxloop);
    } else if (sniffLogicCfg.protocol == SNIFFER_PROTOCOL_I2C) {
        ui->progressBarSnifferBufferFill->setMaximum(SNIFFER_I2C_BUFFER_LENGTH - 1);
        sniffProtocolI2CPrint(sniffLogicData->i2c.decoded, sniffLogicData->i2c.addr_ack, sniffLogicCfg.indexrx, ui->i2cPlainTextEdit);
    } else if (sniffLogicCfg.protocol == SNIFFER_PROTOCOL_Parallel) {
        ui->progressBarSnifferBufferFill->setMaximum(SNIFFER_PARALLEL_BUFFER_LENGTH - 1);
        sniffProtocolParallelPrint(sniffLogicData->parallel.data, sniffLogicCfg.indexrx, ui->plainTextEditParallel, sniffbufferfirstrxloop);
    }
    ui->progressBarSnifferBufferFill->setValue(sniffbufferfulfill);
}

void XlabInterface::sniffStop()
{
    sendSnifferStop();
    sniffControlsStartSet(BUTTON_STOP_SNIFFER_TEXT_STOP);
}

void XlabInterface::sniffControlsStartSet(QString command)
{
    if (command == BUTTON_STOP_SNIFFER_TEXT_START) {
        sniffProtocolWidgetClear();
        sniffControlsEnable(false);
        ui->startSnifferButton->setText(tr(BUTTON_STOP_SNIFFER_TEXT_STOP));
        ui->pauseSnifferButton->setEnabled(true);
        ui->pauseSnifferButton->setChecked(false);
    } else {
        sniffControlsEnable(true);
        ui->startSnifferButton->setText(tr(BUTTON_STOP_SNIFFER_TEXT_START));
        ui->pauseSnifferButton->setEnabled(false);
        ui->pauseSnifferButton->setChecked(false);
    }
    sniffPauseButtonChanged();
}

void XlabInterface::sniffProtocolWidgetClear()
{
    ui->rxTextEdit->clear();
    ui->txTextEdit->clear();
    ui->misoTextEdit->clear();
    ui->mosiTextEdit->clear();
    ui->i2cPlainTextEdit->clear();
    ui->plainTextEditParallel->clear();
    ui->progressBarSnifferBufferFill->setValue(0);
}

void XlabInterface::sniffControlsEnable(bool value)
{
    ui->comboBoxBaud->setEnabled(value);
    ui->comboBoxStopBits->setEnabled(value);
    ui->comboBoxParity->setEnabled(value);
    ui->comboBoxCPHA->setEnabled(value);
    ui->comboBoxCPOL->setEnabled(value);
    ui->modeGroupBox->setEnabled(value);
    ui->comboBoxUARTData->setEnabled(value);
    ui->checkBoxSynchro->setEnabled(value);
    ui->comboBoxSniffSSINV->setEnabled(value);
    ui->checkBoxSniffParallelInputInvert->setEnabled(value);
    ui->comboBoxSniffParallelStrobe->setEnabled(value);
}

void XlabInterface::sniffProtocolSerialPrint(uint8_t *buffer, uint8_t *validbuffer, uint8_t validator, uint16_t idx, QPlainTextEdit *outwidget, bool firstloop)
{
    const QString noASCII = "_";
    const QString noData = ".";
    int j, max;
    uint16_t i, n;
    byte data;
    QString datastring;
    QByteArray databytes;

    if ((ui->checkBoxCircular->isChecked()) && !(firstloop)) {
        i = idx;
        max = SNIFFER_USART_BUFFER_LENGTH;
        if (i >= SNIFFER_USART_BUFFER_LENGTH) i -= SNIFFER_USART_BUFFER_LENGTH;
    } else {
        i = 0;
        max = idx;
        if (max > SNIFFER_USART_BUFFER_LENGTH) max = SNIFFER_USART_BUFFER_LENGTH;
    }
    for (n = i, j = 0; j < max; n++, j++) {
        if (n >= SNIFFER_USART_BUFFER_LENGTH) n = 0;
        data = *(buffer + n);
        if (((*(validbuffer + n) & validator)) || !(ui->checkBoxSynchro->isChecked())) {
            if (ui->checkBoxASCII->isChecked()) {
                if (data < 0x20) {
                    datastring.append(noASCII);
                } else {
                    datastring.append(QChar::fromLatin1(data));
                }
                datastring.append(" ");
            } else {
                databytes.append(data);
                datastring.append(databytes.toHex());
                datastring.append(" ");
                databytes.clear();
            }
        } else {
            if (ui->checkBoxASCII->isChecked()) {
                datastring.append(noData);
                datastring.append(" ");
            } else {
                datastring.append(noData);
                datastring.append(noData);
                datastring.append(" ");
            }
        }
    }
    outwidget->setPlainText(datastring);
    outwidget->verticalScrollBar()->setValue(outwidget->verticalScrollBar()->maximum());
}

void XlabInterface::sniffProtocolSPIPrint(uint8_t *buffer, uint16_t idx, QPlainTextEdit *outwidget, bool firstloop)
{
    const QString noASCII = "_";
    int j, max;
    uint16_t i, n;
    byte data;
    QString datastring;
    QByteArray databytes;

    if ((ui->checkBoxCircular->isChecked()) && !(firstloop)) {
        i = idx;
        max = SNIFFER_SPI_BUFFER_LENGTH;
        if (i >= SNIFFER_SPI_BUFFER_LENGTH) i -= SNIFFER_SPI_BUFFER_LENGTH;
    } else {
        i = 0;
        max = idx;
        if (max > SNIFFER_SPI_BUFFER_LENGTH) max = SNIFFER_SPI_BUFFER_LENGTH;
    }
    for (n = i, j = 0; j < max; n++, j++) {
        if (n >= SNIFFER_SPI_BUFFER_LENGTH) n = 0;
        data = *(buffer + n);
        if (ui->checkBoxASCII->isChecked()) {
            if (data < 0x20) {
                datastring.append(noASCII);
            } else {
                datastring.append(QChar::fromLatin1(data));
            }
            datastring.append(" ");
        } else {
            databytes.append(data);
            datastring.append(databytes.toHex());
            datastring.append(" ");
            databytes.clear();
        }
    }
    outwidget->setPlainText(datastring);
    outwidget->verticalScrollBar()->setValue(outwidget->verticalScrollBar()->maximum());
}

void XlabInterface::sniffProtocolI2CPrint(uint8_t *buffer, uint8_t *ackbuffer, uint16_t idx, QPlainTextEdit *outwidget)
{
    const uint8_t i2c_ReadBitMask = 0b00000001;
    const QString i2c_ACK_String = "<A>";
    const QString i2c_NOACK_String = "<" + QString(QChar(0x1FB9)) + ">";
    const QString i2c_ADDRESS_String = "<Address:>";
    const QString i2c_WRITE_String = "<W>";
    const QString i2c_READ_String = "<R>";
    const QString i2c_START_String = "<Start>";
    const QString i2c_REPEATSTART_String = "<Restart>";
    const QString i2c_TIMEOUT_String = "<Timeout>";
    const QString i2c_FASTSTOP_String = "<Stop/Timeout>";
    const QString i2c_STOP_String = "<Stop>";
    const QString noASCII = "_";
    byte data;
    QByteArray bdata;
    QString i2cData;
    int max, j;
    uint16_t i, n;
    uint8_t ack, address;

    if (ui->checkBoxCircular->isChecked()) {
        i = idx;
        max = SNIFFER_I2C_BUFFER_LENGTH;
        if (i >= SNIFFER_I2C_BUFFER_LENGTH) i -= SNIFFER_I2C_BUFFER_LENGTH;
    } else {
        i = 0;
        max = idx;
        if (max > SNIFFER_I2C_BUFFER_LENGTH) max = SNIFFER_I2C_BUFFER_LENGTH;
    }
    for (n = i, j = 0; j < max; n++, j++) {
        if (n >= SNIFFER_I2C_BUFFER_LENGTH) n = 0;
        data = *(buffer + n);
        address = *(ackbuffer + n) & SNIFFER_I2C_Address;
        ack = *(ackbuffer + n) & SNIFFER_I2C_ACK;

        if (!(*(ackbuffer + n) & SNIFFER_I2C_ValidData)) {
            continue;
        }
        if (*(ackbuffer + n) & SNIFFER_I2C_Start) {
            i2cData.append(i2c_START_String);
            i2cData.append(" ");
            continue;
        }
        if (*(ackbuffer + n) & SNIFFER_I2C_RepeatStart) {
            i2cData.append("\n");
            i2cData.append(i2c_REPEATSTART_String);
            i2cData.append(" ");
            continue;
        }
        if (*(ackbuffer + n) & SNIFFER_I2C_Stop) {
            i2cData.append(i2c_STOP_String);
            i2cData.append("\n");
            continue;
        }
        if (*(ackbuffer + n) & SNIFFER_I2C_FastStop) {
            i2cData.append(i2c_FASTSTOP_String);
            i2cData.append("\n");
            continue;
        }
        if (*(ackbuffer + n) & SNIFFER_I2C_Timeout) {
            i2cData.append(i2c_TIMEOUT_String);
            i2cData.append("\n");
            continue;
        }
        if (address) {
            i2cData.append(i2c_ADDRESS_String);
            bdata.append(data >> 1);
            i2cData.append(bdata.toHex());
            bdata.clear();
            if ((data & i2c_ReadBitMask) != 0) {
                i2cData.append(i2c_READ_String);
            } else {
                i2cData.append(i2c_WRITE_String);
            }
        } else {
            if (ui->checkBoxASCII->isChecked()) {
                if (data < 0x20) {
                    i2cData.append(noASCII);
                } else {
                    i2cData.append(QChar::fromLatin1(data));
                }
            } else {
                bdata.append(data);
                i2cData.append(bdata.toHex());
                bdata.clear();
            }
        }
        if (ack) {
            i2cData.append(i2c_ACK_String);
        } else {
            i2cData.append(i2c_NOACK_String);
        }
        i2cData.append(" ");
    }
    outwidget->setPlainText((QString)i2cData);
    outwidget->verticalScrollBar()->setValue(outwidget->verticalScrollBar()->maximum());
}

void XlabInterface::sniffProtocolParallelPrint(uint8_t *buffer, uint16_t idx, QPlainTextEdit *outwidget, bool firstloop)
{
    const QString noASCII = "_";
    int j, max;
    uint16_t i, n;
    byte data;
    QString datastring;
    QByteArray databytes;

    if ((ui->checkBoxCircular->isChecked()) && !(firstloop)) {
        i = idx;
        max = SNIFFER_PARALLEL_BUFFER_LENGTH;
        if (i >= SNIFFER_PARALLEL_BUFFER_LENGTH) i -= SNIFFER_PARALLEL_BUFFER_LENGTH;
    } else {
        i = 0;
        max = idx;
        if (max > SNIFFER_PARALLEL_BUFFER_LENGTH) max = SNIFFER_PARALLEL_BUFFER_LENGTH;
    }
    for (n = i, j = 0; j < max; n++, j++) {
        if (n >= SNIFFER_PARALLEL_BUFFER_LENGTH) n = 0;
        data = *(buffer + n);
        if (ui->checkBoxASCII->isChecked()) {
            if (data < 0x20) {
                datastring.append(noASCII);
            } else {
                datastring.append(QChar::fromLatin1(data));
            }
            datastring.append(" ");
        } else {
            databytes.append(data);
            datastring.append(databytes.toHex());
            datastring.append(" ");
            databytes.clear();
        }
    }
    outwidget->setPlainText(datastring);
    outwidget->verticalScrollBar()->setValue(outwidget->verticalScrollBar()->maximum());
}

void XlabInterface::channelSetBoxTitle(QGroupBox *boxac, QGroupBox *boxdc, int ch, int gain, bool ac)
{
    QString title;

    title = "Channel " + QString::number(ch) + "  " + gainSliderText[gain];
    if (ac) title += ", AC ";
    else title += ", AC/DC ";
    boxac->setTitle(title);
    boxdc->setTitle(title);
}

void XlabInterface::channel1DCButtonCheck(int gain, bool ac)
{
    if (ac) {
        ui->toolButtonCh1ADC2V4->setChecked(false);
        ui->toolButtonCh1ADC24V->setChecked(false);
        return;
    }
    if (gain == 0) {
        ui->toolButtonCh1ADC2V4->setChecked(false);
        ui->toolButtonCh1ADC24V->setChecked(true);
    } else if (gain == 3) {
        ui->toolButtonCh1ADC2V4->setChecked(true);
        ui->toolButtonCh1ADC24V->setChecked(false);
    } else {
        ui->toolButtonCh1ADC2V4->setChecked(false);
        ui->toolButtonCh1ADC24V->setChecked(false);
    }
}

void XlabInterface::channel2DCButtonCheck(int gain, bool ac)
{
    if (ac) {
        ui->toolButtonCh2ADC2V4->setChecked(false);
        ui->toolButtonCh2ADC24V->setChecked(false);
        return;
    }
    if (gain == 0) {
        ui->toolButtonCh2ADC2V4->setChecked(false);
        ui->toolButtonCh2ADC24V->setChecked(true);
    } else if (gain == 3) {
        ui->toolButtonCh2ADC2V4->setChecked(true);
        ui->toolButtonCh2ADC24V->setChecked(false);
    } else {
        ui->toolButtonCh2ADC2V4->setChecked(false);
        ui->toolButtonCh2ADC24V->setChecked(false);
    }
}

void XlabInterface::channel1GainSliderChanged(int value, bool press)
{
    ui->labelCH1Gain->setText(gainSliderText[value]);
    channelSetBoxTitle(ui->groupBoxCH1AC, ui->groupBoxCH1DC, 1, value, ui->checkBoxCH1ACDC->isChecked());
    if (press) {
        channel1DCButtonCheck(value, ui->checkBoxCH1ACDC->isChecked());
    }
    sendChannel1GainControls(value);
}

void XlabInterface::channel2GainSliderChanged(int value, bool press)
{
    ui->labelCH2Gain->setText(gainSliderText[value]);
    channelSetBoxTitle(ui->groupBoxCH2AC, ui->groupBoxCH2DC, 2, value, ui->checkBoxCH2ACDC->isChecked());
    if (press) {
        channel2DCButtonCheck(value, ui->checkBoxCH2ACDC->isChecked());
    }
    sendChannel2GainControls(value);
}

void XlabInterface::channel1ACDCChanged(bool press)
{
    channelSetBoxTitle(ui->groupBoxCH1AC, ui->groupBoxCH1DC, 1, ui->ch1GainSlider->value(), ui->checkBoxCH1ACDC->isChecked());
    if (press) {
        channel1DCButtonCheck(ui->ch1GainSlider->value(), ui->checkBoxCH1ACDC->isChecked());
    }
    sendCH1Controls();
}

void XlabInterface::channel2ACDCChanged(bool press)
{
    channelSetBoxTitle(ui->groupBoxCH2AC, ui->groupBoxCH2DC, 2, ui->ch2GainSlider->value(), ui->checkBoxCH2ACDC->isChecked());
    if (press) {
        channel2DCButtonCheck(ui->ch2GainSlider->value(), ui->checkBoxCH2ACDC->isChecked());
    }
    sendCH2Controls();
}

void XlabInterface::channelPositionSliderChanged(int channel, int value, int *chzeropos, bool enabled)
{
    double tmp_pos = *chzeropos;
    *chzeropos = yAxisMax / 2 + value;
    if (enabled) {
        hCursorAHead->topLeft->setPixelPosition(QPointF(14, (ui->plotterWidget->yAxis->coordToPixel(hCursorAHead->topLeft->coords().y() + *chzeropos - tmp_pos))));
        hCursorBHead->topLeft->setPixelPosition(QPointF(14, (ui->plotterWidget->yAxis->coordToPixel(hCursorBHead->topLeft->coords().y() + *chzeropos - tmp_pos))));
    }
    zeroHeadRefresh();
    triggerSetChannelOffset(channel, value);
    triggerRefreshPos();
    if ((ui->comboBoxTrigSource->currentIndex() <= 1)) {
        triggerMoveTrigger(triggerPosXIntToPixel(trigger.coord[trigger.channel].posX), triggerPosYIntToPixel(trigger.coord[trigger.channel].posY));
        triggerMoveWinTrigger(triggerPosXIntToPixel(trigger.coord[trigger.channel].posX), triggerPosYIntToPixel(trigger.coord[trigger.channel].posY1), triggerPosYIntToPixel(trigger.coord[trigger.channel].posY2));
    }
    m_repaint = true;
}

void XlabInterface::channel1TraceChanged(bool checked)
{
    unsigned char result[2];

    clearBuffer(result, 2);
    if (sendCH1Controls(result, 2) < 0) return;
    if (result[0] == MAIN_CONFIGURATION_RESULT_NoEnoughMemory) {
        logToFile("No enough memory for CH buffer: " + QString::number(result[1]), LOGGING_LEVEL_Debug);
        scopeADCBufferChanged(result[1] - 1);
        if (sendCH1Controls(result, 2) != 0) return;
    }
    ui->groupBoxCH1AC->setChecked(checked);
    ui->groupBoxCH1DC->setChecked(checked);
    bufferCheckMaxSize();
    m_repaint = true;
}

void XlabInterface::channel2TraceChanged(bool checked)
{
    unsigned char result[2];

    clearBuffer(result, 2);
    if (sendCH2Controls(result, 2) < 0) return;
    if (result[0] == MAIN_CONFIGURATION_RESULT_NoEnoughMemory) {
        logToFile("No enough memory for CH buffer: " + QString::number(result[1]), LOGGING_LEVEL_Debug);
        scopeADCBufferChanged(result[1] - 1);
        if (sendCH2Controls(result, 2) != 0) return;
    }
    ui->groupBoxCH2AC->setChecked(checked);
    ui->groupBoxCH2DC->setChecked(checked);
    bufferCheckMaxSize();
    m_repaint = true;
}

void XlabInterface::channelLogicSizeSliderChanged()
{
    triggerSetChannelSize(MAIN_CHANNEL_CHD, ui->chdSizeSlider->value());
    triggerRefreshPos();
    m_repaint = true;
}

void XlabInterface::channelLogicPositionSliderChanged(int value)
{
    triggerSetChannelOffset(MAIN_CHANNEL_CHD, ui->chdPositionSlider->value());
    triggerRefreshPos();
    chdZeroPos = yAxisMax / 2 + value;
    zeroHeadRefresh();  
    m_repaint = true;
}

void XlabInterface::channelLogicTraceChanged()
{
    unsigned char result[2];

    clearBuffer(result, 2);
    if (sendCHDControls(result, 2) < 0) return;
    if (result[0] == MAIN_CONFIGURATION_RESULT_NoEnoughMemory) {
        logToFile("No enough memory for CH buffer: " + QString::number(result[1]), LOGGING_LEVEL_Debug);
        scopeADCBufferChanged(result[1] - 1);
        if (sendCHDControls(result, 2) != 0) return;
    }
    bufferCheckMaxSize();
    m_repaint = true;
}

void XlabInterface::displayIntensitySliderChanged(int value)
{
    qreal h, s, lmaxCh1, lmaxCh2;
    ch1Pen.color().getHslF(&h, &s, &lmaxCh1);
    ch2Pen.color().getHslF(&h, &s, &lmaxCh2);
    qreal h1, h2, s1, s2, l1, l2;
    QPen t1Pen, t2Pen;

    for (int i = 0; i < TG; i++) {
        t1Pen = ch1PPen[i];
        t1Pen.color().getHslF(&h1, &s1, &l1);
        t2Pen = ch2PPen[i];
        t2Pen.color().getHslF(&h2, &s2, &l2);
        if (tomap_ScopeOptions->theme == DISPLAY_THEME_Dark) {
            t1Pen.setColor(QColor::fromHslF(h1, s1, mapRangeF(((i + 1.0) * (value / 10.0)), i + 1.0, 0, lmaxCh1, 0.05)));
            t2Pen.setColor(QColor::fromHslF(h2, s2, mapRangeF(((i + 1.0) * (value / 10.0)), i + 1.0, 0, lmaxCh2, 0.05)));
            ch1PGraphs[i]->setPen(t1Pen);
            ch2PGraphs[i]->setPen(t2Pen);
        } else if (tomap_ScopeOptions->theme == DISPLAY_THEME_Light) {
            t1Pen.setColor(QColor::fromHslF(h1, s1, 1.0 - mapRangeF(((i + 1.0) * (value / 10.0)), i + 1.0, 0, lmaxCh1, 0.05)));
            t2Pen.setColor(QColor::fromHslF(h2, s2, 1.0 - mapRangeF(((i + 1.0) * (value / 10.0)), i + 1.0, 0, lmaxCh2, 0.05)));
            ch1PGraphs[i]->setPen(t1Pen);
            ch2PGraphs[i]->setPen(t2Pen);
        } else {
            if (customThemeDialog.idealForegroundColor(customThemeDialog.customColors.background.color())=="#fffff") {
                t1Pen.setColor(QColor::fromHslF(h1, s1, 1.0 - mapRangeF(((i + 1.0) * (value / 10.0)), i + 1.0, 0, lmaxCh1, 0.05)));
                t2Pen.setColor(QColor::fromHslF(h2, s2, 1.0 - mapRangeF(((i + 1.0) * (value / 10.0)), i + 1.0, 0, lmaxCh2, 0.05)));
                ch1PGraphs[i]->setPen(t1Pen);
                ch2PGraphs[i]->setPen(t2Pen);
            } else {
                t1Pen.setColor(QColor::fromHslF(h1, s1, mapRangeF(((i + 1.0) * (value / 10.0)), i + 1.0, 0, lmaxCh1, 0.05)));
                t2Pen.setColor(QColor::fromHslF(h2, s2, mapRangeF(((i + 1.0) * (value / 10.0)), i + 1.0, 0, lmaxCh2, 0.05)));
                ch1PGraphs[i]->setPen(t1Pen);
                ch2PGraphs[i]->setPen(t2Pen);
            }
        }
    }
}

void XlabInterface::displayVectorsChanged(bool checked) {
    setupScatterStyles(checked);
    displayControlsSet();
}

void XlabInterface::displayThemeChanged(int theme)
{
    if (initializing) return;
    if (theme == DISPLAY_THEME_Custom){
        customThemeDialog.loadTheme();
        customThemeDialog.show();
    } else setTheme(theme);
}

void XlabInterface::displayGridStyleChanged(int index)
{
    if (index == 0)      gridPenStyle = Qt::NoPen;
    else if (index == 1) gridPenStyle = Qt::DotLine;
    else if (index == 2) gridPenStyle = Qt::SolidLine;
    else if (index == 3) gridPenStyle = Qt::SolidLine;
    gridBoxSet(gridPenColor, gridPenStyle);
}

void XlabInterface::cursorsVerticalChanged(bool checked)
{
    if (ui->xyMode->isChecked()) vCursorSetVisible(false);
    else vCursorSetVisible(checked);
    vCursorRefresh();
    m_repaint = true;
}

void XlabInterface::cursorsHorizontalChanged(bool checked)
{
    if (ui->xyMode->isChecked() && ui->comboBoxScopeCursorsChannel->currentIndex() == 0) hCursorSetVisible(false);
    else hCursorSetVisible(checked);
    m_repaint = true;
}

void XlabInterface::cursorsHorizontalTypeSet(int index)
{
    if (index == 0) {
        if (!(ui->checkBoxCursorVertical->isChecked())) {
            ui->checkBoxCursorVertical->setChecked(true);
            cursorsVerticalChanged(ui->checkBoxCursorVertical->isChecked());
        }
    }
    bool cursorTrack = false;
    if(index == 0) cursorTrack = true;
    hCursorSetTracerVisible(cursorTrack);
    m_repaint = true;
}

void XlabInterface::cursorsSetChannelActive(int index)
{
    QColor color;

    cursorsHorizontalChanged(ui->checkBoxScopeCursorsHorizontal->isChecked());
    if (index == 0) {
        hCursor.channel = MAIN_CHANNEL_CH1;
        hCursorAHead->topLeft->setPixelPosition(QPointF(14, (ui->plotterWidget->yAxis->coordToPixel(hCursorAPosCh1 + ch1ZeroPos)) - 8));
        hCursorBHead->topLeft->setPixelPosition(QPointF(14, (ui->plotterWidget->yAxis->coordToPixel(hCursorBPosCh1 + ch1ZeroPos)) - 8));
        color = ch1Pen.color();
    } else if (index == 1) {
        hCursor.channel = MAIN_CHANNEL_CH2;
        hCursorAHead->topLeft->setPixelPosition(QPointF(14, (ui->plotterWidget->yAxis->coordToPixel(hCursorAPosCh2 + ch2ZeroPos)) - 8));
        hCursorBHead->topLeft->setPixelPosition(QPointF(14, (ui->plotterWidget->yAxis->coordToPixel(hCursorBPosCh2 + ch2ZeroPos)) - 8));
        color = ch2Pen.color();
    }
    hCursorRefreshVisible();
    vCursorSetLabelColor(color);
    hCursorSetLabelColor(color);
    bool cursorTrack = false;
    if(ui->comboBoxScopeCursorsHorizontalType->currentIndex() == 0) cursorTrack = true;
    hCursorSetTracerVisible(cursorTrack);
    m_repaint = true;
}

void XlabInterface::scopeRollModeChanged()
{
    sendMixedControls();
    if (ui->rollMode->isChecked()) ui->elasticMode->setEnabled(false);
    else ui->elasticMode->setEnabled(true);
    m_repaint = true;
}

void XlabInterface::scopeElasticModeChanged()
{
    sendMixedControls();
    m_repaint = true;
}

void XlabInterface::scopeRefreshSamplingRateLabel()
{
    QString unit;
    int sr;

    sr = ui->samplingSlider->value();
    ui->labelSamplingRate->setText("Time = " + rateText[sr] + ", time range = " +
            get4Digits(vCursorDeltaTime(0, 2 * DATA_BUFFER_BLOCK_SIZE * (ui->comboBoxADCBuffer->currentIndex() + 1), &unit, 1, sr)) + unit);
}

void XlabInterface::scopeSamplingSliderChanged(int value)
{
    if (value >= 11) {
        ui->rollMode->setEnabled(true);
        if (ui->rollMode->isChecked())
            ui->elasticMode->setEnabled(false);
        else
            ui->elasticMode->setEnabled(true);
        triggerSetSlowSamplingRate(true);
    } else {
        ui->rollMode->setEnabled(false);
        ui->elasticMode->setEnabled(true);
        triggerSetSlowSamplingRate(false);
    }
    if (!usbDevice.isDeviceConnected) return;
    triggerRefreshPos();
    scopeRefreshSamplingRateLabel();
    sendSamplingRate(value);
}

void XlabInterface::meterFrequencyModeChanged()
{
    clearMMLabelsFreq();
    if (ui->radioFreq->isChecked()) {
        ui->groupBoxSourceGate->setEnabled(false);
        ui->multimeterdisplayDutyValue->setVisible(false);
        ui->multimeterdisplayDutyValue_s->setVisible(false);
        ui->multimeterdisplayPeriodValue->setVisible(true);
        ui->multimeterdisplayCounter->setVisible(false);
        ui->multimeterdisplayFrequencyValue->setVisible(true);
        counterButtonEnable(false);
        mode = MAIN_MODE_MeterFrequency;
        m_mmTimer.start(METER_RESULT_REFRESH_Ms);
    } else if (ui->radioPulse->isChecked()) {
        ui->groupBoxSourceGate->setEnabled(false);
        ui->multimeterdisplayDutyValue->setVisible(false);
        ui->multimeterdisplayDutyValue_s->setVisible(false);
        ui->multimeterdisplayPeriodValue->setVisible(false);
        ui->multimeterdisplayCounter->setVisible(true);
        ui->multimeterdisplayFrequencyValue->setVisible(false);
        counterButtonEnable(true);
        mode = MAIN_MODE_MeterCount;
        m_mmTimer.stop();
    } else if (ui->radioPeriod->isChecked()) {
        ui->groupBoxSourceGate->setEnabled(false);
        ui->multimeterdisplayDutyValue->setVisible(true);
        ui->multimeterdisplayDutyValue_s->setVisible(true);
        ui->multimeterdisplayPeriodValue->setVisible(true);
        ui->multimeterdisplayCounter->setVisible(false);
        ui->multimeterdisplayFrequencyValue->setVisible(true);
        counterButtonEnable(false);
        mode = MAIN_MODE_MeterPeriod;
        m_mmTimer.start(METER_RESULT_REFRESH_Ms);
    } else if (ui->radioExtGate->isChecked()) {
        ui->groupBoxSourceGate->setEnabled(true);
        ui->multimeterdisplayDutyValue->setVisible(false);
        ui->multimeterdisplayDutyValue_s->setVisible(false);
        ui->multimeterdisplayPeriodValue->setVisible(false);
        ui->multimeterdisplayCounter->setVisible(true);
        ui->multimeterdisplayFrequencyValue->setVisible(false);
        counterButtonEnable(false);
        mode = MAIN_MODE_MeterGate;
        m_mmTimer.start(METER_RESULT_REFRESH_Ms);
    }
    sendMeterFrequencyControl();
    sendMainModeControls(mode);
}

void XlabInterface::meterTabChanged(int index)
{
    m_mmTimer.stop();
    if (ui->mainTabWidget->currentIndex() == MAIN_TAB_Mater){
        if (index == 0) {             // VAC
            mode = MAIN_MODE_MeterVAC;
            sendMainModeControls(mode);
            m_mmTimer.start(METER_RESULT_REFRESH_Ms);
        } else if (index == 1)  {     // VDC
            mode = MAIN_MODE_MeterVDC;
            sendMainModeControls(mode);
            m_mmTimer.start(METER_RESULT_REFRESH_Ms);
        } else if (index == 2) {      // f, pulse, counter, gate
            sendTriggerSource();
            sendGateSource();
            meterFrequencyModeChanged();
         }
     }
}

void XlabInterface::awgDACBufferChanged(int index)
{
    index = index + 1;
    index = dataValueRangeTrunk(index, DAC_BUFFER_BLOCKS_MIN, DAC_BUFFER_BLOCKS_MAX);
    sendDACBufferBlocks((byte)index);
    bufferCheckMaxSize();
}

void XlabInterface::bufferCheckMaxSize()
{
    unsigned char maxbuffer[1];
    int size;

    clearBuffer(maxbuffer, 1);
    size = usbDevice.controlReadTransfer(USB_REQUEST_DATA, maxbuffer, 1, MAIN_DATA_REQUEST_DACBufferMax);
    if (size == 1) {
        bufferComboBoxItemChange(ui->comboBoxDACBuffer, maxbuffer[0]);
    }
    size = usbDevice.controlReadTransfer(USB_REQUEST_DATA, maxbuffer, 1, MAIN_DATA_REQUEST_CHBufferMax);
    if (size == 1) {
        bufferComboBoxItemChange(ui->comboBoxADCBuffer, maxbuffer[0]);
    }
}

void XlabInterface::bufferComboBoxItemChange(QComboBox *combo, int count)
{
    int idx;

    if (count < 1) count = 1;
    if (combo->count() != count) {
        idx = combo->currentIndex();
        combo->blockSignals(true);
        combo->clear();
        for (int i = 0; i < count; i++) {
            combo->addItem(QString::number((i + 1) * DATA_BUFFER_BLOCK_SIZE));
        }
        if (idx >= combo->count()) idx = combo->count() - 1;
        combo->setCurrentIndex(idx);
        combo->blockSignals(false);
    }
}

void XlabInterface::awgEnableChanged()
{
    unsigned char result[2];

    clearBuffer(result, 2);
    if (sendDACControls(result, 2) < 0) return;
    if (result[0] == MAIN_CONFIGURATION_RESULT_NoEnoughMemory) {
        logToFile("No enough memory for CH buffer: " + QString::number(result[1]), LOGGING_LEVEL_Debug);
        scopeADCBufferChanged(result[1] - 1);
        if (sendDACControls(result, 2) != 0) return;
    }
    ui->comboBoxDACBuffer->setEnabled(ui->checkBoxAWBEnable->isChecked());
    bufferCheckMaxSize();
}

void XlabInterface::awgAmplitudeSliderChanged(int value)
{
    ui->doubleSpinBoxAmp->setValue((double)(value) / AMPLITUDE_SLIDER_FACTOR);
    sendWaveAmplitude(value);
}

void XlabInterface::awgDutyCycleSliderChanged(int value)
{
    ui->doubleSpinBoxDuty->setValue((double)(value / DUTY_SLIDER_FACTOR));
    sendWaveDuty(value);
}

void XlabInterface::awgOffsetSliderChanged(int value)
{
    ui->doubleSpinBoxOffset->setValue((double)(value / OFFSET_SLIDER_FACTOR));
    sendWaveOffset(value);
}

void XlabInterface::sweepCursorsUpdate()
{
    double sweepmin, sweepmax;
    double freqmin, freqmax;

    sweepmin = (byte)(ui->sweepStartFreqSlider->value());
    sweepmax = (byte)(ui->sweepEndFreqSlider->value());
    freqmin = pow(10, sweepmin / 51);
    freqmax = pow(10, sweepmax / 51);
    if (ui->radioButtonSweepFrequency->isChecked()) {
        if (freqmin < 1000) {
            ui->startFreqText->setDecimals(0);
            ui->startFreqText->setValue(round(freqmin));
            ui->labelUnitStart->setText("Hz");
        } else {
            ui->startFreqText->setDecimals(1);
            ui->startFreqText->setValue(round(freqmin / 1000));
            ui->labelUnitStart->setText("kHz");
        }
        if (freqmax < 1000) {
            ui->endFreqText->setDecimals(0);
            ui->endFreqText->setValue(round(freqmax));
            ui->labelUnitEnd->setText("Hz");
        } else {
            ui->endFreqText->setDecimals(1);
            ui->endFreqText->setValue(round(freqmax / 1000));
            ui->labelUnitEnd->setText("kHz");
        }
    } else if (ui->radioButtonSweepAmplitude->isChecked()) {
        ui->startFreqText->setDecimals(2);
        ui->endFreqText->setDecimals(2);
        ui->startFreqText->setValue(sweepmin * 0.5018 / 56);
        ui->endFreqText->setValue(sweepmax * 0.5018 / 56);
        ui->labelUnitStart->setText("V");
        ui->labelUnitEnd->setText("V");
    } else if (ui->radioButtonSweepOffset->isChecked()) {
        ui->startFreqText->setDecimals(2);
        ui->endFreqText->setDecimals(2);
        ui->startFreqText->setValue((sweepmin - 127.5) * 0.5017 / 28);
        ui->endFreqText->setValue((sweepmax - 127.5) * 0.5017 / 28);
        ui->labelUnitStart->setText("V");
        ui->labelUnitEnd->setText("V");
    } else if (ui->radioButtonSweepDutyCycle->isChecked()) {
        ui->startFreqText->setDecimals(1);
        ui->endFreqText->setDecimals(1);
        ui->startFreqText->setValue(sweepmin * 100.0 / 255);
        if ((sweepmin * 100.0 / 255)<1) ui->startFreqText->setValue(1);
        ui->endFreqText->setValue(sweepmax * 100.0 / 255);
        if ((sweepmax * 100.0 / 255)<1) ui->endFreqText->setValue(1);
        ui->labelUnitStart->setText("%");
        ui->labelUnitEnd->setText("%");
    }
}

void XlabInterface::sweepSpeedSliderChanged(int value)
{
    ui->sweepSpeedText->setText(QString::number(value));
    sendSweepSpeed(value);
}

void XlabInterface::sweepTypeChanged()
{
    sweepCursorsUpdate();
    sendSweepTypeControls();
}

void XlabInterface::sweepStartFreqChanged(QSlider *start, QSlider *end) {
    if (start->value() > end->value()) {
        start->setValue(end->value());
    }
    sendSweepStart((byte)(start->value()));
}

void XlabInterface::sweepEndFreqChanged(QSlider *start, QSlider *end)
{
    if (end->value() < start->value()) {
        end->setValue(start->value());
    }
    sendSweepEnd((byte)(end->value()));
}

void XlabInterface::sniffProtocolTabChanged(int index)
{
    sniffStop();
    mode = MAIN_MODE_Scope;
    sendMainModeControls(mode);
    sendDecodeProtocol(index);
}

void XlabInterface::scopeADCBufferChanged(int index)
{
    index = index + 1;
    index = dataValueRangeTrunk(index, CH_BUFFER_BLOCKS_MIN, CH_BUFFER_BLOCKS_MAX);
    sendCHBufferBlocks((byte)index);
    bufferCheckMaxSize();
    scopeRefreshSamplingRateLabel();
    hCursorRefreshVisible();
    sendCH1Controls();
    sendCH2Controls();
    vCursorRefreshLabel();
    vCursorRefresh();
    m_repaint = true;
}

void XlabInterface::scopeSetSizeX(int blocks)
{
    bufferSize = DATA_BUFFER_BLOCK_SIZE * blocks;
    xAxisMax = 2 * bufferSize;
    triggerSetRange(4, (bufferSize - 8), 127, TRIGGRT_WINDOW_VALUE_MAX, TRIGGRT_WINDOW_DISTANCE_MIN);
    triggerRefreshPos();
    sendTriggerPost(trigger.coord[trigger.channel].posX);
    vCursorSetRange(8, xAxisMax - 8);
    vCursorRefresh();
    xAxisChanged(ui->plotterWidget->xAxis->range());
}

void XlabInterface::mainTabChanged(int index)
{
    if (index != MAIN_TAB_Mater) {
        m_mmTimer.stop();
        sniffStop();
        sendTriggerSource(ui->comboBoxTrigSource->currentIndex());
        sendCH1Controls();
        sendCH2Controls();
        mode = MAIN_MODE_Scope;
        sendMainModeControls(mode);
        m_repaint = true;
    }
    if (index == MAIN_TAB_Mater){
        meterTabChanged(ui->mmTabWidget->currentIndex());
    }
    if (index == MAIN_TAB_Graph) ui->stackedWidget->setCurrentIndex(1);
    else if(index == MAIN_TAB_ProtocolSniffer) ui->stackedWidget->setCurrentIndex(2);
    else ui->stackedWidget->setCurrentIndex(0);
}

void XlabInterface::horizontalScrollBarChanged(int position)
{
    if (!(ui->plotterWidget->xAxis->range().lower - position == 0)) {
        ui->plotterWidget->xAxis->setRange(position, ui->plotterWidget->xAxis->range().size(), Qt::AlignLeft);
        ui->plotterWidget->replot();
    }
}

void XlabInterface::sniffPauseButtonChanged()
{
    if (ui->pauseSnifferButton->isChecked()) {
        ui->pauseSnifferButton->setText(tr("Continue"));
    } else {
        ui->pauseSnifferButton->setText(tr("Pause"));
    }
}

byte XlabInterface::scopeCH1ControlsValue()
{
    byte field = 0;

    if (ui->checkBoxCH1Trace->isChecked()) field += (1 << MAIN_CHANNEL_CONTROL_Enable);
    if (ui->checkBoxCH1ACDC->isChecked()) field += (1 << MAIN_CHANNEL_CONTROL_FilterDC);
    if (ui->checkBoxCH1Invert->isChecked()) field += (1 << MAIN_CHANNEL_CONTROL_Invert);
    if (ui->checkBoxCH1Average->isChecked()) field += (1 << MAIN_CHANNEL_CONTROL_Average);
    if (ui->checkBoxCH1Math->isChecked()) field += (1 << MAIN_CHANNEL_CONTROL_Math);
    if (ui->radioButtonCH1Sub->isChecked()) field += (1 << MAIN_CHANNEL_CONTROL_Subtract);
    if (ui->radioButtonCH1Multiply->isChecked()) field += (1 << MAIN_CHANNEL_CONTROL_Multiply);
    return field;
}

byte XlabInterface::scopeCH2ControlsValue()
{
    byte field = 0;

    if (ui->checkBoxCH2Trace->isChecked()) field += (1 << MAIN_CHANNEL_CONTROL_Enable);
    if (ui->checkBoxCH2ACDC->isChecked()) field += (1 << MAIN_CHANNEL_CONTROL_FilterDC);
    if (ui->checkBoxCH2Invert->isChecked()) field += (1 << MAIN_CHANNEL_CONTROL_Invert);
    if (ui->checkBoxCH2Average->isChecked()) field += (1 << MAIN_CHANNEL_CONTROL_Average);
    if (ui->checkBoxCH2Math->isChecked()) field += (1 << MAIN_CHANNEL_CONTROL_Math);
    if (ui->radioButtonCH2Sub->isChecked()) field += (1 << MAIN_CHANNEL_CONTROL_Subtract);
    if (ui->radioButtonCH2Multiply->isChecked()) field += (1 << MAIN_CHANNEL_CONTROL_Multiply);
    return field;
}

byte XlabInterface::scopeCHDControlsValue()
{
    byte field = 0;

    if (ui->checkBoxCHDTrace->isChecked()) field += (1 << MAIN_CHANNEL_CONTROL_Enable);
    if (ui->checkBoxCHDInvert->isChecked()) field += (1 << MAIN_CHANNEL_CONTROL_Invert);
    return field;
}

// MAIN_CONFIGURATION_Mode
void XlabInterface::sendMainModeControls(int mode)
{
    sendMainConfiguration(MAIN_CONFIGURATION_Mode, (byte)mode);
}

// MAIN_CONFIGURATION_SamplingRate
void XlabInterface::sendSamplingRate(int value)
{
    sendMainConfiguration(MAIN_CONFIGURATION_SamplingRate, (byte)value);
}

// MAIN_CONFIGURATION_DACctrl
int XlabInterface::sendDACControls(unsigned char *result, int length)
{
    byte field = 0;

    if (ui->checkBoxAWBEnable->isChecked()) field += (1 << MAIN_CHANNEL_CONTROL_Enable);
    return sendMainConfigurationConfirm(result, length, MAIN_CONFIGURATION_DACctrl, field);
}

// MAIN_CONFIGURATION_CH1ctrl
int XlabInterface::sendCH1Controls(unsigned char *result, int length)
{
    return sendMainConfigurationConfirm(result, length, MAIN_CONFIGURATION_CH1ctrl, scopeCH1ControlsValue());
}

void XlabInterface::sendCH1Controls()
{
    sendMainConfiguration(MAIN_CONFIGURATION_CH1ctrl, scopeCH1ControlsValue());
}

// MAIN_CONFIGURATION_CH2ctrl
int XlabInterface::sendCH2Controls(unsigned char *result, int length)
{
    return sendMainConfigurationConfirm(result, length, MAIN_CONFIGURATION_CH2ctrl, scopeCH2ControlsValue());
}

void XlabInterface::sendCH2Controls()
{
    sendMainConfiguration(MAIN_CONFIGURATION_CH2ctrl, scopeCH2ControlsValue());
}

// MAIN_CONFIGURATION_CHDctrl
int XlabInterface::sendCHDControls(unsigned char *result, int length) {
    return sendMainConfigurationConfirm(result, length, MAIN_CONFIGURATION_CHDctrl, scopeCHDControlsValue());
}

void XlabInterface::sendCHDControls() {
    sendMainConfiguration(MAIN_CONFIGURATION_CHDctrl, scopeCHDControlsValue());
}

// MAIN_CONFIGURATION_CHDmask
void XlabInterface::sendCHDBitControls()
{
    byte field = 0;

    if ((bitChecked[0] = ui->checkBoxCHD0->isChecked()) == true) field += (1 << 0);
    if ((bitChecked[1] = ui->checkBoxCHD1->isChecked()) == true) field += (1 << 1);
    if ((bitChecked[2] = ui->checkBoxCHD2->isChecked()) == true) field += (1 << 2);
    if ((bitChecked[3] = ui->checkBoxCHD3->isChecked()) == true) field += (1 << 3);
    if ((bitChecked[4] = ui->checkBoxCHD4->isChecked()) == true) field += (1 << 4);
    if ((bitChecked[5] = ui->checkBoxCHD5->isChecked()) == true) field += (1 << 5);
    if ((bitChecked[6] = ui->checkBoxCHD6->isChecked()) == true) field += (1 << 6);
    if ((bitChecked[7] = ui->checkBoxCHD7->isChecked()) == true) field += (1 << 7);
    sendMainConfiguration(MAIN_CONFIGURATION_CHDmask, field);
}

// MAIN_CONFIGURATION_Trigger
void XlabInterface::sendTriggerControls()
{
    byte field = 0;

    if (ui->radioButtonFree->isChecked()) field += (1 << MAIN_TRIGGER_Free);
    if (ui->radioButtonNormal->isChecked()) field += (1 << MAIN_TRIGGER_Normal);
    if (ui->radioButtonSingle->isChecked()) field += (1 << MAIN_TRIGGER_Single);
    if (ui->radioButtonAuto->isChecked()) field += (1 << MAIN_TRIGGER_Auto);
    if (ui->radioButtonFallingEdge->isChecked() || ui->radioButtonNegativeSlope->isChecked()) field += (1 << MAIN_TRIGGER_Falling);
    if (ui->radioButtonPositiveSlope->isChecked() || ui->radioButtonNegativeSlope->isChecked()) field += (1 << MAIN_TRIGGER_Slope);
    if (ui->radioButtonRisingEdge->isChecked() || ui->radioButtonFallingEdge->isChecked()) field += (1 << MAIN_TRIGGER_Edge);
    if (ui->radioButtonWindow->isChecked()) field += (1 << MAIN_TRIGGER_Window);
    sendMainConfiguration(MAIN_CONFIGURATION_Trigger, field);
}

// MAIN_CONFIGURATION_TriggerSource
void XlabInterface::sendTriggerSource()
{
    int value;

    if (ui->fcR0->isChecked()) value = MAIN_TRIGGER_SOURCE_CHD_Bit0;
    else if (ui->fcR1->isChecked()) value = MAIN_TRIGGER_SOURCE_CHD_Bit1;
    else if (ui->fcR2->isChecked()) value = MAIN_TRIGGER_SOURCE_CHD_Bit2;
    else if (ui->fcR3->isChecked()) value = MAIN_TRIGGER_SOURCE_CHD_Bit3;
    else if (ui->fcR4->isChecked()) value = MAIN_TRIGGER_SOURCE_CHD_Bit4;
    else if (ui->fcR5->isChecked()) value = MAIN_TRIGGER_SOURCE_CHD_Bit5;
    else if (ui->fcR6->isChecked()) value = MAIN_TRIGGER_SOURCE_CHD_Bit6;
    else if (ui->fcR7->isChecked()) value = MAIN_TRIGGER_SOURCE_CHD_Bit7;
    else if (ui->fcR10->isChecked()) value = MAIN_TRIGGER_SOURCE_EXT;
    else if (ui->fcR11->isChecked()) value = MAIN_TRIGGER_SOURCE_SCHMITT;
    sendTriggerSource(value);
}

void XlabInterface::sendTriggerSource(int value)
{
    sendMainConfiguration(MAIN_CONFIGURATION_TriggerSource, value);
}

// MAIN_CONFIGURATION_GateSource
void XlabInterface::sendGateSource()
{
    int value = 0;

    if (ui->fcR0_2->isChecked()) value = MAIN_TRIGGER_SOURCE_CHD_Bit0;
    else if (ui->fcR1_2->isChecked()) value = MAIN_TRIGGER_SOURCE_CHD_Bit1;
    else if (ui->fcR2_2->isChecked()) value = MAIN_TRIGGER_SOURCE_CHD_Bit2;
    else if (ui->fcR3_2->isChecked()) value = MAIN_TRIGGER_SOURCE_CHD_Bit3;
    else if (ui->fcR4_2->isChecked()) value = MAIN_TRIGGER_SOURCE_CHD_Bit4;
    else if (ui->fcR5_2->isChecked()) value = MAIN_TRIGGER_SOURCE_CHD_Bit5;
    else if (ui->fcR6_2->isChecked()) value = MAIN_TRIGGER_SOURCE_CHD_Bit6;
    else if (ui->fcR7_2->isChecked()) value = MAIN_TRIGGER_SOURCE_CHD_Bit7;
    else if (ui->fcR10_2->isChecked()) value = MAIN_TRIGGER_SOURCE_EXT;
    else if (ui->fcR11_2->isChecked()) value = MAIN_TRIGGER_SOURCE_SCHMITT;
    sendMainConfiguration(MAIN_CONFIGURATION_GateSource, value);
}

// MAIN_CONFIGURATION_Mixed
void XlabInterface::sendMixedControls()
{
    byte field = 0;

    if (ui->rollMode->isChecked()) field += (1 << MAIN_MSO_MODE_Roll);
    if (ui->elasticMode->isChecked()) field += (1 << MAIN_MSO_MODE_Elastic);
    if (ui->checkBoxmeterondev->isChecked()) field += (1 << MAIN_METER_AlwaysOn);
    if (awg_wave_update_disable) field += (1 << MAIN_AWG_UpdateDisable);
    if (ui->radioButtonSniffSingle->isChecked()) field += (1 << MAIN_SNIFFER_SingleSniff);
    sendMainConfiguration(MAIN_CONFIGURATION_Mixed, field);
}

// MAIN_CONFIGURATION_MeterFrequency
void XlabInterface::sendMeterFrequencyControl()
{
    byte field = 0;

    if (ui->checkBoxFreqFastMode->isChecked()) field += (1 << METER_FREQUENCY_FastMode);
    if (ui->radioFrequencyGatePeriod->isChecked()) field += (1 << METER_FREQUENCY_GATE_Period);
    sendMainConfiguration(MAIN_CONFIGURATION_MeterFrequency, field);
}

// MAIN_CONFIGURATION_SniffOption
void XlabInterface::sendSnifferOptionControl()
{
    byte field = 0;

    if (ui->comboBoxCPOL->currentIndex() == 1) field += (1 << SNIFFER_SPI_CPOL);
    if (ui->comboBoxCPHA->currentIndex() == 1) field += (1 << SNIFFER_SPI_CPHA);
    if (ui->comboBoxSniffSSINV->currentIndex() == 1) field += (1 << SNIFFER_SPI_SSINV);
    if (ui->comboBoxParity->currentIndex() > 0) field += (1 << SNIFFER_UART_ParityEnable);
    if (ui->comboBoxParity->currentIndex() == 2) field += (1 << SNIFFER_UART_ParityOdd);
    if (ui->comboBoxStopBits->currentIndex()== 1) field += (1 << SNIFFER_UART_StopBit2);
    if (ui->checkBoxSynchro->isChecked()) field += (1 << SNIFFER_RxTx_Synchronization);
    if (ui->checkBoxSniffParallelInputInvert->isChecked()) field += (1 << SNIFFER_PARALLEL_InputInvert);
    sendMainConfiguration(MAIN_CONFIGURATION_SniffOption, field);
}

// MAIN_CONFIGURATION_CH1gain
void XlabInterface::sendChannel1GainControls(int value)
{
    sendMainConfiguration(MAIN_CONFIGURATION_CH1gain, (byte)value);
}

// MAIN_CONFIGURATION_CH2gain
void XlabInterface::sendChannel2GainControls(int value)
{
    sendMainConfiguration(MAIN_CONFIGURATION_CH2gain, (byte)value);
}

// MAIN_CONFIGURATION_TriggerHold
void XlabInterface::sendTriggerHold(int value)
{
    sendMainConfiguration(MAIN_CONFIGURATION_TriggerHold, (byte)value);
}

// MAIN_CONFIGURATION_TriggerPos
void XlabInterface::sendTriggerPost(int value)
{
    sendMainConfiguration(MAIN_CONFIGURATION_TriggerPos, (bufferSize - value));
}

// MAIN_CONFIGURATION_TriggerLevel
void XlabInterface::sendTriggerLevel(int value)
{
    sendMainConfiguration(MAIN_CONFIGURATION_TriggerLevel, value);
}

// MAIN_CONFIGURATION_Window1
void XlabInterface::sendTriggerWin1Level(int value)
{
    sendMainConfiguration(MAIN_CONFIGURATION_Window1, value);
}

// MAIN_CONFIGURATION_Window2
void XlabInterface::sendTriggerWin2Level(int value)
{
    sendMainConfiguration(MAIN_CONFIGURATION_Window2, value);
}

// MAIN_CONFIGURATION_AutoTrigger
void XlabInterface::sendAutoTrigger(int value)
{
    sendMainConfiguration(MAIN_CONFIGURATION_AutoTrigger, value);
}

// MAIN_CONFIGURATION_SniffProtocol
void XlabInterface::sendDecodeProtocol(int protocol)
{
    sendMainConfiguration(MAIN_CONFIGURATION_SniffProtocol, (byte)protocol);
}

// MAIN_CONFIGURATION_USARTBitRate
void XlabInterface::sendSnifferUSARTControl()
{
    byte field = 0;

    field += (byte)ui->comboBoxBaud->currentIndex();
    field += ((byte)ui->comboBoxUARTData->currentIndex() << SNIFFER_UART_DATABITS_bp);
    sendMainConfiguration(MAIN_CONFIGURATION_USARTBitRate, field);
}

// MAIN_CONFIGURATION_SniffStrobeEdge
void XlabInterface::sendSnifferStrobeEdge(int value)
{
    sendMainConfiguration(MAIN_CONFIGURATION_SniffStrobeEdge, value);
}

// MAIN_CONFIGURATION_DACBufferBlocks
void XlabInterface::sendDACBufferBlocks(int value)
{
    sendMainConfiguration(MAIN_CONFIGURATION_DACBufferBlocks, (byte)value);
}

// MAIN_CONFIGURATION_CHBufferBlocks
void XlabInterface::sendCHBufferBlocks(int value)
{
    sendMainConfiguration(MAIN_CONFIGURATION_CHBufferBlocks, (byte)value);
}

// MAIN_CONFIGURATION_SweepType
void XlabInterface::sendSweepTypeControls()
{
    byte field = 0;

    if (ui->radioButtonSweepFalling->isChecked()) field += (1 << AWG_SWEEP_TYPE_SweepDown);
    if (ui->radioButtonSweepPingPong->isChecked()) field += (1 << AWG_SWEEP_TYPE_PingPong);
    if (ui->collapsiblePanelAWGSweep->isChecked()) {
        if (ui->radioButtonSweepFrequency->isChecked()) field += (1 << AWG_SWEEP_TYPE_SweepFrequency);
        if (ui->radioButtonSweepAmplitude->isChecked()) field += (1 << AWG_SWEEP_TYPE_SweepAmplitude);
        if (ui->radioButtonSweepOffset->isChecked()) field += (1 << AWG_SWEEP_TYPE_SweepOffset);
        if (ui->radioButtonSweepDutyCycle->isChecked()) field += (1 << AWG_SWEEP_TYPE_SweepDuty);
    }
    sendMainConfiguration(MAIN_CONFIGURATION_SweepType, field);
}

// MAIN_CONFIGURATION_SweepStart
void XlabInterface::sendSweepStart(int value)
{
    sweepCursorsUpdate();
    sendMainConfiguration(MAIN_CONFIGURATION_SweepStart, (byte)value);
}

// MAIN_CONFIGURATION_SweepEnd
void XlabInterface::sendSweepEnd(int value)
{
    sweepCursorsUpdate();
    sendMainConfiguration(MAIN_CONFIGURATION_SweepEnd, (byte)value);
}

// MAIN_CONFIGURATION_SweepSpeed
void XlabInterface::sendSweepSpeed(int value)
{
    value = dataValueRangeTrunk(value, 1, AWG_SWEEP_SPEED_MAX);
    sendMainConfiguration(MAIN_CONFIGURATION_SweepSpeed, (byte)value);
}

// MAIN_CONFIGURATION_WaveAmplitude
void XlabInterface::sendWaveAmplitude(int value)
{
    value = dataValueRangeTrunk(value, 0, AWG_WAVE_MAX_Amplitude);
    sendMainConfiguration(MAIN_CONFIGURATION_WaveAmplitude, (byte)value);
}

// MAIN_CONFIGURATION_WaveType
void XlabInterface::sendWaveType()
{
    byte value = 0;
    QString waveName = "";

    if (ui->radioButtonConstant->isChecked()) {
        value = AWG_WAVE_Constant;
        waveName = "Constant";
    }
    else if (ui->radioButtonCustom->isChecked()) {
        value = AWG_WAVE_Custom;
        waveName = "Custom";
    }
    else if (ui->radioButtonExpo->isChecked()) {
        value = AWG_WAVE_Exponential;
        waveName = "Exponential";
    }
    else if (ui->radioButtonTriangle->isChecked()) {
        value = AWG_WAVE_Triangle;
        waveName = "Triangle";
    }
    else if (ui->radioButtonSquare->isChecked()) {
        value = AWG_WAVE_Square;
        waveName = "Square";
    }
    else if (ui->radioButtonSine->isChecked()) {
        value = AWG_WAVE_Sine;
        waveName = "Sine";
    }
    else if (ui->radioButtonPinkNoise->isChecked()) {
        value = AWG_WAVE_PinkNoise;
        waveName = "Pink Noise";
    }
    else if (ui->radioButtonWhiteNoise->isChecked()) {
        value = AWG_WAVE_WhiteNoise;
        waveName = "White Noise";
    }
    ui->collapsiblePanelAWGWaveform->setText(waveName + " ");
    sendMainConfiguration(MAIN_CONFIGURATION_WaveType, value);
    if (value == AWG_WAVE_Constant) sendWaveConstantLevel(ui->spinBoxConstant->value());
}

// MAIN_CONFIGURATION_WaveDuty
void XlabInterface::sendWaveDuty(int value)
{
    value = dataValueRangeTrunk(value, AWG_WAVE_MIN_Duty, 255);
    sendMainConfiguration(MAIN_CONFIGURATION_WaveDuty, (byte)value);
}

// MAIN_CONFIGURATION_WaveOffset
void XlabInterface::sendWaveOffset(int value)
{
    value = dataValueRangeTrunk(value, -127, 127);
    sendMainConfiguration(MAIN_CONFIGURATION_WaveOffset, (byte)value);
}

// MAIN_CONFIGURATION_WaveFrequency
void XlabInterface::sendWaveFrequency(qint64 value)
{
    value = dataValueRangeTrunk64(value, AWG_WAVE_MIN_Frequency, AWG_WAVE_MAX_Frequency);
    sendMainConfigurationFrequency(MAIN_CONFIGURATION_WaveFrequency, value);
}

// MAIN_CONFIGURATION_LevelConstant
void XlabInterface::sendWaveConstantLevel(int value)
{
    sendMainConfiguration(MAIN_CONFIGURATION_LevelConstant, (int16_t)value);
}

void XlabInterface::sendMainConfiguration(byte config, int value)
{
    if (writeBulkConfig) {
        *((uint32_t *)(awgBuffer) + config) = (uint32_t)value;
        logToFile("Bulk write to USB: index(" + QString::number(config) + ") value:" + QString::number(value) , LOGGING_LEVEL_Debug);
    } else {
        if (!usbDevice.isDeviceConnected) return;
        usbDevice.controlWriteTransfer(USB_SEND_CONFIGURATION, config, value);
        logToFile("Write to USB: index(" + QString::number(config) + ") value:" + QString::number(value) , LOGGING_LEVEL_Debug);
    }
}

int XlabInterface::sendMainConfigurationConfirm(unsigned char *result, int length, byte config, int value)
{
    if (!usbDevice.isDeviceConnected) return -1;
    int size = usbDevice.controlReadTransfer(USB_SEND_CONFIGURATION_ACK, result, length, config, value);
    if (length != size) return -1;
    logToFile("Write to USB: index(" + QString::number(config) + ") value:" + QString::number(value) + ", return:" + QString::number(result[0]), LOGGING_LEVEL_Debug);
    return result[0];
}

void XlabInterface::sendMainConfigurationFrequency(byte config, qint64 value)
{
    if (writeBulkConfig) {
        *((uint32_t *)(awgBuffer) + config) = (uint32_t)value;
        logToFile("Bulk write to USB: index(" + QString::number(config) + ") value:" + QString::number(value) , LOGGING_LEVEL_Debug);
    } else {
        if (!usbDevice.isDeviceConnected) return;
        usbDevice.controlWriteTransfer(USB_SEND_FREQUENCY, (uint16_t)(value & 0x0000ffff), (uint16_t)(value >> 16));
        logToFile("Write to USB: Send Frequency value:" + QString::number(value) , LOGGING_LEVEL_Debug);
    }
}

// MAIN_CONTROL_SCOPE_Start
void XlabInterface::sendScopeStart()
{
    if (!usbDevice.isDeviceConnected) return;
    usbDevice.controlWriteTransfer(USB_SEND_CONTROL, MAIN_CONTROL_SCOPE_Start, 0);
    ui->stopButton->setText(tr(BUTTON_STOP_SCOPE_TEXT_STOP));
}

// MAIN_CONTROL_SCOPE_Stop
void XlabInterface::sendScopeStop()
{
    if (!usbDevice.isDeviceConnected) return;
    usbDevice.controlWriteTransfer(USB_SEND_CONTROL, MAIN_CONTROL_SCOPE_Stop, 0);
    ui->stopButton->setText(tr(BUTTON_STOP_SCOPE_TEXT_START));
}

// MAIN_CONTORL_SNIFFER_Start
void XlabInterface::sendSnifferStart()
{
    usbDevice.controlWriteTransfer(USB_SEND_CONTROL, MAIN_CONTORL_SNIFFER_Start, 0);
}

// MAIN_CONTORL_SNIFFER_Stop
void XlabInterface::sendSnifferStop()
{
    usbDevice.controlWriteTransfer(USB_SEND_CONTROL, MAIN_CONTORL_SNIFFER_Stop, 0);
}

// MAIN_CONTORL_SNIFFER_Clear
void XlabInterface::sendSnifferClear()
{
    usbDevice.controlWriteTransfer(USB_SEND_CONTROL, MAIN_CONTORL_SNIFFER_Clear, 0);
}

// MAIN_CONTROL_COUNTER_Start
void XlabInterface::sendCouterStart()
{
    usbDevice.controlWriteTransfer(USB_SEND_CONTROL, MAIN_CONTROL_COUNTER_Start, 0);
}

// MAIN_CONTROL_COUNTER_Stop
void XlabInterface::sendCouterStop()
{
    usbDevice.controlWriteTransfer(USB_SEND_CONTROL, MAIN_CONTROL_COUNTER_Stop, 0);
}

// MAIN_CONTROL_TRIGGER_Force
void XlabInterface::sendScopeForce()
{
    usbDevice.controlWriteTransfer(USB_SEND_CONTROL, MAIN_CONTROL_TRIGGER_Force, 0);
}

void XlabInterface::on_stopButton_clicked()
{
    if (ui->stopButton->text() == tr(BUTTON_STOP_SCOPE_TEXT_START))
        sendScopeStart();
    else
        sendScopeStop();
}

void XlabInterface::on_playButton_clicked()
{
    ui->label_13->setText("Reading data from file. Please wait");
    QString fpath = QFileDialog::getOpenFileName(0, "Choose file with recorded wave", "", "XLab files (*.xsp)");
    qApp->setActiveWindow(this);
    if (fpath != "")
        wavePlay(fpath);
    else
        ui->label_13->setText("");
}

void XlabInterface::on_connectButton_clicked()
{
    setInfoText("Please wait...", true);
    if (!usbDevice.isDeviceConnected) {
        usbDevice.initializeDevice();
        usbDevice.openDevice();
        if (usbDevice.isDeviceConnected) {
            //Sleep(100);
            if (deviceInitialize() == 0) ui->connectButton->setText("Disconnect");
        }
    } else {
        disconnectDevice();
        ui->connectButton->setText("Connect");
        ui->textBrowserDeviceInfo->clear();
        setInfoText();
    }
}

void XlabInterface::on_samplingSlider_valueChanged(int value)
{
    scopeSamplingSliderChanged(value);
}

void XlabInterface::on_openCSVButton_clicked()
{
    QString path;
    QString csvString;

    path = QFileDialog::getOpenFileName(this, tr("Open File"), defaultDir, "CSV files (*.csv *.txt);;All files (*.*)");
    if (path.isEmpty()) return;
    QFile csvFile;
    csvFile.setFileName(path);
    csvFile.open(QIODevice::ReadOnly);
    csvString = csvFile.readAll();
    csvFile.close();
    parseCSV(csvString, awgBuffer);
}

void XlabInterface::on_saveAWGButton_clicked()
{
    if (usbDevice.isDeviceConnected) {
        // AWG shuld by enabled and sweep off
        ui->progressBar->setValue(0);
        awg_wave_update_disable = true;
        sendMixedControls();
        // send 512 salpes of int16
        // first half
        ui->progressBar->setValue(4);
        //Sleep(300);
        usbDevice.bulkWriteTransfer(awgBuffer, USB_ENDPOINT1_AUXDATA_LENGTH);
        usbDevice.controlWriteTransfer(USB_SEND_DATA_BULK, MAIN_BULK_DATA_AWG_WaveFB);
        ui->progressBar->setValue(50);
        // second half
        usbDevice.bulkWriteTransfer(awgBuffer + USB_ENDPOINT1_AUXDATA_LENGTH, USB_ENDPOINT1_AUXDATA_LENGTH);
        usbDevice.controlWriteTransfer(USB_SEND_DATA_BULK, MAIN_BULK_DATA_AWG_WaveLB);
        awg_wave_update_disable = false;
        sendMixedControls();
        ui->progressBar->setValue(100);
        ui->radioButtonCustom->setChecked(true);
        //Sleep(300);
        sendWaveType();
        ui->progressBar->setValue(0);
    }
}

void XlabInterface::on_checkBoxCH1Invert_clicked()
{
    sendCH1Controls();
}

void XlabInterface::on_checkBoxCH1Invert_toggled()
{
    if ((trigger.source == MAIN_TRIGGER_SOURCE_CH1) && (trigger.mode == TRIGGER_MODE_Rising)) {
        ui->radioButtonFallingEdge->setChecked(true);
        scopeSetTriggerMode();
        triggerSetPosition(trigger.coord[MAIN_CHANNEL_CH1].posX, -trigger.coord[MAIN_CHANNEL_CH1].posY);
        sendTriggerLevel(trigger.coord[MAIN_CHANNEL_CH1].posY);
    } else if ((trigger.source == MAIN_TRIGGER_SOURCE_CH1) && (trigger.mode == TRIGGER_MODE_Falling)) {
        ui->radioButtonRisingEdge->setChecked(true);
        scopeSetTriggerMode();
        triggerSetPosition(trigger.coord[MAIN_CHANNEL_CH1].posX, -trigger.coord[MAIN_CHANNEL_CH1].posY);
        sendTriggerLevel(trigger.coord[MAIN_CHANNEL_CH1].posY);
    } else if ((trigger.source == MAIN_TRIGGER_SOURCE_CH1) && (trigger.mode == TRIGGER_MODE_Dual)) {
        triggerSetPosition(trigger.coord[MAIN_CHANNEL_CH1].posX, -trigger.coord[MAIN_CHANNEL_CH1].posY);
        sendTriggerLevel(trigger.coord[MAIN_CHANNEL_CH1].posY);
    }
}

void XlabInterface::on_checkBoxCH1Trace_clicked()
{
    channel1TraceChanged(ui->checkBoxCH1Trace->isChecked());
}

void XlabInterface::on_checkBoxCH1Average_clicked()
{
    sendCH1Controls();
}

void XlabInterface::on_checkBoxCH1Math_clicked()
{
    sendCH1Controls();
}

void XlabInterface::on_radioButtonCH1Sub_clicked()
{
    sendCH1Controls();
}

void XlabInterface::on_radioButtonCH1Multiply_clicked()
{
    sendCH1Controls();
}

void XlabInterface::on_checkBoxCH2Invert_clicked()
{
    sendCH2Controls();
}

void XlabInterface::on_checkBoxCH2Invert_toggled()
{
    if ((trigger.source == MAIN_TRIGGER_SOURCE_CH2) && (trigger.mode == TRIGGER_MODE_Rising)) {
        ui->radioButtonFallingEdge->setChecked(true);
        scopeSetTriggerMode();
        triggerSetPosition(trigger.coord[MAIN_CHANNEL_CH2].posX, -trigger.coord[MAIN_CHANNEL_CH2].posY);
        sendTriggerLevel(trigger.coord[MAIN_CHANNEL_CH2].posY);
    } else if ((trigger.source == MAIN_TRIGGER_SOURCE_CH2) && (trigger.mode == TRIGGER_MODE_Falling)) {
        ui->radioButtonRisingEdge->setChecked(true);
        scopeSetTriggerMode();
        triggerSetPosition(trigger.coord[MAIN_CHANNEL_CH2].posX, -trigger.coord[MAIN_CHANNEL_CH2].posY);
        sendTriggerLevel(trigger.coord[MAIN_CHANNEL_CH2].posY);
    } else if ((trigger.source == MAIN_TRIGGER_SOURCE_CH2) && (trigger.mode == TRIGGER_MODE_Dual)) {
        triggerSetPosition(trigger.coord[MAIN_CHANNEL_CH2].posX, -trigger.coord[MAIN_CHANNEL_CH2].posY);
        sendTriggerLevel(trigger.coord[MAIN_CHANNEL_CH2].posY);
    }
}

void XlabInterface::on_checkBoxCH2Trace_clicked()
{
    channel2TraceChanged(ui->checkBoxCH2Trace->isChecked());
}

void XlabInterface::on_checkBoxCH2Average_clicked()
{
    sendCH2Controls();
}

void XlabInterface::on_checkBoxCH2Math_clicked()
{
    sendCH2Controls();
}

void XlabInterface::on_radioButtonCH2Sub_clicked()
{
    sendCH2Controls();
}

void XlabInterface::on_radioButtonCH2Multiply_clicked()
{
    sendCH2Controls();
}

void XlabInterface::on_checkBoxCHDTrace_clicked()
{
    channelLogicTraceChanged();
}

void XlabInterface::on_checkBoxCHDInvert_clicked()
{
    sendCHDControls();
}

void XlabInterface::on_checkBoxASCII_clicked()
{
}

void XlabInterface::on_checkBoxCHD0_clicked()
{
    sendCHDBitControls();
}

void XlabInterface::on_checkBoxCHD1_clicked()
{
    sendCHDBitControls();
}

void XlabInterface::on_checkBoxCHD2_clicked()
{
    sendCHDBitControls();
}

void XlabInterface::on_checkBoxCHD3_clicked()
{
    sendCHDBitControls();
}

void XlabInterface::on_checkBoxCHD4_clicked()
{
    sendCHDBitControls();
}

void XlabInterface::on_checkBoxCHD5_clicked()
{
    sendCHDBitControls();
}

void XlabInterface::on_checkBoxCHD6_clicked()
{
    sendCHDBitControls();
}

void XlabInterface::on_checkBoxCHD7_clicked()
{
    sendCHDBitControls();
}

void XlabInterface::on_radioButtonRisingEdge_clicked()
{
    scopeSetTriggerMode();
}

void XlabInterface::on_radioButtonFallingEdge_clicked()
{
    scopeSetTriggerMode();
}

void XlabInterface::on_radioButtonDual_clicked()
{
    scopeSetTriggerMode();
}

void XlabInterface::on_radioButtonPositiveSlope_clicked()
{
    scopeSetTriggerMode();
}

void XlabInterface::on_radioButtonNegativeSlope_clicked()
{
    scopeSetTriggerMode();
}

void XlabInterface::on_radioButtonWindow_clicked()
{
    scopeSetTriggerMode();
}

void XlabInterface::on_radioButtonFree_clicked()
{
    scopeSetTriggerType();
}

void XlabInterface::on_radioButtonNormal_clicked()
{
    scopeSetTriggerType();
}

void XlabInterface::on_radioButtonAuto_clicked()
{
    scopeSetTriggerType();
}

void XlabInterface::on_radioButtonSingle_clicked()
{
    scopeSetTriggerType();
}

void XlabInterface::on_checkBoxCircular_clicked()
{
    sendMixedControls();
}

void XlabInterface::on_rollMode_clicked()
{
    scopeRollModeChanged();
}

void XlabInterface::on_checkBoxCursorVertical_clicked()
{
    cursorsVerticalChanged(ui->checkBoxCursorVertical->isChecked());
}

void XlabInterface::on_comboBoxScopeCursorsChannel_currentIndexChanged(int index)
{
    cursorsSetChannelActive(index);
}

void XlabInterface::on_checkBoxPersistence_clicked()
{
    displayControlsSet();
}

void XlabInterface::on_checkBoxVectors_clicked()
{
    displayVectorsChanged(ui->checkBoxVectors->isChecked());
}

void XlabInterface::on_elasticMode_clicked()
{
    scopeElasticModeChanged();
}

void XlabInterface::on_checkBoxLogY_clicked()
{
    fftControlsPlotSet();
}

void XlabInterface::on_checkBoxFFTTrace_clicked()
{
    fftControlsPlotSet();
}

void XlabInterface::on_checkBoxIQFFT_clicked()
{
    fftControlsPlotSet();
}

void XlabInterface::on_radioButtonRect_clicked()
{
    fftWindowFunctionSet();
    fftControlsPlotSet();
}

void XlabInterface::on_radioButtonHamming_clicked()
{
    fftWindowFunctionSet();
    fftControlsPlotSet();
}

void XlabInterface::on_radioButtonHann_clicked()
{
    fftWindowFunctionSet();
    fftControlsPlotSet();
}

void XlabInterface::on_radioButtonBlackman_clicked()
{
    fftWindowFunctionSet();
    fftControlsPlotSet();
}

void XlabInterface::on_xyMode_clicked()
{
    cursorsVerticalChanged(ui->checkBoxCursorVertical->isChecked());
    cursorsHorizontalChanged(ui->checkBoxScopeCursorsHorizontal->isChecked());
    displayVectorsChanged(ui->checkBoxVectors->isChecked());
    scopeSetTriggerType();
}

void XlabInterface::on_doubleSpinBoxTrigHold_valueChanged(double value)
{
    sendTriggerHold(value);
}

void XlabInterface::on_radioButtonSniffNormal_clicked()
{
    sendMixedControls();
}

void XlabInterface::on_radioButtonSniffSingle_clicked()
{
    sendMixedControls();
}

void XlabInterface::on_comboBoxBaud_currentIndexChanged(int)
{
    sendSnifferUSARTControl();
}

void XlabInterface::on_comboBoxParity_currentIndexChanged(int)
{
    sendSnifferOptionControl();
}

void XlabInterface::on_comboBoxStopBits_currentIndexChanged(int) {
    sendSnifferOptionControl();
}

void XlabInterface::on_comboBoxUARTData_currentIndexChanged(int)
{
    sendSnifferUSARTControl();
}

void XlabInterface::on_checkBoxSynchro_clicked()
{
    sendSnifferOptionControl();
}

void XlabInterface::on_comboBoxCPOL_currentIndexChanged(int)
{
    sendSnifferOptionControl();
}

void XlabInterface::on_comboBoxCPHA_currentIndexChanged(int)
{
    sendSnifferOptionControl();
}

void XlabInterface::on_comboBoxSniffSSINV_currentIndexChanged(int)
{
    sendSnifferOptionControl();
}

void XlabInterface::on_mainTabWidget_currentChanged(int index)
{
    mainTabChanged(index);
}

void XlabInterface::on_mmTabWidget_currentChanged(int index)
{
    meterTabChanged(index);
}

void XlabInterface::on_radioPeriod_clicked()
{
    ui->radioPeriod->setChecked(true);
    meterFrequencyModeChanged();
}

void XlabInterface::on_radioFreq_clicked() {
    ui->radioFreq->setChecked(true);
    meterFrequencyModeChanged();
}

void XlabInterface::on_radioPulse_clicked()
{
    ui->radioPulse->setChecked(true);
    meterFrequencyModeChanged();
}

void XlabInterface::on_pushButtonCounterStart_clicked()
{
    sendCouterStop();
    clearMMLabelsFreq();
    sendCouterStart();
    ui->pushButtonCounteStop->setEnabled(true);
    m_mmTimer.start(METER_RESULT_REFRESH_Ms / 2);
}

void XlabInterface::on_pushButtonCounteStop_clicked()
{
    sendCouterStop();
    ui->pushButtonCounteStop->setEnabled(false);
    m_mmTimer.start(100);
}

void XlabInterface::on_radioExtGate_clicked()
{
    ui->radioExtGate->setChecked(true);
    meterFrequencyModeChanged();
}

void XlabInterface::on_startSnifferButton_clicked()
{
    sendSnifferOptionControl();
    sendSnifferUSARTControl();
    sendSnifferStrobeEdge(ui->comboBoxSniffParallelStrobe->currentIndex());
    sniffbufferfulfill = 0;
    sniffbufferfulfillrx = 0;
    sniffbufferfulfilltx = 0;
    sniffbufferfirstrxloop = true;
    sniffbufferfirsttxloop = true;
    sniffControlsStartSet(ui->startSnifferButton->text());
    if (ui->startSnifferButton->text() == (tr(BUTTON_STOP_SNIFFER_TEXT_STOP))) {
        mode = MAIN_MODE_Sniff;
    } else {
        mode = MAIN_MODE_Scope;
    }
    sendMainModeControls(mode);
    sendSnifferStart();
}

void XlabInterface::on_protocolTabWidget_currentChanged(int index)
{
    sniffProtocolTabChanged(index);
    ui->stackedWidgetSniffer->setCurrentIndex(index);
}

void XlabInterface::on_forceButton_clicked()
{
    sendScopeForce();
}

void XlabInterface::on_ch1GainSlider_valueChanged(int value)
{
    channel1GainSliderChanged(value, true);
}

void XlabInterface::on_ch2GainSlider_valueChanged(int value)
{
    channel2GainSliderChanged(value, true);
}

void XlabInterface::on_horizontalScrollBar_valueChanged(int position)
{
    horizontalScrollBarChanged(position);
}

void XlabInterface::on_verticalScrollBar_valueChanged(int value)
{
    if (!(ui->plotterWidget->yAxis->range().lower - value == 0)) {
        ui->plotterWidget->yAxis->setRange((ui->verticalScrollBar->maximum()) - value, ui->plotterWidget->yAxis->range().size(), Qt::AlignLeft);
        ui->plotterWidget->replot();
    }
}

void XlabInterface::on_comboBoxTrigSource_currentIndexChanged(int index)
{
    scopeTriggerSourceChanged(index);
}

void XlabInterface::on_doubleSpinBoxTrigAuto_valueChanged(double value)
{
    scopeSetTriggerAutoTimer(value);
}

void XlabInterface::on_ch1PositionSlider_valueChanged(int value)
{
    bool ch1CursorUpdate = false;
    if (ui->comboBoxScopeCursorsChannel->currentIndex() == 0) ch1CursorUpdate = true;
    channelPositionSliderChanged(MAIN_CHANNEL_CH1, value, &ch1ZeroPos, ch1CursorUpdate);
}

void XlabInterface::on_ch2PositionSlider_valueChanged(int value)
{
    bool ch2CursorUpdate = false;
    if (ui->comboBoxScopeCursorsChannel->currentIndex() == 1) ch2CursorUpdate = true;
    channelPositionSliderChanged(MAIN_CHANNEL_CH2, value, &ch2ZeroPos, ch2CursorUpdate);
}

void XlabInterface::on_chdPositionSlider_valueChanged(int value)
{
    channelLogicPositionSliderChanged(value);
}

void XlabInterface::on_sweepStartFreqSlider_valueChanged(int)
{
    sweepStartFreqChanged(ui->sweepStartFreqSlider, ui->sweepEndFreqSlider);
}

void XlabInterface::on_sweepEndFreqSlider_valueChanged(int)
{
    sweepEndFreqChanged(ui->sweepStartFreqSlider, ui->sweepEndFreqSlider);
}

void XlabInterface::on_sweepSpeedSlider_valueChanged(int value)
{
    sweepSpeedSliderChanged(value);
}

void XlabInterface::on_doubleSpinBoxAmp_valueChanged(double value) {
    ui->amplitudeSlider->setSliderPosition(round(value * AMPLITUDE_SLIDER_FACTOR));
}

void XlabInterface::on_amplitudeSlider_valueChanged(int value)
{
    awgAmplitudeSliderChanged(value);
}

void XlabInterface::on_radioButtonCustom_clicked()
{
    sendWaveType();
}

void XlabInterface::clickRadioButtonCustom()
{
    ui->radioButtonCustom->click();
}

void XlabInterface::on_radioButtonExpo_clicked()
{
    sendWaveType();
}

void XlabInterface::clickRadioButtonExpo()
{
    ui->radioButtonExpo->click();
}

void XlabInterface::on_radioButtonTriangle_clicked()
{
    sendWaveType();
}

void XlabInterface::clickRadioButtonTriangle()
{
    ui->radioButtonTriangle->click();
}

void XlabInterface::on_radioButtonSquare_clicked()
{
    sendWaveType();
}

void XlabInterface::clickRadioButtonSquare()
{
    ui->radioButtonSquare->click();
}

void XlabInterface::on_radioButtonSine_clicked()
{
    sendWaveType();
}

void XlabInterface::clickRadioButtonSine()
{
    ui->radioButtonSine->click();
}

void XlabInterface::on_radioButtonWhiteNoise_clicked()
{
    sendWaveType();
}

void XlabInterface::clickRadioButtonWhiteNoise()
{
    ui->radioButtonWhiteNoise->click();
}

void XlabInterface::on_radioButtonPinkNoise_clicked()
{
    sendWaveType();
}

void XlabInterface::clickRadioButtonPinkNoise()
{
    ui->radioButtonPinkNoise->click();
}

void XlabInterface::on_doubleSpinBoxDuty_valueChanged(double value) {
    ui->dutyCycleSlider->setSliderPosition(round(value * DUTY_SLIDER_FACTOR));
}

void XlabInterface::on_dutyCycleSlider_valueChanged(int value)
{
    awgDutyCycleSliderChanged(value);
}

void XlabInterface::on_doubleSpinBoxOffset_valueChanged(double value)
{
    ui->offsetSlider->setSliderPosition(round(value * OFFSET_SLIDER_FACTOR));
}

void XlabInterface::on_offsetSlider_valueChanged(int value)
{
    awgOffsetSliderChanged(value);
}

float XlabInterface::frequencyRealValueCheck()
{
    double result = 0;
    byte tmp_buff[4];

    clearBuffer(tmp_buff, 4);
    int size = usbDevice.controlReadTransfer(USB_REQUEST_DATA, tmp_buff, 4, MAIN_DATA_REQUEST_ActualFrequency);
    if (size == 4) result = *((uint32_t *)tmp_buff);
    return result / 100;
}

void XlabInterface::frequencyDigitsShowValue(double value)
{
    QString f;
    int length;

    f = "00000000" + QString::number(mathRoundTo(value, 1));
    length  = f.length();
    f = f.mid(length - 8, 8);
    ui->labelFrequencyDigit001->setText(f.at(7));
    ui->labelFrequencyDigit01->setText(f.at(6));
    ui->labelFrequencyDigit1->setText(f.at(5));
    ui->labelFrequencyDigit10->setText(f.at(4));
    ui->labelFrequencyDigit100->setText(f.at(3));
    ui->labelFrequencyDigit1000->setText(f.at(2));
    ui->labelFrequencyDigit10000->setText(f.at(1));
    ui->labelFrequencyDigit100000->setText(f.at(0));
}

double XlabInterface::frequencyDigitsToFrequency()
{
    double result = 0;
    QString v;

    v = ui->labelFrequencyDigit100000->text();
    result += 10000000 * v.toInt();
    v = ui->labelFrequencyDigit10000->text();
    result += 1000000 * v.toInt();
    v = ui->labelFrequencyDigit1000->text();
    result += 100000 * v.toInt();
    v = ui->labelFrequencyDigit100->text();
    result += 10000 * v.toInt();
    v = ui->labelFrequencyDigit10->text();
    result += 1000 * v.toInt();
    v = ui->labelFrequencyDigit1->text();
    result += 100 * v.toInt();
    v = ui->labelFrequencyDigit01->text();
    result += 10 * v.toInt();
    v = ui->labelFrequencyDigit001->text();
    result += v.toInt();
    return result;
}

void XlabInterface::updateFrequency(double value)
{
    sendWaveFrequency(value);
    ui->multimeterdisplayAWGFrequency->setValue(frequencyRealValueCheck());
    ui->collapsiblePanelAWGFrequency->setText(QString::number(frequencyDigitsToFrequency() / 100, 'f', 2) + "Hz ");
}

void XlabInterface::on_frequencySlider_valueChanged(int value)
{
    double f;

    f = frequencySliderValueToInt(value);
    frequencyDigitsShowValue(f);
    updateFrequency(f);
}

void XlabInterface::setAwgFrequency(double f)
{
    frequencyDigitsShowValue(f * 100);
    frequencyDigitsValueChanged();
}

void XlabInterface::frequencyDigitsValueChanged()
{
    double value;

    ui->frequencySlider->blockSignals(true);
    value = frequencyDigitsToFrequency();
    ui->frequencySlider->setValue(frequencyIntToSliderValue(value));
    ui->frequencySlider->blockSignals(false);
    updateFrequency(value);
}

double XlabInterface::frequencySliderValueToInt(double value)
{
    double result;

    result = pow(10, value / 100);
    if (result >= 10000000) result = mathRoundTo(result, 100000);
    else if (result >= 1000000) result = mathRoundTo(result, 10000);
    else if (result >= 100000) result = mathRoundTo(result, 1000);
    else if (result >= 10000) result = mathRoundTo(result, 100);
    else if (result >= 1000) result = mathRoundTo(result, 10);
    result = dataValueRangeTrunk64(result, AWG_WAVE_MIN_Frequency, AWG_WAVE_MAX_Frequency);
    return result;
}

int XlabInterface::frequencyIntToSliderValue(double value)
{
    return log10(value) * 100;
}

void XlabInterface::on_toolButtonFrequencyDigit100000Up_clicked()
{
    frequencyDigitIncrease(ui->labelFrequencyDigit100000, 1);
    frequencyDigitsValueChanged();
}

void XlabInterface::on_toolButtonFrequencyDigit100000Down_clicked()
{
    frequencyDigitDecrease(ui->labelFrequencyDigit100000, 1);
    frequencyDigitsValueChanged();
}

void XlabInterface::on_toolButtonFrequencyDigit10000Up_clicked()
{
    frequencyDigitIncrease(ui->labelFrequencyDigit10000, 9);
    frequencyDigitsValueChanged();
}

void XlabInterface::on_toolButtonFrequencyDigit10000Down_clicked()
{
    frequencyDigitDecrease(ui->labelFrequencyDigit10000, 9);
    frequencyDigitsValueChanged();
}

void XlabInterface::on_toolButtonFrequencyDigit1000Up_clicked()
{
    frequencyDigitIncrease(ui->labelFrequencyDigit1000, 9);
    frequencyDigitsValueChanged();
}

void XlabInterface::on_toolButtonFrequencyDigit1000Down_clicked()
{
    frequencyDigitDecrease(ui->labelFrequencyDigit1000, 9);
    frequencyDigitsValueChanged();
}

void XlabInterface::on_toolButtonFrequencyDigit100Up_clicked()
{
    frequencyDigitIncrease(ui->labelFrequencyDigit100, 9);
    frequencyDigitsValueChanged();
}

void XlabInterface::on_toolButtonFrequencyDigit100Down_clicked()
{
    frequencyDigitDecrease(ui->labelFrequencyDigit100, 9);
    frequencyDigitsValueChanged();
}

void XlabInterface::on_toolButtonFrequencyDigit10Up_clicked()
{
    frequencyDigitIncrease(ui->labelFrequencyDigit10, 9);
    frequencyDigitsValueChanged();
}

void XlabInterface::on_toolButtonFrequencyDigit10Down_clicked()
{
    frequencyDigitDecrease(ui->labelFrequencyDigit10, 9);
    frequencyDigitsValueChanged();
}

void XlabInterface::on_toolButtonFrequencyDigit1Up_clicked()
{
    frequencyDigitIncrease(ui->labelFrequencyDigit1, 9);
    frequencyDigitsValueChanged();
}

void XlabInterface::on_toolButtonFrequencyDigit1Down_clicked()
{
    frequencyDigitDecrease(ui->labelFrequencyDigit1, 9);
    frequencyDigitsValueChanged();
}

void XlabInterface::on_toolButtonFrequencyDigit01Up_clicked()
{
    frequencyDigitIncrease(ui->labelFrequencyDigit01, 9);
    frequencyDigitsValueChanged();
}

void XlabInterface::on_toolButtonFrequencyDigit01Down_clicked()
{
    frequencyDigitDecrease(ui->labelFrequencyDigit01, 9);
    frequencyDigitsValueChanged();
}

void XlabInterface::on_toolButtonFrequencyDigit001Up_clicked()
{
    frequencyDigitIncrease(ui->labelFrequencyDigit001, 9);
    frequencyDigitsValueChanged();
}

void XlabInterface::on_toolButtonFrequencyDigit001Down_clicked()
{
    frequencyDigitDecrease(ui->labelFrequencyDigit001, 9);
    frequencyDigitsValueChanged();
}

void XlabInterface::frequencyDigitIncrease(QLabel *digit, int range)
{
    QString v;

    v = digit->text();
    if (v.toInt() > (range - 1)) digit->setText("0");
    else digit->setText(QString::number(v.toInt() + 1));
}

void XlabInterface::frequencyDigitDecrease(QLabel *digit, int range)
{
    QString v;

    v = digit->text();
    if (v.toInt() < 1) digit->setText(QString::number(range));
    else digit->setText(QString::number(v.toInt() - 1));
}

void XlabInterface::on_captureButton_clicked()
{
    captureRef = true;
}

void XlabInterface::on_screenShotButton_clicked()
{
    QPixmap sshot = ui->widget->grab(QRect(QPoint(0, 0), QSize(-1, ui->widget->height() - 20)));
    QString fileName = "screenShot" + QTime::currentTime().toString() + ".png";
    fileName.replace(":", "");
    if (filePath.isEmpty()) {
        filePath=QFileDialog::getExistingDirectory(this, tr("Choose Destination Folder"), defaultDir);
    }
    if (filePath.isEmpty())
        return;
    defaultDir = filePath;
    sshot.save(filePath + QDir::separator() + fileName);
    ui->notifLabel->setText("File " + fileName + " created!");
}

void XlabInterface::on_saveWave_clicked()
{
    saveWave = true;
}

void XlabInterface::on_loadWave_clicked()
{
    on_clearWaveButton_clicked();
    QString buffer;
    QString fpath;
    fpath = QFileDialog::getOpenFileName(this, tr("Open File"), defaultDir, "XLab files (*.xsp)");
    if (fpath.isEmpty())
        return;
    QFile waveData(fpath);
    if (waveData.open(QFile::ReadOnly)) {
        buffer = waveData.readAll();
        QStringList rawParsedData = buffer.split("<*>");
        rawParsedData.removeAll("\n");
        rawParsedData.removeAll("\r");
        QStringList parsedData = rawParsedData[0].split(",");
        for (int i = 1; i < parsedData.length() - 1; i++) {
            ch1SaveBuffer.push_back(parsedData[i].toDouble());
        }
        parsedData.clear();
        parsedData = rawParsedData[1].split(",");
        for (int i = 1; i < parsedData.length() - 1; i++) {
            ch2SaveBuffer.push_back(parsedData[i].toDouble());
        }
        for (int k = 0 ; k < 8; k++) {
            parsedData.clear();
            parsedData = rawParsedData[k + 2].split(",");
            for (int i = 1; i < parsedData.length() - 1; i++) {
                bitSaveBuffer[k].push_back(parsedData[i].toDouble());
            }
        }
        displayLoadedWave = true;
    }
}

void XlabInterface::on_clearWaveButton_clicked()
{
    displayLoadedWave = false;
    ch1SaveBuffer.clear();
    ch2SaveBuffer.clear();
    for (int k = 0; k < 8; k++) {
        bitSaveBuffer[k].clear();
    }
}

void XlabInterface::on_intensitySlider_valueChanged(int value)
{
    displayIntensitySliderChanged(value);
}

void XlabInterface::on_pauseSnifferButton_clicked()
{
    sniffPauseButtonChanged();
}

void XlabInterface::on_recordButton_clicked()
{
    if (!recordingWaves) {
        ui->recordButton->setText("Stop");
        ui->recordButton->setIcon(QIcon(":/icons/ico-record-active.png"));
        ui->label_13->setText("Recording wave.");
        recordingWaves = true;
        recordWaveFile->setFileName("record" + QDateTime::currentDateTime().toString("yyyy-MM-ddhh-mm-ss") + ".xsp");
        if (!recordWaveFile->open(QIODevice::Text | QIODevice::WriteOnly)) {
            recordingWaves = false;
            return;
        }
        recordWaveTextStream=new QTextStream(recordWaveFile);
        dataToSave = "";
    } else {
        ui->recordButton->setText("Record");
        ui->recordButton->setIcon(QIcon(":/icons/ico-record.png"));
        ui->label_13->setText("");
        (*recordWaveTextStream) << dataToSave;
        recordingWaves = false;
        delete recordWaveTextStream;
        recordWaveFile->close();
        ui->label_13->setText("");
        QMessageBox msgBox(QMessageBox::Information, "Success", "File " + recordWaveFile->fileName() + " saved.");
        msgBox.setWindowIcon(QIcon(":/icons/gt.ico"));
        msgBox.exec();
    }
}

void XlabInterface::resizeEvent(QResizeEvent *)
{
    updateLabelPosition();
    vCursorRefresh();
}

void XlabInterface::on_fcR10_clicked()
{
    sendTriggerSource();
}

void XlabInterface::on_fcR7_clicked()
{
    sendTriggerSource();
}

void XlabInterface::on_fcR6_clicked()
{
    sendTriggerSource();
}

void XlabInterface::on_fcR5_clicked()
{
    sendTriggerSource();
}

void XlabInterface::on_fcR4_clicked()
{
    sendTriggerSource();
}

void XlabInterface::on_fcR3_clicked()
{
    sendTriggerSource();
}

void XlabInterface::on_fcR2_clicked()
{
    sendTriggerSource();
}

void XlabInterface::on_fcR1_clicked()
{
    sendTriggerSource();
}

void XlabInterface::on_fcR0_clicked()
{
    sendTriggerSource();
}

void XlabInterface::on_fcR11_clicked()
{
    sendTriggerSource();
}

void XlabInterface::on_fcR11_2_clicked()
{
    sendGateSource();
}

void XlabInterface::on_fcR10_2_clicked()
{
    sendGateSource();
}

void XlabInterface::on_fcR7_2_clicked()
{
    sendGateSource();
}

void XlabInterface::on_fcR6_2_clicked()
{
    sendGateSource();
}

void XlabInterface::on_fcR5_2_clicked()
{
    sendGateSource();
}

void XlabInterface::on_fcR4_2_clicked()
{
    sendGateSource();
}

void XlabInterface::on_fcR3_2_clicked()
{
    sendGateSource();
}

void XlabInterface::on_fcR2_2_clicked()
{
    sendGateSource();
}

void XlabInterface::on_fcR1_2_clicked()
{
    sendGateSource();
}

void XlabInterface::on_fcR0_2_clicked()
{
    sendGateSource();
}

void XlabInterface::on_checkBoxCH1ACDC_clicked()
{
    channel1ACDCChanged(true);
}

void XlabInterface::on_checkBoxCH2ACDC_clicked()
{
    channel2ACDCChanged(true);
}

void XlabInterface::on_radioButtonSweepFrequency_clicked()
{
    ui->radioButtonSweepFrequency->setChecked(true);
    sweepTypeChanged();
}

void XlabInterface::on_radioButtonSweepAmplitude_clicked()
{
    ui->radioButtonSweepAmplitude->setChecked(true);
    sweepTypeChanged();
}

void XlabInterface::on_radioButtonSweepDutyCycle_clicked()
{
    ui->radioButtonSweepDutyCycle->setChecked(true);
    sweepTypeChanged();
}

void XlabInterface::on_radioButtonSweepOffset_clicked()
{
    ui->radioButtonSweepOffset->setChecked(true);
    sweepTypeChanged();
}

void XlabInterface::on_radioButtonSweepRising_clicked()
{
    ui->radioButtonSweepRising->setChecked(true);
    sendSweepTypeControls();
}

void XlabInterface::on_radioButtonSweepPingPong_clicked()
{
    ui->radioButtonSweepPingPong->setChecked(true);
    sendSweepTypeControls();
}

void XlabInterface::on_radioButtonSweepFalling_clicked()
{
    ui->radioButtonSweepFalling->setChecked(true);
    sendSweepTypeControls();
}

void XlabInterface::setAwgLevelsAmplitude(double val)
{
    ui->doubleSpinBoxAmp->setValue(val);
}

void XlabInterface::setAwgLevelsDutyCycle(double val)
{
    ui->doubleSpinBoxDuty->setValue(val);
}

void XlabInterface::setAwgLevelsOffset(double val)
{
    ui->doubleSpinBoxOffset->setValue(val);
}

void XlabInterface::setAwgSweepEnabled(bool enabled)
{
    ui->collapsiblePanelAWGSweep->setChecked(enabled);
}

void XlabInterface::clickRadioButtonSweepFrequency()
{
    ui->radioButtonSweepFrequency->click();
}

void XlabInterface::clickRadioButtonSweepAmplitude()
{
    ui->radioButtonSweepAmplitude->click();
}

void XlabInterface::clickRadioButtonSweepDutyCycle()
{
    ui->radioButtonSweepDutyCycle->click();
}

void XlabInterface::clickRadioButtonSweepOffset()
{
    ui->radioButtonSweepOffset->click();
}

void XlabInterface::clickRadioButtonSweepDirectionRising()
{
    ui->radioButtonSweepRising->click();
}

void XlabInterface::clickRadioButtonSweepDirectionPingPong()
{
    ui->radioButtonSweepPingPong->click();
}

void XlabInterface::clickRadioButtonSweepDirectionFaling()
{
    ui->radioButtonSweepFalling->click();
}

void XlabInterface::setAwgSweepSpeed(int val)
{
    ui->sweepSpeedSlider->setValue(val);
}

void XlabInterface::setAwgSweepStart(double val)
{
    ui->sweepStartFreqSlider->setValue(int(val));
}

void XlabInterface::setAwgSweepEnd(double val)
{
    ui->sweepEndFreqSlider->setValue(int(val));
}

void XlabInterface::on_radioButtonConstant_clicked()
{
    sendWaveType();
}

void XlabInterface::clickRadioButtonConstant()
{
    ui->radioButtonConstant->click();
}

void XlabInterface::on_pushButtonCalibrateADCOffset_clicked()
{
    usbDevice.controlWriteTransfer(USB_SEND_CALIBRATE, CALIBRATION_ADC_Auto);
}

void XlabInterface::on_pushButtonCalibrateADCOffsetClear_clicked()
{
    usbDevice.controlWriteTransfer(USB_SEND_CALIBRATE, CALIBRATION_ADC_Clear);
}

void XlabInterface::on_toolButtonDACOffsetPlus_clicked()
{
    usbDevice.controlWriteTransfer(USB_SEND_CALIBRATE, CALIBRATION_DAC_Offset, 1);
    ui->toolButtonDACOffsetOK->setEnabled(true);
}

void XlabInterface::on_toolButtonDACOffsetMinus_clicked()
{
    usbDevice.controlWriteTransfer(USB_SEND_CALIBRATE, CALIBRATION_DAC_Offset, -1);
    ui->toolButtonDACOffsetOK->setEnabled(true);
}

void XlabInterface::on_toolButtonDACOffsetOK_clicked()
{
    usbDevice.controlWriteTransfer(USB_SEND_CALIBRATE, CALIBRATION_DAC_Offset, 0);
    ui->toolButtonDACOffsetOK->setEnabled(false);
}

void XlabInterface::on_toolButtonDACGainPlus_clicked()
{
    usbDevice.controlWriteTransfer(USB_SEND_CALIBRATE, CALIBRATION_DAC_Gain, 1);
    ui->toolButtonDACGainOK->setEnabled(true);
}

void XlabInterface::on_toolButtonDACGainMinus_clicked()
{
    usbDevice.controlWriteTransfer(USB_SEND_CALIBRATE, CALIBRATION_DAC_Gain, -1);
    ui->toolButtonDACGainOK->setEnabled(true);
}

void XlabInterface::on_toolButtonDACGainOK_clicked()
{
    usbDevice.controlWriteTransfer(USB_SEND_CALIBRATE, CALIBRATION_DAC_Gain, 0);
    ui->toolButtonDACGainOK->setEnabled(false);
}

void XlabInterface::on_toolButtonCH1GainPlus_clicked()
{
    ui->toolButtonCH1GainOK->setEnabled(true);
}

void XlabInterface::on_toolButtonCH1GainMinus_clicked()
{
    ui->toolButtonCH1GainOK->setEnabled(true);
}

void XlabInterface::on_toolButtonCH1GainOK_clicked()
{
    ui->toolButtonCH1GainOK->setEnabled(false);
}

void XlabInterface::on_toolButtonCH2GainPlus_clicked()
{
    ui->toolButtonCH2GainOK->setEnabled(true);
}

void XlabInterface::on_toolButtonCH2GainMinus_clicked()
{
    ui->toolButtonCH2GainOK->setEnabled(true);
}

void XlabInterface::on_toolButtonCH2GainOK_clicked()
{
    ui->toolButtonCH2GainOK->setEnabled(false);
}

void XlabInterface::on_comboBoxScopeCursorsHorizontalType_currentIndexChanged(int index)
{
    cursorsHorizontalTypeSet(index);
}

void XlabInterface::on_spinBoxConstant_valueChanged(int arg1)
{
    ui->horizontalSliderConstant->setSliderPosition(arg1);
}

void XlabInterface::on_horizontalSliderConstant_valueChanged(int value)
{
    ui->spinBoxConstant->setValue(value);
    sendWaveConstantLevel(value);
}

void XlabInterface::on_toolButtonCh1ADC24V_clicked()
{
    ui->toolButtonCh1ADC2V4->setChecked(false);
    ui->ch1GainSlider->setValue(0);
    channel1GainSliderChanged(ui->ch1GainSlider->value(), false);
    ui->checkBoxCH1ACDC->setChecked(false);
    channel1ACDCChanged(false);
}

void XlabInterface::on_toolButtonCh1ADC2V4_clicked()
{
    ui->toolButtonCh1ADC24V->setChecked(false);
    ui->ch1GainSlider->setValue(3);
    channel1GainSliderChanged(ui->ch1GainSlider->value(), false);
    ui->checkBoxCH1ACDC->setChecked(false);
    channel1ACDCChanged(false);
}

void XlabInterface::on_toolButtonCh2ADC24V_clicked()
{
    ui->toolButtonCh2ADC2V4->setChecked(false);
    ui->ch2GainSlider->setValue(0);
    channel2GainSliderChanged(ui->ch2GainSlider->value(), false);
    ui->checkBoxCH2ACDC->setChecked(false);
    channel2ACDCChanged(false);
}

void XlabInterface::on_toolButtonCh2ADC2V4_clicked()
{
    ui->toolButtonCh2ADC24V->setChecked(false);
    ui->ch2GainSlider->setValue(3);
    channel2GainSliderChanged(ui->ch2GainSlider->value(), false);
    ui->checkBoxCH2ACDC->setChecked(false);
    channel2ACDCChanged(false);
}

void XlabInterface::on_groupBoxCH2AC_clicked()
{
    ui->checkBoxCH2Trace->setChecked(ui->groupBoxCH2AC->isChecked());
    ui->groupBoxCH2DC->setChecked(ui->groupBoxCH2AC->isChecked());
    channel2TraceChanged(ui->checkBoxCH2Trace->isChecked());
}

void XlabInterface::on_groupBoxCH1AC_clicked()
{
    ui->checkBoxCH1Trace->setChecked(ui->groupBoxCH1AC->isChecked());
    ui->groupBoxCH1DC->setChecked(ui->groupBoxCH1AC->isChecked());
    channel1TraceChanged(ui->checkBoxCH1Trace->isChecked());
}

void XlabInterface::on_groupBoxCH1DC_clicked()
{
    ui->checkBoxCH1Trace->setChecked(ui->groupBoxCH1DC->isChecked());
    ui->groupBoxCH1AC->setChecked(ui->groupBoxCH1DC->isChecked());
    channel1TraceChanged(ui->checkBoxCH1Trace->isChecked());
}

void XlabInterface::on_groupBoxCH2DC_clicked()
{
    ui->checkBoxCH2Trace->setChecked(ui->groupBoxCH2DC->isChecked());
    ui->groupBoxCH2AC->setChecked(ui->groupBoxCH2DC->isChecked());
    channel2TraceChanged(ui->checkBoxCH2Trace->isChecked());
}

void XlabInterface::on_checkBoxScopeCursorsHorizontal_clicked()
{
    cursorsHorizontalChanged(ui->checkBoxScopeCursorsHorizontal->isChecked());
}

void XlabInterface::on_chdSizeSlider_valueChanged()
{
    channelLogicSizeSliderChanged();
}

void XlabInterface::on_checkBoxAWBEnable_clicked()
{
    awgEnableChanged();
}

void XlabInterface::on_comboBoxDACBuffer_currentIndexChanged(int index)
{
    awgDACBufferChanged(index);
}

void XlabInterface::on_comboBoxADCBuffer_currentIndexChanged(int index)
{
    scopeADCBufferChanged(index);
}

void XlabInterface::on_checkBoxmeterondev_clicked()
{
    sendMixedControls();
    m_repaint = true;
}

void XlabInterface::on_checkBoxFreqFastMode_clicked()
{
    sendMeterFrequencyControl();
}

void XlabInterface::on_radioFrequencyGatePeriod_clicked()
{
    sendMeterFrequencyControl();
}

void XlabInterface::on_radioFrequencyGatePulse_clicked()
{
    sendMeterFrequencyControl();
}

void XlabInterface::on_pushButtonSnifferClear_clicked()
{
    sendSnifferClear();
}

void XlabInterface::on_comboBoxSniffParallelStrobe_currentIndexChanged(int index)
{
    sendSnifferStrobeEdge(index);
}

void XlabInterface::on_checkBoxSniffParallelInputInvert_clicked()
{
    sendSnifferOptionControl();
}

void XlabInterface::on_collapsiblePanelAWGWaveform_collapsedChanged(bool collapsed)
{
    ui->frameAWGWaveform->setVisible(!collapsed);
    if (collapsed) {
        ui->collapsiblePanelAWGWaveform->setTextVisible(true);
    } else {
        ui->collapsiblePanelAWGWaveform->setTextVisible(false);
    }
}

void XlabInterface::on_collapsiblePanelAWGFrequency_collapsedChanged(bool collapsed)
{
    ui->frameAWGFrequency->setVisible(!collapsed);
    if (collapsed) {
        ui->collapsiblePanelAWGFrequency->setTextVisible(true);
    } else {
        ui->collapsiblePanelAWGFrequency->setTextVisible(false);
    }
}

void XlabInterface::on_collapsiblePanelAWGLevels_collapsedChanged(bool collapsed)
{
    ui->frameAWGLevels->setVisible(!collapsed);
}

void XlabInterface::on_collapsiblePanelAWGSweep_checkedChanged(bool checked)
{
    ui->frameAWGSweep->setEnabled(checked);
    sweepCursorsUpdate();
    sendSweepTypeControls();
}

void XlabInterface::on_collapsiblePanelAWGSweep_collapsedChanged(bool collapsed)
{
    ui->frameAWGSweep->setVisible(!collapsed);
}

void XlabInterface::on_collapsiblePanelScopeCursors_collapsedChanged(bool collapsed)
{
    ui->frameScopeCursors->setVisible(!collapsed);
}

void XlabInterface::on_collapsiblePanelScopeDisplay_collapsedChanged(bool collapsed)
{
    ui->frameScopeDisplay->setVisible(!collapsed);
}

void XlabInterface::on_collapsiblePanelScopeRecord_collapsedChanged(bool collapsed)
{
    ui->frameScopeRecord->setVisible(!collapsed);
}

void XlabInterface::on_actionExit_triggered()
{
    logToFile("Exit triggered", LOGGING_LEVEL_Warning);
    closeApplication();
    exit(0);
}

void XlabInterface::on_actionAbout_triggered()
{
    about = new About(this, tomap_GlobalOptions);
    about->exec();
}

void XlabInterface::on_actionConsole_triggered()
{
    qDebug() << "Check Python console status";
    if (pyconsole == nullptr) qDebug() << "Create new Python console";
    else if ((pyconsole != nullptr) && (!pythonConsoleIsRunning)) qDebug() << "Recreate Python console";
    else {
        qDebug() << "Show Ptyhon console";
        pyconsole->setVisible(true);
        return;
    }
    pyconsole = new PythonConsole(this, tomap_PythonConsoleOptions, &pythonConsoleIsRunning);
    pyconsole->setAttribute(Qt::WA_DeleteOnClose);
    connect(pyconsole, SIGNAL(configurationRequest(QString)), this, SLOT(callSettings(QString)));
    pyconsole->show();
}

void XlabInterface::on_actionSettings_triggered()
{
    callSettings("");
}

void XlabInterface::callSettings(QString item)
{
    qDebug() << "Call settings by: " << item;
    settingsdialog = new Settings(this, map_Option, item);
    connect(settingsdialog, SIGNAL(applyRequest()), this, SLOT(applyModuleConfiguration()));
    settingsdialog->exec();
}

void XlabInterface::applyModuleConfiguration()
{
    emit applyConfigurationRequest();
}

void XlabInterface::applyScopeConfiguration()
{
    displayThemeChanged(tomap_ScopeOptions->theme);
    logicLabelPositionSet(tomap_ScopeOptions->logicLabelPosition);
    displayGridStyleChanged(tomap_ScopeOptions->gridStyle);
    m_repaint = true;

}

